#include <stdio.h>
#include <math.h>
#include <mpi.h>
#include "rule75.h"
#pragma hdrstop

#define useMPI 2 /*  0 => PC, 1 => Unix, 2 => cluster  */

#define HSnum 2           /*  health states  */
#define rounder 0.01
#define zeroprobwages 0.00000005 /*  Used to skip states in integration  */
#define zeroprobhc 0.0000000001
#define negpen 1000       /*  scales penalty for violating liquidity constraint */
#define miserpen 1000000  /*  scales penalty for exceeding asset maximum  */
#define xvstep 3          /*  Gap between "main body" of search area and extreme values  */
#define nopart 0
#define doespart 1
#define notransfers 0
#define dieage 95
#define penavailage 95    /*  Pensions are treated as fully liquid from here on  */

/*  Tax brackets and marginal tax rates  */
#define tbk1 6250
#define tbk2 40200
#define tbk3 68400
#define tbk4 93950
#define tbk5 148250
#define tbk6 284700
#define mtr1 0.0765
#define mtr2 0.2616
#define mtr3 0.4119
#define mtr4 0.3499
#define mtr5 0.3834
#define mtr6 0.4360
#define mtr7 0.4761

/*  After-tax income at bracket points:  */
/*  ati(j) = ati(j-1) + [1-mtr(j)]*[tbk(j)-tbk(j-1)]  */
#define ati1 5771.88
#define ati2 30840.56
#define ati3 47424.98
#define ati4 64035.03
#define ati5 97516.41
#define ati6 174474.21

/*  AIME and PIA  */
#define AIMEage 60
#define startwork 25
#define AIMEyrs 35      /*  AIMEage - startwork  */
#define disabincometest 3600  

/*  HRS Coefficients for spousal income equation  */

#define spicns   476640.6
#define spiage1 -24223.72
#define spiage2  460.255 
#define spiage3 -3.785257
#define spiage4  0.011295
#define spiage5  0.0  
#define spibadh  1030.71  
#define spibadhage -14.60367
#define cpi8798  1.0  

/*  PSID Coefficients for spousal income equation  */ 
/* 
#define spicns  -35856.04
#define spiage1  3579.82
#define spiage2 -137.8337 
#define spiage3  2.791257
#define spiage4  -.0273952  
#define spiage5  0.0001012  
#define spibadh  0.0  
#define spibadhage 0.0
#define cpi8798  1.354  
*/
/*----------------------------------GLOBALS----------------------------------*/
/*  Used in interpolation and quadrature  */
    double Amovein1, Amovein2, AIMEmovein1, AIMEmovein2, _Aint, _AIMEint;
    int Anint, AIMEnint, _Anum, _Wnum, AIMEnum, hcnum, ndnumHC, HITnum, 
        PrefTypeNum, HasEntryCost;
/*  Used in the utility function and bequest functions  */
    double conspref, _nu, _gamma, leispref, leisprefbad, fixcost, L, 
           bequest, Onemnu, Onemgamma, beqk, EntryPenalty, keepjob;
/*  Relates to Preference Heterogeneity or Job Tenure */
    int PrefType, HasTenure;
/*  Used in the budget constraint  */
    double earnlev, taxfrac, ror, cfloor, _Afloor, AIMEaccrue, agep, 
           AIMEmax, AIMEbk1, AIMEbk2, mpr1, mpr2, mpr3, PIA1, PIA2, 
           penbalfrac, nppenbalfrac, penbenfrac, penbenprob, eretadj,  
           cumeretadj, bigcred, litcred, nospinc, AIMElowfrac;
    int nret, etstage, penbensstart, manretage, mappage, eretage, medicage,
        age, apply, tiedwage, cfloorindx;
/*  Used in computing the value function  */
    double _beta, survprobpen;
/*  Miscellaneous  */
    double _Amax, fracAR1, fracwn, rhohc, hminsim, aggwagegrowth;

/*---------------------------------------------------------------------------*/
/*---------------------------------------------------------------------------*/
/*  MAIN:  entry point for the main program  */

int main(int argc, char *argv[]) 
   {
    unsigned int recsize, searchtype;
    int gotderef, gotrules, rank, size, statespermachine, iAssetsmin,
        iAssetsmax;
    double spm2;
    double *tempvec;
    double ********valftpoMat, ********clevmaxMat, ********clevlazMat, 
           ********maxconMat, ********lazconMat, ********hoursMat, 
           ********partMat, ********valfMat, ********holdclevmaxMat, 
           ********holdclevlazMat, ********holdmaxconMat,
           ********holdlazconMat, ********holdhoursMat,
           ********pensionMat;
 
    char fullpath[70];

    Matrix interpparmPtr, gridptnumPtr, searchparmPtr, prefparmPtr, 
           ageparmPtr, miscparmPtr, sscredsPtr, AstatePtr, WstatePtr, 
           AIMstatePtr, hcstatePtr, cdecPtr, wagetransprobPtr, 
           HITtransprobwPtr, HITtransprobnwPtr, hctransprobPtr, 
           rawhcnodesPtr, hcwgtsPtr, pencfsPtr, penaccPtr, penaccprobPtr, 
           AIMEbkPtr, mprPtr, PIAbkPtr, penkinkPtr, agePtr, applyPtr, 
           wagetrendPtr, survprobPtr, probbadPtr, mnlnhcwPtr, mnlnhcnwPtr, 
           stderrhcwPtr, stderrhcnwPtr, clevmaxPtr, clevlazPtr, valftpoPtr, 
           maxconPtr, lazconPtr, hoursPtr, partPtr, valfPtr, holdclevmaxPtr, 
           holdclevlazPtr, holdmaxconPtr, holdlazconPtr, holdhoursPtr,
           pensionPtr;

    char rootdir[] = "/home/jjones/ssretire_final/iofiles/";  /*  Unix or Cluster  */
    if (useMPI==0)  /*  PC  */
       {
        strcpy(rootdir,"d:\\ssretire\\iofiles\\");
       }

    rank = 1;
    size = 1;

    if (useMPI==2)
       {
        MPI_Init(&argc, &argv);
        MPI_Comm_size(MPI_COMM_WORLD, &size);
        MPI_Comm_rank(MPI_COMM_WORLD, &rank);
       }

/*  Read in parameter vectors from the *.fmt files  */

    interpparmPtr = gau5read(strcat(strcpy(fullpath,rootdir),"interpol.fmt"));
    gridptnumPtr  = gau5read(strcat(strcpy(fullpath,rootdir),"gridpnum.fmt"));
    prefparmPtr   = gau5read(strcat(strcpy(fullpath,rootdir),"prefparm.fmt"));
    ageparmPtr    = gau5read(strcat(strcpy(fullpath,rootdir),"ageparm.fmt"));
    miscparmPtr   = gau5read(strcat(strcpy(fullpath,rootdir),"miscparm.fmt"));
    applyPtr      = gau5read(strcat(strcpy(fullpath,rootdir),"apply.fmt"));
    survprobPtr   = gau5read(strcat(strcpy(fullpath,rootdir),"survprob.fmt"));
    agePtr        = gau5read(strcat(strcpy(fullpath,rootdir),"age.fmt"));
    AIMEbkPtr     = gau5read(strcat(strcpy(fullpath,rootdir),"AIMEbend.fmt"));
    mprPtr        = gau5read(strcat(strcpy(fullpath,rootdir),"mpr.fmt"));
    PIAbkPtr      = gau5read(strcat(strcpy(fullpath,rootdir),"PIAbend.fmt"));

    searchparmPtr = gau5read(strcat(strcpy(fullpath,rootdir),"srchparm.fmt"));
    sscredsPtr    = gau5read(strcat(strcpy(fullpath,rootdir),"sscreds.fmt"));    
    AstatePtr     = gau5read(strcat(strcpy(fullpath,rootdir),"Astate.fmt"));
    WstatePtr     = gau5read(strcat(strcpy(fullpath,rootdir),"Wstate.fmt"));
    AIMstatePtr   = gau5read(strcat(strcpy(fullpath,rootdir),"AIMstate.fmt"));
    hcstatePtr    = gau5read(strcat(strcpy(fullpath,rootdir),"hcstate.fmt"));
    cdecPtr       = gau5read(strcat(strcpy(fullpath,rootdir),"cdec.fmt")); 

    wagetransprobPtr  = gau5read(strcat(strcpy(fullpath,rootdir),"wtrmat.fmt")); 
    HITtransprobwPtr  = gau5read(strcat(strcpy(fullpath,rootdir),"hitranw.fmt")); 
    HITtransprobnwPtr = gau5read(strcat(strcpy(fullpath,rootdir),"hitrannw.fmt")); 
    hctransprobPtr    = gau5read(strcat(strcpy(fullpath,rootdir),"hctrmat.fmt")); 
    rawhcnodesPtr = gau5read(strcat(strcpy(fullpath,rootdir),"hcnodes.fmt")); 
    hcwgtsPtr     = gau5read(strcat(strcpy(fullpath,rootdir),"hcwgts.fmt")); 

    pencfsPtr     = gau5read(strcat(strcpy(fullpath,rootdir),"pencfs.fmt")); 
    penaccPtr     = gau5read(strcat(strcpy(fullpath,rootdir),"penacc.fmt")); 
    penaccprobPtr = gau5read(strcat(strcpy(fullpath,rootdir),"paccprob.fmt")); 
    penkinkPtr    = gau5read(strcat(strcpy(fullpath,rootdir),"pbbend.fmt")); 
    wagetrendPtr  = gau5read(strcat(strcpy(fullpath,rootdir),"wagetrnd.fmt")); 
    probbadPtr    = gau5read(strcat(strcpy(fullpath,rootdir),"probbad.fmt")); 

    mnlnhcwPtr    = gau5read(strcat(strcpy(fullpath,rootdir),"mnlnhcw.fmt")); 
    mnlnhcnwPtr   = gau5read(strcat(strcpy(fullpath,rootdir),"mnlnhcnw.fmt"));
    stderrhcwPtr  = gau5read(strcat(strcpy(fullpath,rootdir),"stdhcw.fmt")); 
    stderrhcnwPtr = gau5read(strcat(strcpy(fullpath,rootdir),"stdhcnw.fmt")); 

/*  Update the C++ globals */    
    gotderef = globderef(interpparmPtr.data, gridptnumPtr.data, 
                         ageparmPtr.data, miscparmPtr.data, 
                         applyPtr.data, survprobPtr.data, 
                         agePtr.data, AIMEbkPtr.data,                       
                         mprPtr.data, PIAbkPtr.data);

    valftpoPtr = gau5read(strcat(strcpy(fullpath,rootdir),"vtpovec.fmt"));
    valftpoMat = VectorToMatrix(valftpoPtr.data);	

/*  Now read existing (time-t+1) decisions and set up output vectors  */
/*  These will vary by search methodology  */

    searchtype    = (int) floor(rounder+*(searchparmPtr.data));
    if (searchtype < 3)
       {
        clevmaxPtr = gau5read(strcat(strcpy(fullpath,rootdir),"clevmax.fmt")); 
        clevlazPtr = gau5read(strcat(strcpy(fullpath,rootdir),"clevlaz.fmt")); 
        clevmaxMat = VectorToMatrix(clevmaxPtr.data);	
        clevlazMat = VectorToMatrix(clevlazPtr.data);	
       }
    else
       {
        maxconPtr = gau5read(strcat(strcpy(fullpath,rootdir),"maxcons.fmt")); 
        lazconPtr = gau5read(strcat(strcpy(fullpath,rootdir),"lazycons.fmt")); 
        hoursPtr  = gau5read(strcat(strcpy(fullpath,rootdir),"maxhours.fmt"));
        maxconMat = VectorToMatrix(maxconPtr.data);	
        lazconMat = VectorToMatrix(lazconPtr.data);	
        hoursMat  = VectorToMatrix(hoursPtr.data);	
       }
	
/*  Set up output vectors  */
/*  These will vary by search methodology  */
    recsize = (valftpoPtr.m)*(valftpoPtr.n);  
    if (searchtype < 3)
       {
        maxconPtr  = zeromat(recsize);
        lazconPtr  = zeromat(recsize);
        hoursPtr   = zeromat(recsize);
        maxconMat  = VectorToMatrix(maxconPtr.data);
        lazconMat  = VectorToMatrix(lazconPtr.data);
        hoursMat   = VectorToMatrix(hoursPtr.data);
       }
    else
       {
        clevmaxPtr = zeromat(recsize);
        clevlazPtr = zeromat(recsize);
        clevmaxMat = VectorToMatrix(clevmaxPtr.data);
        clevlazMat = VectorToMatrix(clevlazPtr.data);
       }

    partPtr        = zeromat(recsize);
    valfPtr        = zeromat(recsize);
    holdclevmaxPtr = zeromat(recsize);
    holdclevlazPtr = zeromat(recsize);
    holdmaxconPtr  = zeromat(recsize);
    holdlazconPtr  = zeromat(recsize);
    holdhoursPtr   = zeromat(recsize);
    pensionPtr     = zeromat(recsize);

    partMat        = VectorToMatrix(partPtr.data);
    valfMat        = VectorToMatrix(valfPtr.data);
    holdclevmaxMat = VectorToMatrix(holdclevmaxPtr.data);
    holdclevlazMat = VectorToMatrix(holdclevlazPtr.data);
    holdmaxconMat  = VectorToMatrix(holdmaxconPtr.data);
    holdlazconMat  = VectorToMatrix(holdlazconPtr.data);
    holdhoursMat   = VectorToMatrix(holdhoursPtr.data);
    pensionMat     = VectorToMatrix(pensionPtr.data);

    tempvec = (double *)calloc(recsize,sizeof(double));


    if (useMPI==2)
       {
        spm2 = ((double) _Anum)/ ((double) size);
        statespermachine = (int) ceil(spm2);
        iAssetsmin = rank*statespermachine;
        iAssetsmax = (rank+1)*statespermachine;
        if (iAssetsmax > _Anum) iAssetsmax = _Anum;
       }
    else
       {
        iAssetsmin=0;
        iAssetsmax=_Anum;
       }

/*  Now we get the decision rules, which will be stored  */ 
/*  vector parts of the maxtrix structures  */
    gotrules = getrules(
        searchparmPtr.data, sscredsPtr.data, AstatePtr.data, 
        WstatePtr.data, AIMstatePtr.data, hcstatePtr.data, 
        cdecPtr.data, wagetransprobPtr.data, HITtransprobwPtr.data, 
        HITtransprobnwPtr.data, hctransprobPtr.data, 
        rawhcnodesPtr.data, hcwgtsPtr.data, pencfsPtr.data, 
        penaccPtr.data, penaccprobPtr.data, penkinkPtr.data, 
        wagetrendPtr.data, probbadPtr.data, mnlnhcwPtr.data, 
        mnlnhcnwPtr.data, stderrhcwPtr.data, stderrhcnwPtr.data, 
        clevmaxMat, clevlazMat, holdclevmaxMat, holdclevlazMat, 
        valftpoMat, maxconMat, lazconMat, hoursMat, holdmaxconMat, 
        holdlazconMat, holdhoursMat, partMat, valfMat, pensionMat,
        survprobPtr.data, prefparmPtr.data, iAssetsmin, iAssetsmax);



    if (useMPI==2)
       {
        MPI_Reduce((void *) holdmaxconPtr.data, (void *) tempvec, recsize, MPI_DOUBLE,
                   MPI_SUM, size - 1, MPI_COMM_WORLD);
        switchem(&holdmaxconPtr.data, &tempvec);
	
        MPI_Reduce((void *) holdlazconPtr.data, (void *) tempvec, recsize, MPI_DOUBLE,
                   MPI_SUM, size - 1, MPI_COMM_WORLD);
        switchem(&holdlazconPtr.data, &tempvec);

        MPI_Reduce((void *) holdhoursPtr.data, (void *) tempvec, recsize, MPI_DOUBLE,
                   MPI_SUM, size - 1, MPI_COMM_WORLD);
        switchem(&holdhoursPtr.data, &tempvec);

        MPI_Reduce((void *) partPtr.data, (void *) tempvec, recsize, MPI_DOUBLE,
                   MPI_SUM, size - 1, MPI_COMM_WORLD);
        switchem(&partPtr.data, &tempvec);

        MPI_Reduce((void *) valfPtr.data, (void *) tempvec, recsize, MPI_DOUBLE,
                   MPI_SUM, size - 1, MPI_COMM_WORLD);
        switchem(&valfPtr.data, &tempvec);

        MPI_Reduce((void *) pensionPtr.data, (void *) tempvec, recsize, MPI_DOUBLE,
                   MPI_SUM, size - 1, MPI_COMM_WORLD);
        switchem(&pensionPtr.data, &tempvec);

        if (searchtype < 3)
           {
            MPI_Reduce((void *) holdclevmaxPtr.data, (void *) tempvec, recsize, MPI_DOUBLE,
                       MPI_SUM, size - 1, MPI_COMM_WORLD);
            switchem(&holdclevmaxPtr.data, &tempvec);

            MPI_Reduce((void *) holdclevlazPtr.data, (void *) tempvec, recsize, MPI_DOUBLE,
                       MPI_SUM, size - 1, MPI_COMM_WORLD);
            switchem(&holdclevlazPtr.data, &tempvec);
           }
       }

/*  Write the data to *.fmt files that GAUSS can read  */

    if ( (useMPI<2)||(rank==(size-1)) )
       {
        gau5write(strcat(strcpy(fullpath,rootdir),"maxcons.fmt"), holdmaxconPtr); 
        gau5write(strcat(strcpy(fullpath,rootdir),"lazycons.fmt"), holdlazconPtr); 
        gau5write(strcat(strcpy(fullpath,rootdir),"maxhours.fmt"), holdhoursPtr); 
        gau5write(strcat(strcpy(fullpath,rootdir),"partvec.fmt"), partPtr); 
        gau5write(strcat(strcpy(fullpath,rootdir),"Vvec.fmt"), valfPtr);
        gau5write(strcat(strcpy(fullpath,rootdir),"imppen.fmt"), pensionPtr); 

        if (searchtype < 3)
           {
            gau5write(strcat(strcpy(fullpath,rootdir),"clevmax.fmt"), holdclevmaxPtr); 
            gau5write(strcat(strcpy(fullpath,rootdir),"clevlaz.fmt"), holdclevlazPtr); 
           }
       }

    MPI_Finalize();

    return (0);
   }
/*---------------------------------------------------------------------------*/
/*---------------------------------------------------------------------------*/
/*  ZEROMAT:  Initializes matrix structure and fills it with zeros  */

Matrix zeromat(unsigned int recsize)
   {
    Matrix mat = {1, 1};
    mat.m = recsize;
    mat.n = 1;
    mat.data = (double *)calloc(recsize,sizeof(double));
    return mat;
   }
/*---------------------------------------------------------------------------*/
/*---------------------------------------------------------------------------*/
/*  ZEROMAT2:  Initializes multidimensional array and fills it with zeros  */

double ********zeromat2()
   {

    int iType, iTenure, iHIT, iHS, iHC, iAIME, iWage;
    double ********dataMat;

    dataMat = (double ********)malloc(PrefTypeNum*sizeof(double *******));

    for(iType=0; iType<PrefTypeNum; iType++)
       {
        dataMat[iType] = (double *******)malloc((HasEntryCost+1)*sizeof(double ******));
        for(iTenure=0; iTenure<(HasEntryCost+1); iTenure++)
           {
            dataMat[iType][iTenure] = (double ******)malloc(HITnum*sizeof(double *****));

            for(iHIT=0; iHIT<HITnum; iHIT++)
               {
                dataMat[iType][iTenure][iHIT] = (double *****)malloc(HSnum*sizeof(double ****));
                for(iHS=0; iHS<HSnum; iHS++)
                   {
                    dataMat[iType][iTenure][iHIT][iHS] = 
                        (double ****)malloc(hcnum*sizeof(double ***));
                    for(iHC=0; iHC<hcnum; iHC++)
                       {
                        dataMat[iType][iTenure][iHIT][iHS][iHC] = 
                            (double ***)malloc(AIMEnum*sizeof(double **));
                        for(iAIME=0; iAIME<AIMEnum; iAIME++)
                           {
                            dataMat[iType][iTenure][iHIT][iHS][iHC][iAIME] = 
                                (double **)malloc(_Wnum*sizeof(double *));
                            for(iWage=0; iWage<_Wnum; iWage++) 
                               {
                                dataMat[iType][iTenure][iHIT][iHS][iHC][iAIME][iWage] = (double *)calloc(_Anum,sizeof(double));;
                               }  /*  End loop through wages  */
                           }  /*  End loop through AIME  */
                       }  /*  End loop through persistent health costs  */
                   }  /*  End loop through health states  */
               }  /*  End loop through health insurance types  */ 
           }  /*  End loop through job tenure statuses  */
       }  /*  End loop through preference types  */ 

    return dataMat;
   }

/*---------------------------------------------------------------------------*/
/*---------------------------------------------------------------------------*/
/*  SWITCHEM:  Switches two pointer addresses  */

void switchem(double **dataPtr, double **tempPtr)
   {
    double *tempPtr2;
    tempPtr2 = *dataPtr;  /*  dataPtr, tempPtr are pointers to pointers.  */
    *dataPtr = *tempPtr;  /*  Dereferencing them yields memory addresses, */
    *tempPtr = tempPtr2;  /*  rather than values.  */
   }

/*---------------------------------------------------------------------------*/
/*---------------------------------------------------------------------------*/
/*  VECTORTOMATRIX:  Converts column vector into multidimensional array.     */
/*                   This is NOT a smart program.  The user will have to     */
/*                   ensure compatibility.                                   */

double ********VectorToMatrix(double *dataVec)
   {
    int i, iType, iTenure, iHIT, iHS, iHC, iAIME, iWage;
    double ********dataMat;

    i = 0;  /*  Position in vector  */
    dataMat = (double ********)malloc(PrefTypeNum*sizeof(double *******));

    for(iType=0; iType<PrefTypeNum; iType++)
       {
        dataMat[iType] = (double *******)malloc((HasEntryCost+1)*sizeof(double ******));
        for(iTenure=0; iTenure<(HasEntryCost+1); iTenure++)
           {
            dataMat[iType][iTenure] = (double ******)malloc(HITnum*sizeof(double *****));

            for(iHIT=0; iHIT<HITnum; iHIT++)
               {
                dataMat[iType][iTenure][iHIT] = (double *****)malloc(HSnum*sizeof(double ****));
                for(iHS=0; iHS<HSnum; iHS++)
                   {
                    dataMat[iType][iTenure][iHIT][iHS] = 
                        (double ****)malloc(hcnum*sizeof(double ***));
                    for(iHC=0; iHC<hcnum; iHC++)
                       {
                        dataMat[iType][iTenure][iHIT][iHS][iHC] = 
                            (double ***)malloc(AIMEnum*sizeof(double **));
                        for(iAIME=0; iAIME<AIMEnum; iAIME++)
                           {
                            dataMat[iType][iTenure][iHIT][iHS][iHC][iAIME] = 
                                (double **)malloc(_Wnum*sizeof(double *));
                            for(iWage=0; iWage<_Wnum; iWage++) 
                               {
                            /*  Let dataMat share an _Anum-sized block of memory with datavec.  */
                            /*  This block of the state space will vary only by asset level.  */
                                i = iType*(_Anum*_Wnum*AIMEnum*hcnum*HSnum*HITnum*(HasEntryCost+1)) +
                                    iTenure*(_Anum*_Wnum*AIMEnum*hcnum*HSnum*HITnum) +
                                    iHIT*(_Anum*_Wnum*AIMEnum*hcnum*HSnum) + 
                                    iHS*(_Anum*_Wnum*AIMEnum*hcnum) +
                                    iHC*(_Anum*_Wnum*AIMEnum) + 
                                    iAIME*(_Anum*_Wnum) + iWage*_Anum;
                                dataMat[iType][iTenure][iHIT][iHS][iHC][iAIME][iWage] = &dataVec[i];
                               }  /*  End loop through wages  */
                           }  /*  End loop through AIME  */
                       }  /*  End loop through persistent health costs  */
                   }  /*  End loop through health states  */
               }  /*  End loop through health insurance types  */ 
           }  /*  End loop through job tenure statuses  */
       }  /*  End loop through preference types  */ 

    return dataMat;
   }

/*---------------------------------------------------------------------------*/
/*---------------------------------------------------------------------------*/
/* U:  Utility function  */

double U(double c, double h, double particip, int badheal)
   {
    double leisure, utils, within, ReentryCost;

    ReentryCost = 0;
    if ((HasEntryCost==1) && (HasTenure==0) && (h>=keepjob))
       {
        ReentryCost = EntryPenalty; 
       }
    leisure = leispref - h - (fixcost*particip) - (leisprefbad*badheal)
              - ReentryCost;

    within  = pow(c,_gamma)*pow(leisure,Onemgamma);

    if (_nu == 1)
        utils = conspref*log(within);
    else
        utils = conspref*pow(within,Onemnu)/Onemnu;
        
    return utils;
   }

/*---------------------------------------------------------------------------*/
/*---------------------------------------------------------------------------*/
/* BEQ:  Bequest function  */

double beq(double assets)
   {
    double val;

    if (assets<0) assets=0;

    if (_nu == 1)
        val=log(assets+beqk);
    else
        val=pow(assets+beqk,Onemnu*_gamma)/Onemnu;

    val=val*bequest*conspref;

    return val;
   }

/*---------------------------------------------------------------------------*/
/*---------------------------------------------------------------------------*/
/*  ASS: Finds next period's assets, while checking earnings test  */
 
double ass(double AfterTaxIncome, double c, double laborincome, double assets, 
           double *reduc, double ssbenes)
   {
    double nextass;

/*  Didn't apply for benefits  */
    if (apply==0)
       {
        nextass = assets+AfterTaxIncome+ssbenes-c;
    /*  ssbenes > 0 via disability  */
        *reduc=0;
/*      if (nextass>_Amax) nextass=_Amax;   */
        return nextass;
       }

/*  Earnings below earnings test threshhold levels  */
    if (laborincome < earnlev) 
       {
        nextass=assets+ssbenes+AfterTaxIncome-c;
        *reduc=0;
/*      if (nextass>_Amax) nextass=_Amax;   */
        return nextass;
       }

/*  Only part of benefits are reduced  */
    if ( ((laborincome-earnlev)*taxfrac) < ssbenes ) 
       {
        double ssearntest=(laborincome-earnlev)*taxfrac; /*  Lost benefits  */ 
        *reduc=ssearntest/(ssbenes); /*  fraction of all benefits reduced  */
        ssbenes=ssbenes-ssearntest;
        nextass=assets+AfterTaxIncome+ssbenes-c;  /*  ssbenes are post-reduction  */
/*      if (nextass>_Amax) nextass=_Amax;   */
        return nextass;
       }

/*  All benefits are reduced  */
    nextass=assets+AfterTaxIncome-c; 
    *reduc=1;
    ssbenes=0;
/*  if (nextass>_Amax) nextass=_Amax;  */
    return nextass;
   }

/*---------------------------------------------------------------------------*/
/*---------------------------------------------------------------------------*/
/*  COMPUTELABORINCOME  */

double computelaborincome(double wage, double h)
   {
    if (tiedwage==0) return (wage*h);
    if ((tiedwage==1) && (h>0)) return (wage*pow(h,1.415)*pow(2000,-.415));
    return 0;                                   /*  if tiedwage==1 and h==0  */
   }

/*---------------------------------------------------------------------------*/
/*---------------------------------------------------------------------------*/
/*  COMPUTEAFTERTAXINCOME  */

double computeAfterTaxIncome(double laborincome, double assets, double *MTR,
                             int badheal, double penbenefits)
   {
    double badhealp, spousalinc, Y;

    badhealp   = (double) badheal;
    spousalinc = spicns + spiage1*agep + spiage2*pow(agep,2) 
                 + spiage3*pow(agep,3) + spiage4*pow(agep,4) 
                 + spiage5*pow(agep,5) + spibadh*badhealp
                 + spibadhage*badhealp*agep;
    spousalinc = spousalinc*cpi8798*(1-nospinc);

    if (spousalinc < 0) spousalinc = 0;

    Y = ror*assets + laborincome + spousalinc + penbenefits;

    if (Y<0) 
       {
        *MTR=0;
        return Y;
       }

    if (Y<tbk1) 
       {
        *MTR=mtr1;
        return ((1-mtr1)*Y);
       }

    if (Y<tbk2) 
       {
        *MTR=mtr2;
        return (ati1+(1-mtr2)*(Y-tbk1));
       }

    if (Y<tbk3)
       {
        *MTR=mtr3;
        return (ati2+(1-mtr3)*(Y-tbk2));
       }

    if (Y<tbk4) 
       {
        *MTR=mtr4;
        return (ati3+(1-mtr4)*(Y-tbk3));
       }

    if (Y<tbk5) 
       {
        *MTR=mtr5;
        return (ati4+(1-mtr5)*(Y-tbk4));
       }

    if (Y<tbk6) 
       {
        *MTR=mtr6;
        return (ati5+(1-mtr6)*(Y-tbk5));
       }

    *MTR=mtr7;
    return ati6+(1-mtr7)*(Y-tbk6);
   }

/*---------------------------------------------------------------------------*/
/*---------------------------------------------------------------------------*/
/*  LOSSY:  Generate a loss function for assets outside a certain range  */

double lossy(double valfunction, double assets, int noliquidity)
   {
    double loss=0;
    if (noliquidity == 1) 
       {
        loss = pow((assets/negpen),2);
        if (loss < 1) loss = 1;
       }
/*  if (assets>_Amax) loss=pow((assets/miserpen),2);  */

    if (valfunction<0) valfunction=valfunction*(1+loss);
    else valfunction=valfunction*(1-loss);

    return valfunction;
   }

/*---------------------------------------------------------------------------*/
/*---------------------------------------------------------------------------*/
/*  INTERP:  Linearly interpolates between points on the AIME/asset/hcar grid  */

double interp(double Avar, double AIMEvar, double *AstatePtr, 
              double *AIMstatePtr, int prtype, int tentype, int hitype, 
              int badheal, int hctype, int wagetype, double ********VvarPtr)

   {
    double Amove, Alev, AIMEmove, AIMElev, Achange, AIMEchange, 
           Afrac, AIMEfrac, LoAssetLoAIME,  LoAssetHiAIME, 
           HiAssetLoAIME, HiAssetHiAIME, retr;
    int AlevShort, AIMEShort;

/*  First find the appropriate interpolation factors by seeing where the  */
/*  grid point lies in relation to the grid breaks  */
/*  With the interpolation factors, we can also find the grid point indices  */ 
/*  corresponding to aVar, AIMEvar and hcvar  */

    if (Avar<_Aint)
       {
        Amove= Amovein1;
        Alev = ((Avar-AstatePtr[0])*Amove);
       }
    else
       {
        Amove=Amovein2;
        Alev=((Avar-AstatePtr[Anint-1])*Amove)+Anint-1;
       }

    if (AIMEvar<_AIMEint)
       {
        AIMEmove=AIMEmovein1;
        AIMElev=((AIMEvar-AIMstatePtr[0])*AIMEmove);
       }
    else
       {
        AIMEmove=AIMEmovein2;
        AIMElev=((AIMEvar-AIMstatePtr[AIMEnint-1])*AIMEmove)+AIMEnint-1;
       }

/*  Find the grid points bounding avar, wvar and AIMEvar from below  */
/*  Recall C++ indices run from 0 to *num-1  */

    AlevShort = (int) floor(Alev);
    AIMEShort = (int) floor(AIMElev);

    if (AlevShort>_Anum-2) AlevShort=_Anum-2; 
    if (AIMEShort>AIMEnum-2) AIMEShort=AIMEnum-2;

    if (AlevShort<0) AlevShort=0;
    if (AIMEShort<0) AIMEShort=0;

/*  Find the weights to be placed along each dimension  */

    Achange    = Avar-AstatePtr[AlevShort];
    AIMEchange = AIMEvar-AIMstatePtr[AIMEShort];
    Afrac      = Achange*Amove;
    AIMEfrac   = AIMEchange*AIMEmove;

/*  Evaluate at the grid points bounding avar, AIMEvar and hcvar  */

    LoAssetLoAIME = VvarPtr[prtype][tentype][hitype][badheal][hctype][AIMEShort][wagetype][AlevShort];
    LoAssetHiAIME = VvarPtr[prtype][tentype][hitype][badheal][hctype][AIMEShort+1][wagetype][AlevShort];
    HiAssetLoAIME = VvarPtr[prtype][tentype][hitype][badheal][hctype][AIMEShort][wagetype][AlevShort+1];
    HiAssetHiAIME = VvarPtr[prtype][tentype][hitype][badheal][hctype][AIMEShort+1][wagetype][AlevShort+1];

/*  Weighted average  */
    retr = Afrac*AIMEfrac*         HiAssetHiAIME +
           Afrac*(1-AIMEfrac)*     HiAssetLoAIME +
           (1-Afrac)*AIMEfrac*     LoAssetHiAIME +
           (1-Afrac)*(1-AIMEfrac)* LoAssetLoAIME ;

    return retr;
   }

/*---------------------------------------------------------------------------*/
/*---------------------------------------------------------------------------*/
/*  COMPUTEAIME  */

double computeAIME(double AIME, double laborincome)
   {
    double nextAIME, LowYearIncome;

    if (laborincome>AIMEmax) laborincome=AIMEmax;
    if (age<(AIMEage+1)) AIME=AIME*(1+aggwagegrowth); /* Index year increases until age 60 */
    if (AIME>AIMEmax) AIME=AIMEmax;

    LowYearIncome = AIMElowfrac*AIME;

    if (laborincome>LowYearIncome)
       {
        nextAIME= AIME + ((laborincome-LowYearIncome)*AIMEaccrue);
        if (nextAIME>AIMEmax) nextAIME=AIMEmax;
        return nextAIME; 
       }

    nextAIME=AIME;
    return nextAIME;
   }

/*---------------------------------------------------------------------------*/
/*---------------------------------------------------------------------------*/
/*         below are the bend points for AIME and PIA, respectively          */
/*             AIME                    PIA                                   */
/*             5724                    AIME*.9 = 5151.6                      */
/*             34500                   5151.6 +.32*(AIME-5724) = 14359.9     */
/*             34500+                  14359.9 +.15*(AIME-22392)             */                        
/*         Pension benefits are imputed by:                                  */
/*             pb = p0 + p1*PIA + p2*max{0,PIA-6751.6}                       */
/*---------------------------------------------------------------------------*/
/*---------------------------------------------------------------------------*/
/*  COMPUTEPIA: convert AIME to PIA  */

double computePIA(double AIME)
   {
    double PIA;

    if (AIME<AIMEbk1) 
       {
        PIA=mpr1*AIME;
        return PIA;
       }

    if (AIME<AIMEbk2)
       {
        PIA=PIA1+mpr2*(AIME-AIMEbk1);
        return PIA;
       }

    PIA=PIA2+mpr3*(AIME-AIMEbk2);
    return PIA;
   }

/*---------------------------------------------------------------------------*/
/*---------------------------------------------------------------------------*/
/*  FINDAIME:  find the AIME associated with a given PIA  */

double findAIME(double PIA, int imposemax)
   {
    double AIMEstar;
    if (PIA<0)
       {
        return 0.0;
       }
    if (PIA<PIA1)
       {
        return (PIA/mpr1);
       }
    if (PIA<PIA2)
       {
        return (AIMEbk1+(PIA-PIA1)/mpr2);
       }

    AIMEstar=AIMEbk2+(PIA-PIA2)/mpr3;
    if ( (AIMEstar>AIMEmax) && (imposemax==1) ) AIMEstar = AIMEmax;
    return AIMEstar;
   }

/*---------------------------------------------------------------------------*/
/*---------------------------------------------------------------------------*/
/*  PREDICTPENBEN: impute pension benefits using PIA                    */
/*                 these are "full" benefits for someone with complete  */
/*                 access to their pension wealth                       */

double predictpenben(double PIA, double *pencfsPtr, double PIAkink[2], int HItype)
   {
    double penben;

    if (PIA < PIAkink[0])
       {
        penben = pencfsPtr[HItype] + pencfsPtr[HITnum]*PIA;
        if (penben < 0.0) penben = 0;
        return penben;
       }

    if (PIA < PIAkink[1])
       {
        penben = pencfsPtr[HItype] + pencfsPtr[HITnum]*PIA + 
                 pencfsPtr[HITnum+1]*(PIA-PIAkink[0]);
        if (penben < 0.0) penben = 0;
        return penben;
       }

    penben = pencfsPtr[HItype] + pencfsPtr[HITnum]*PIA + 
             pencfsPtr[HITnum+1]*(PIA-PIAkink[0])+
             pencfsPtr[HITnum+2]*(PIA-PIAkink[1]);
    return penben;

   }

/*---------------------------------------------------------------------------*/
/*---------------------------------------------------------------------------*/
/*  FINDPIA:  find the PIA associated with a given total of pension and  */
/*            SS benefits  */

double findPIA(double totben, double *pencfsPtr, double PIAkink[2], 
               double totbenkink[2], int HItype)
   {
    double PIA;

    if (totben<pencfsPtr[HItype])
       {
        return 0.0;
       }
    if (totben<totbenkink[0])
       {
        PIA = (totben-pencfsPtr[HItype])/(1+pencfsPtr[HITnum]);
       }
    else
       {
        if (totben<totbenkink[1])
           {
            PIA = PIAkink[0] + (totben-totbenkink[0])/
                  (1+pencfsPtr[HITnum]+pencfsPtr[HITnum+1]);
           }
        else
           {
            PIA = PIAkink[1] + (totben-totbenkink[1])/
                  (1+pencfsPtr[HITnum]+pencfsPtr[HITnum+1]+pencfsPtr[HITnum+2]);
           }
       }

    return PIA;
   }

/*---------------------------------------------------------------------------*/
/*---------------------------------------------------------------------------*/
/*  COMPUTEPENACCRUE  */

double computepenaccrue(double laborincome, double *penaccPtr)
   {
    double penaccruerate, penaccrue, penaccrueprob, lli;

    penaccruerate = 0.0;
    if (laborincome>0)
       {
        lli = log(laborincome);
        penaccruerate = penaccPtr[0] + penaccPtr[1]*lli;
        if (penaccruerate <-0.1) penaccruerate = -0.1; 
        penaccrueprob = penaccPtr[2]+penaccPtr[3]*lli + penaccPtr[4]*lli*lli;
        penaccrueprob = exp(penaccrueprob)/(1+exp(penaccrueprob));
        penaccruerate = penaccruerate*penaccrueprob;
       }
    penaccrue = penaccruerate*laborincome;

    return penaccrue;
   }

/*---------------------------------------------------------------------------*/
/*---------------------------------------------------------------------------*/

/*  GETADJ:  find adjustment factors for early/late retirement and the  */
/*           crediting back of the earnings test.                       */
/*           NOTE:  Works around a benchmark nret age of 65             */


int getadj(double *sscredsPtr)
   {
    double cumadj2;
    double bigcred62, bigcred63, bigcred64, bigcred65, bigcred66,
           eretadj62, eretadj63, eretadj64, eretadj65, eretadj66,
           litcred65, litcred66, litcred67, litcred68, litcred69;

    bigcred62 = sscredsPtr[0];  /*  Parameters for crediting back SS benefits lost to earnings test  */
    bigcred63 = sscredsPtr[1];
    bigcred64 = sscredsPtr[2];
    bigcred65 = sscredsPtr[3];
    bigcred66 = sscredsPtr[4];
    eretadj62 = sscredsPtr[5];  /*  Early application penalty factors  */
    eretadj63 = sscredsPtr[6];
    eretadj64 = sscredsPtr[7];
    eretadj65 = sscredsPtr[8];
    eretadj66 = sscredsPtr[9],
    litcred65 = sscredsPtr[10];  /*  Parameters for crediting back SS benefits lost to earnings test  */
    litcred66 = sscredsPtr[11];
    litcred67 = sscredsPtr[12];
    litcred68 = sscredsPtr[13];
    litcred69 = sscredsPtr[14];

    eretadj=1;  /*  These are all globals  */
    cumeretadj=1;
    bigcred=0;
    litcred=0;
    cumadj2 = 1;

    if (age < nret)
       {
        if (age > 66)  /* nret > 67  */
           {
            eretadj=eretadj66;
            cumeretadj=pow(eretadj66,(nret-age));
            bigcred=bigcred66;            
           }

        if (age==66)
           {
            if (nret > 67) cumadj2 = pow(eretadj66,(nret-67));
            eretadj=eretadj66;
            cumeretadj=eretadj66*cumadj2;
            bigcred=bigcred66;            
           }

        if (age==65)
           {
            if (nret==67) cumadj2 = eretadj66;
            if (nret > 67) cumadj2 = eretadj66*pow(eretadj66,(nret-67));
            eretadj=eretadj65;
            cumeretadj=eretadj65*cumadj2;
            bigcred=bigcred65;            
           }

        if (age < 65)  /*  Cumulative adjustments benchmarked to nret==65  */
           {           /*  This allows for nret > 65  */
            if (nret==66)
               {
                cumadj2 = eretadj65;
               }
            if (nret==67)
               {
                cumadj2 = eretadj65*eretadj66;
               }
            if (nret > 67)
               {
                cumadj2 = eretadj65*eretadj66*pow(eretadj66,(nret-67));
               }
           }

        if (age==64)
           {
            eretadj=eretadj64;
            cumeretadj=eretadj64*cumadj2;
            bigcred=bigcred64;
           }
        if (age==63)
           {
            eretadj=eretadj63;
            cumeretadj=eretadj64*eretadj63*cumadj2;
            bigcred=bigcred63;
           }
        if (age==62)
           {
            eretadj=eretadj62;
            cumeretadj=eretadj64*eretadj63*eretadj62*cumadj2;
            bigcred=bigcred62;
           }
        if (age<62)  /*  Allow for policy experiments  */
           {
            eretadj=eretadj62;
            cumeretadj=eretadj64*eretadj63*eretadj62;
            cumeretadj=cumeretadj*pow(eretadj,(62-age))*cumadj2;
            bigcred=bigcred62;
           }
       }

    if ((age>=nret) && (age<etstage))
       {
        if (age<65)  litcred=litcred65;  /*  Allow for policy experiments  */
        if (age==65) litcred=litcred65;
        if (age==66) litcred=litcred66;
        if (age==67) litcred=litcred67;
        if (age==68) litcred=litcred68;
        if (age==69) litcred=litcred69;
        if (age>69)  litcred=litcred69;  /*  Allow for policy experiments  */
       }

    return 1;
   }

/*---------------------------------------------------------------------------*/
/*---------------------------------------------------------------------------*/
/*  COMPUTEPENSTUFF  */

double computepenstuff(double AIME, double *pencfsPtr, double *PIA, 
                       double *ssbenes, double *penbal, double PIAkink[2],
                       int HItype)
   {
    double penbenefits, actualpenben;

/*  Compute current income and future assets and AIME  */

    *PIA=computePIA(AIME);
    *ssbenes=apply*(*PIA);

/*  Impute pension benefits and wealth  */
    penbenefits = predictpenben(*PIA,pencfsPtr,PIAkink,HItype);
    actualpenben = penbenefits*penbenprob;
    *penbal = penbenefits*penbalfrac;

    return actualpenben;
   }

/*---------------------------------------------------------------------------*/
/*---------------------------------------------------------------------------*/
/*  COMPUTETHEVALUEFUNCTION  */

double computethevaluefunction(
    double c, double h, double ********valftpoPtr, double particip, double assets, 
    double wage, double AIME, double *AstatePtr, double *AIMstatePtr, 
    double *hcstatePtr, double *wtransprobPtr, double *HITtransprobPtr, 
    double *hctransprobPtr, double oldval, int *start, double mnlnhc, 
    double stdlnhc, double *hcnodesPtr, double *hcwgtsPtr, double *transfers, 
    double *pencfsPtr, double *penaccPtr, double *penkinkPtr, int *noliquidity, 
    double PIA, double ssbenes, double penbal, double penbenefits, int badheal, 
    double survprob, double probbad) 
   {
    double Evtpo, utils, bequestutils, nextperiodassets, valuefunction, 
           nextAIME, nextPIA, reduc, nppenbal, healthcosts, cashonhand, 
           MTR, totben, imputedpenbal, imputedpenben, penaccrue, 
           assetadj, prehcassets, nphcar1, hcinnovwn, probnpHS, iV, 
           nextPIAunadj, nextPIApartadj,
           laborincome, AfterTaxIncome;
	double pencfsPtr2[6];
    double PIAkink[2], totbenkink[2]; 
    int iCF, iHCp, iHCt, iHIT, iHS, iWage, makeAIMEadj, imposemax, npPrefType,
        npHasTenure;

    *transfers = 0.0;
    imposemax = 1;

/*  Compute current income and future assets and AIME  */

    laborincome = computelaborincome(wage,h);
    AfterTaxIncome = computeAfterTaxIncome(laborincome, assets, &MTR, badheal, 
                                           penbenefits); 
/*  AIME is computed using best years, starting age 25  */
    nextAIME = computeAIME(AIME, laborincome);

/*  Use nextPIA to adjust nextAIME to reflect early retirement penalty or  */
/*  crediting-back of benefits lost via earnings test.  */
    nextPIA = computePIA(nextAIME);
    nextPIAunadj = nextPIA;

/*  Compute next period's pension balances and compare to balances that would  */
/*  be imputed with next period's AIME.  The discrepancy will be added to  */
/*  next period's asset balance.  */

    penaccrue = computepenaccrue(laborincome,penaccPtr);
    nppenbal  = (1+ror)*penbal + penaccrue - penbenefits;
    nppenbal  = nppenbal/survprobpen;

    makeAIMEadj = 0;

/*  Adjust current benefits and next period's AIME to reflect early retirement  */
/*  These adjustments are done for each year of early retirement, and compound. */
    if ((apply==1) && (age<nret))
       {
        nextPIA = eretadj*nextPIA;
        ssbenes = ssbenes*cumeretadj;
        makeAIMEadj = 1;
       }

/*  There is also a benefit to deferring application for nret<=age<mappage  */
    if ((apply==0) && (age<mappage) && (age>=nret))
      {
       nextPIA = (1+litcred)*nextPIA;
       makeAIMEadj = 1;
      }

/*  Compute next period's asset balance.  While current assets are net of  */
/*  health costs, at this point next period's Assets are NOT.  */
    nextperiodassets=ass(AfterTaxIncome, c, laborincome, assets, &reduc, ssbenes); 

/*  Use reduc to increase AIME of people that lost benefits by earnings test  */
    if ((apply==1) && (age<etstage))
      {
       if (age<nret) nextPIA = (1+bigcred*reduc)*nextPIA;
       else nextPIA = (1+litcred*reduc)*nextPIA;
       makeAIMEadj = 1;
      }

/*  Check consumption floor.  transfers > 0 will alert search algorithm  */
/*  Assume that agent consumes any transfers  */
    *noliquidity = 0;
    cashonhand = nextperiodassets + c;
    if (cashonhand < cfloor)
       {
        *transfers = cfloor - cashonhand;
        c = cfloor;
        nextperiodassets = 0.0;
        *noliquidity = 0;
       }
     
/*  Abort procedure if next period's PRE-health-cost assets are negative  */
    else
       {
        if (nextperiodassets < _Afloor) *noliquidity = 1;
        if ((*noliquidity == 1) && (*start == 0))
           {
            valuefunction = oldval-negpen;
            return valuefunction;
           }
        }

/*  Find current utility  */
    utils=U(c,h,particip,badheal);

/*  Find expectation of next period's value function  */

    prehcassets = nextperiodassets;
    nextPIApartadj = nextPIA;
    Evtpo = 0;
    for(iCF=0; iCF<6; iCF++) pencfsPtr2[iCF] = (1-MTR)*pencfsPtr[iCF];

    npPrefType  = PrefType; /*  Preferences are fixed effects  */
    npHasTenure = 0;        /*  Default to no-reentry-cost case */
    if ((HasEntryCost==1) && (h>=keepjob))
       {
        npHasTenure = 1; 
       }

/*  First integrate across future health insurance statuses  */

    for(iHIT=0; iHIT<HITnum; iHIT++)
       {
        if (HITtransprobPtr[iHIT] == 0) continue;

        PIAkink[0] = penkinkPtr[iHIT];  /*  Kinks in pension benefit imputation equation  */
        PIAkink[1] = penkinkPtr[iHIT+HITnum];
        totbenkink[0] = pencfsPtr2[iHIT]+PIAkink[0]*(1+pencfsPtr2[HITnum]);
        totbenkink[1] = totbenkink[0] +
                        (PIAkink[1]-PIAkink[0])*(1+pencfsPtr2[HITnum]+pencfsPtr2[HITnum+1]);
        assetadj = 0;
        imputedpenben = predictpenben(nextPIAunadj, pencfsPtr, PIAkink, iHIT);
        if (age < (dieage-1))
           {
            imputedpenbal = imputedpenben*nppenbalfrac;
            assetadj = (1-MTR)*survprobpen*(nppenbal - imputedpenbal);
           }

    /*  AIME adjustments also change imputed Pension benefits.  Recalculate AIME  */          
    /*  to offset this effect.  */
        if (makeAIMEadj == 1)
           {
            totben = (1-MTR)*imputedpenben + nextPIApartadj;  /*  imputedpenben is pre-AIME adjustments  */
            nextPIA = findPIA(totben,pencfsPtr2,PIAkink,totbenkink,iHIT);
            imposemax = 0;
            nextAIME = findAIME(nextPIA,imposemax); 
           }

    /*  Then integrate over future health statuses  */
        for(iHS=0; iHS<HSnum; iHS++)
           {
            if (iHS==1) probnpHS = probbad;  /*  1=> bad health  */
            else probnpHS = 1-probbad;

        /*  Then integrate across the innovations to the AR1 part of future health costs  */
            for(iHCp=0; iHCp<hcnum; iHCp++)
               {
                if (hctransprobPtr[iHCp] < zeroprobhc) continue; 
                nphcar1 = hcstatePtr[iHCp]*sqrt(fracAR1)*stdlnhc;

            /*  Then integrate across transitory health cost shocks  */

                for(iHCt=0; iHCt<ndnumHC; iHCt++)
                   {
                /*  Now we work with assets net of health costs  */

                    hcinnovwn = hcnodesPtr[iHCt]*sqrt(fracwn)*stdlnhc;
                    healthcosts = exp(mnlnhc+hcinnovwn+nphcar1);

                    nextperiodassets = prehcassets - healthcosts;
                    bequestutils= beq(nextperiodassets);

                /*  Include pension imputation adjustment.  Since these additional pension  */
                /*  balances aren't liquid, they should be added in after checking asset    */
                /*  constraints and calculating bequest utility.  */

                    nextperiodassets = nextperiodassets + assetadj;

                    for(iWage=0; iWage<_Wnum; iWage++)
                       {
                        if (wtransprobPtr[iWage] < zeroprobwages) continue; 
                        iV = interp(nextperiodassets, nextAIME, AstatePtr, 
                                    AIMstatePtr, npPrefType, npHasTenure, 
                                    iHIT, iHS, iHCp, iWage, valftpoPtr); 

                        iV = iV*survprob + (1-survprob)*bequestutils;

                        Evtpo = Evtpo+iV*HITtransprobPtr[iHIT]*probnpHS
                                *hctransprobPtr[iHCp]*hcwgtsPtr[iHCt]
                                *wtransprobPtr[iWage];

                       }  /*  End loop through wages  */
                   }  /*  End loop through transitory health cost shocks  */
               }  /*  End loop through persistent health cost shocks  */
           }  /*  End loop through health statuses  */
       }  /*  End loop through health insurance statuses  */

/*  Sum up, then call lossy to make sure that assets are in specified range  */

    valuefunction = utils + _beta*Evtpo;
    valuefunction = lossy(valuefunction, prehcassets, *noliquidity);
    *start = 0;

    return valuefunction;
   }

/*---------------------------------------------------------------------------*/
/*---------------------------------------------------------------------------*/
/*  CHECKMAX:  Checks to see if maximimizing consumption lies at boundary of */
/*             search space and adjusts space if necessary  */
/*             NOTE:  when we get here, i == hihic  */

int checkmax(int clevi, int cnum, int csteps, int *loloc, int *loc, int *hic, 
             int *hihic, int *i, int illiquidsolution)
   {
    int start = 0;

/*  Reorient search if liquidity constraint is violated.  */
/*  Check loloc-csteps-3 < i < loloc  */
    if (illiquidsolution == 1)
       {
        if (*loloc > 0)
           {
            *hic = *loc+1;
            *hihic = *hic;
            *loc = *loloc-csteps;
            if (*loc < 1) *loc = 1;
            *loloc = *loc - xvstep;
            if (*loloc < 0) *loloc = 0;
            start = 1;
            *i = *loloc - 1;  /*  i gets incremented +1 when returning to top of search loop  */
            }
         return start;
        }

/*  Reorient search if clev >= hic:  Check hihic < i < hihic+csteps+3  */
    if (clevi > (*hic-1))
       {
        if (*hihic < cnum-1)
           { 
            *hic = *hihic+csteps;
            if (*hic > (cnum-2)) *hic = cnum-2;
            *hihic = *hic+xvstep;
            if (*hihic > (cnum-1)) *hihic = cnum-1;
           }
         return start;
        }

    if (clevi > *loc) return start;

/*  Reorient search if clev <= loc:  Check loloc-csteps-3 < i < loloc  */
    if (clevi < (*loc+1))
       {
        if (*loloc > 0)
           {
            *hic = *loc+1;
            *hihic = *hic;
            *loc = *loloc-csteps;
            if (*loc < 1) *loc = 1;
            *loloc = *loc - xvstep;
            if (*loloc < 0) *loloc = 0;
            start = 0;
            *i = *loloc - 1;  /*  i gets incremented +1 when returning to top of search loop  */
            }
         return start;
        }

    return start;
   }


/*---------------------------------------------------------------------------*/
/*---------------------------------------------------------------------------*/
/*  CHECKMAX2:  Checks to see if maximimizing consumption lies at boundary   */
/*              of search space and adjusts space if necessary  */
/*              NOTE:  when we get here, i == maxstepc-1  */

double checkmax2(double *cmin, double *cmax, double *crec, int *start, 
                 int *maxstepc, double cincr, double oldc2, int srch2steps_c, 
                 int illiquidsolution)
   {   
    double c2;
    c2 = oldc2;

    if (illiquidsolution==1) /*  If c_0 is not feasible, c_1 > c_0 is not feasible either  */
       {
        *crec     = *cmin;
        *start    = 1;
        *maxstepc = *maxstepc + srch2steps_c;
        *cmin     = (*cmin)*pow(1/cincr,srch2steps_c);
        *cmax     = (*crec)*cincr;
        c2        = *cmin;
        return c2;
       }

    if (*crec > *cmax)
       {
        *maxstepc = *maxstepc + srch2steps_c;
        *cmin     = (*cmax)/cincr;
        *cmax     = (*cmax)*pow(cincr,srch2steps_c);
        return c2;
       }

    if (*crec > *cmin) return c2;

    if (*cmin > cfloor)  /*  *crec == cmin  */
       {
        *maxstepc = *maxstepc+srch2steps_c; 
        *cmax     = (*cmin)*cincr;
        *cmin     = (*cmin)*pow(1/cincr,srch2steps_c);
	    if (*cmin < cfloor) *cmin = cfloor;
        c2        = *cmin;
       }

    return c2;
   }                


/*---------------------------------------------------------------------------*/
/*---------------------------------------------------------------------------*/
/*  GETPREFPARMS:  Find Type-specific Preference Parameters                  */

int getprefparms(double *PrefParmPtr,int ThisPrefType)
   {
    _beta        = PrefParmPtr[ThisPrefType];
    _gamma       = PrefParmPtr[ThisPrefType+PrefTypeNum];
    _nu          = PrefParmPtr[ThisPrefType+2*PrefTypeNum];
    bequest      = PrefParmPtr[ThisPrefType+3*PrefTypeNum];
    leispref     = PrefParmPtr[ThisPrefType+4*PrefTypeNum];
    leisprefbad  = PrefParmPtr[ThisPrefType+5*PrefTypeNum];
    fixcost      = PrefParmPtr[ThisPrefType+6*PrefTypeNum];
    EntryPenalty = PrefParmPtr[ThisPrefType+7*PrefTypeNum];
    Onemgamma    = PrefParmPtr[ThisPrefType+8*PrefTypeNum];
    Onemnu       = PrefParmPtr[ThisPrefType+9*PrefTypeNum];
    conspref     = PrefParmPtr[ThisPrefType+10*PrefTypeNum];
    beqk         = PrefParmPtr[ThisPrefType+11*PrefTypeNum];
    keepjob      = PrefParmPtr[ThisPrefType+12*PrefTypeNum];

    return 1;
   }                
	
/*---------------------------------------------------------------------------*/
/*---------------------------------------------------------------------------*/
/*  GLOBDEREF:  Dereference a bunch of pointers to C++ globals  */

int globderef(double *interpparmPtr, double *gridptnumPtr, 
              double *ageparmPtr, double *miscparmPtr, 
              double *applyPtr, double *survprobPtr, double *agePtr, 
              double *AIMEbkPtr, double *mprPtr, double *PIAbkPtr)
   {
/*  Dereference variables used in quadrature and interpolation  */

    Amovein1     = interpparmPtr[0];
    Amovein2     = interpparmPtr[1];
    AIMEmovein1  = interpparmPtr[2];
    AIMEmovein2  = interpparmPtr[3];
    Anint        = (int) floor(rounder+interpparmPtr[4]);
    AIMEnint     = (int) floor(rounder+interpparmPtr[5]);
    _Aint        = interpparmPtr[6];
    _AIMEint     = interpparmPtr[7];

    _Anum        = (int) floor(rounder+gridptnumPtr[0]);
    _Wnum        = (int) floor(rounder+gridptnumPtr[1]);
    AIMEnum      = (int) floor(rounder+gridptnumPtr[2]);
    HITnum       = (int) floor(rounder+gridptnumPtr[3]);
    hcnum        = (int) floor(rounder+gridptnumPtr[4]);
    ndnumHC      = (int) floor(rounder+gridptnumPtr[5]);
    PrefTypeNum  = (int) floor(rounder+gridptnumPtr[6]);
    HasEntryCost = (int) floor(rounder+gridptnumPtr[7]);

/*  Dereference parameters used in the budget constraint  */

    nret         = (int) floor(rounder+ageparmPtr[0]);
    etstage      = (int) floor(rounder+ageparmPtr[1]);
    penbensstart = (int) floor(rounder+ageparmPtr[2]);
    manretage    = (int) floor(rounder+ageparmPtr[3]);
    mappage      = (int) floor(rounder+ageparmPtr[4]);
    eretage      = (int) floor(rounder+ageparmPtr[5]);
    medicage     = (int) floor(rounder+ageparmPtr[6]);

    earnlev      = miscparmPtr[0];
    taxfrac      = miscparmPtr[1];
    _Amax        = miscparmPtr[2];
    ror          = miscparmPtr[3];
    cfloor       = miscparmPtr[4];
    _Afloor      = miscparmPtr[5];
    tiedwage     = (int) floor(rounder+miscparmPtr[6]); 
    AIMEaccrue   = miscparmPtr[7];
    penbalfrac   = miscparmPtr[8];
    nppenbalfrac = miscparmPtr[9];
    penbenfrac   = miscparmPtr[10];
    fracAR1      = miscparmPtr[11];
    fracwn       = miscparmPtr[12];
    nospinc      = miscparmPtr[13];
    hminsim      = miscparmPtr[14];
    rhohc        = miscparmPtr[15];
    penbenprob   = miscparmPtr[16];
    aggwagegrowth = miscparmPtr[17];
    AIMElowfrac  = miscparmPtr[18];

    apply        = (int) floor(rounder+*applyPtr);
    age          = (int) floor(rounder+*agePtr);   /*  This is an integer  */
    agep         = floor(rounder+*agePtr);         /*  This is a double    */

/*  Dereference parameters used in computing the value function  */
 
    survprobpen = survprobPtr[2];

/*  Dereference vectors of SS- and pension-related parameters  */
    AIMEbk1 = AIMEbkPtr[0];
    AIMEbk2 = AIMEbkPtr[1];
    AIMEmax = AIMEbkPtr[2];
    mpr1    = mprPtr[0];
    mpr2    = mprPtr[1];
    mpr3    = mprPtr[2];
    PIA1    = PIAbkPtr[0];
    PIA2    = PIAbkPtr[1];

    return 1;
   }                
/*---------------------------------------------------------------------------*/
/*---------------------------------------------------------------------------*/
/*  LAZYRULE:  Finds the optimal consumption choice for non-workers  */
/*             Search over a fixed consumption grid  */

double lazyrule(
    double assets, double wage, double AIME, double *wtransprobPtr, 
    double *HITtransprobPtr, double *hctransprobPtr, double survprob, 
    double probbad, int csteps, int cnum, double *cdecPtr, double *clevPtr, 
    double mnlnhc, double stdlnhc, double *hcnodesPtr, double *hcwgtsPtr, 
    double *pencfsPtr, double *penaccPtr, double *penkinkPtr, 
    double *AstatePtr, double *AIMstatePtr, double *hcstatePtr, double PIA, 
    double ssbenes, double penbal, double penbenefits, double *crec, 
    int badheal, double ********valftpoPtr)

   {
	double val, valc, transfers, c2;
	int start, noliquidity, illiquidsolution, oldclev, loc, loloc, hic,
	    hihic, clazlevi, i;

    val   = 0.0;  /*  U(c,h) + BxE_t{V[t+1]}:  returned to calling function  */
    valc  = 0.0;
    start = 1;
    noliquidity = 0;
    illiquidsolution = 0;
    transfers = 0.0;

    oldclev  = (int) floor(rounder+*clevPtr);
    loc      = oldclev - csteps;
    if (loc < 1) loc = 1;
    loloc    = loc - xvstep;
    if (loloc < 0) loloc = 0;
    hic      = loc + 2*csteps;
    if (hic >(cnum-2)) hic = (cnum-2); 
    hihic    = hic + xvstep;
    if (hihic > cnum-1) hihic = cnum-1;
    clazlevi = loloc;  /*  clazlev is a double  */
    c2       = cdecPtr[clazlevi];
    *crec    = c2;

/*  Focus on consumption indices in neighborhood of old clazlev  */
/*  Reorient search if any point +-3 indices from the neighborhood's  */
/*  edges outperforms points in interior of neighborhood  */

/*  val can be a large negative num, safest to initialize this way  */

    val = computethevaluefunction(
            *crec, nopart, valftpoPtr, nopart, assets, wage, AIME,
            AstatePtr, AIMstatePtr, hcstatePtr, wtransprobPtr, 
            HITtransprobPtr, hctransprobPtr, val, &start, 
            mnlnhc, stdlnhc, hcnodesPtr, hcwgtsPtr, &transfers, 
            pencfsPtr, penaccPtr, penkinkPtr, &illiquidsolution, 
            PIA, ssbenes, penbal, penbenefits, badheal, 
            survprob, probbad);

/*  start = 0;  */

/*  If transfers > 0, cash on hand cannot support the consumption floor,  */
/*  and there is no need to check any other consumption level (we assume  */
/*  that the floor is so low that the agent never saves)  */

    if (transfers > rounder)
       {
        *crec    = cfloor;
        *clevPtr = cfloorindx;
        return val;
       }

    for (i = loloc; i < (hihic+1); i++)
       {
        c2 = cdecPtr[i];

        valc = computethevaluefunction(
                c2, nopart, valftpoPtr, nopart, assets, wage, AIME, 
                AstatePtr, AIMstatePtr, hcstatePtr, wtransprobPtr, 
                HITtransprobPtr, hctransprobPtr, val, &start, 
                mnlnhc, stdlnhc, hcnodesPtr, hcwgtsPtr, &transfers, 
                pencfsPtr, penaccPtr, penkinkPtr, &noliquidity, PIA, 
                ssbenes, penbal, penbenefits, badheal, survprob, 
                probbad);

        if ( (valc > val) || (noliquidity < illiquidsolution) )
           {
            clazlevi = i;
            *crec    = c2;
            val      = valc;
            illiquidsolution = noliquidity;
           }

        if (i == hihic)
           {
        /*  Check to see if utility-maximimizing consumption lies near boundary of  */
        /*  search space and adjust space if necessary   */
            start = checkmax(clazlevi, cnum, csteps, &loloc, &loc, &hic, &hihic,
                             &i, illiquidsolution);
           }
       }

    *clevPtr = clazlevi;
    return val;
   }

/*---------------------------------------------------------------------------*/
/*---------------------------------------------------------------------------*/
/*  LAZYRULE2:  Finds the optimal consumption choice for non-workers  */
/*              Search centered around optimal decision from earlier search  */

double lazyrule2(
    double assets, double wage, double AIME, double *wtransprobPtr, 
    double *HITtransprobPtr, double *hctransprobPtr, double survprob, 
    double probbad, double cincr, int srch2steps_c, double mnlnhc, 
    double stdlnhc, double *hcnodesPtr, double *hcwgtsPtr, 
    double *pencfsPtr, double *penaccPtr, double *penkinkPtr, 
    double *AstatePtr, double *AIMstatePtr, double *hcstatePtr, 
    double PIA, double ssbenes, double penbal, double penbenefits, 
    double *crec, int badheal, double ********valftpoPtr)

   {
    double oldc, val, valc, transfers, cmin, cmax, c2;
    int start, noliquidity, illiquidsolution, i, maxstepc;
    oldc  = *crec;
    val   = 0.0;   /*  U(c,h) + BxE_t{V[t+1]}:  returned to calling function  */
    valc  = 0.0;
    start = 1;
    noliquidity = 0;
    illiquidsolution = 0;
    transfers = 0.0;

/*  Focus on consumption indices in neighborhood of oldc  */
/*  Reorient search if endpoints outperform neighborhood  */

/*  val can be a large negative num, safest to initialize this way  */

    val = computethevaluefunction(
            *crec, nopart, valftpoPtr, nopart, assets, wage, AIME, 
            AstatePtr, AIMstatePtr, hcstatePtr, wtransprobPtr, 
            HITtransprobPtr, hctransprobPtr, val, &start, 
            mnlnhc, stdlnhc, hcnodesPtr, hcwgtsPtr, &transfers, 
            pencfsPtr, penaccPtr, penkinkPtr, &illiquidsolution, 
            PIA, ssbenes, penbal, penbenefits, badheal, survprob, 
            probbad);

/*  If transfers > 0, cash on hand cannot support the consumption floor,  */
/*  and there is no need to check any other consumption level (we assume  */
/*  that the floor is so low that the agent never saves)  */

    if (transfers > rounder)
       {
        *crec = cfloor;
        return val;
       }

/*  start = 0;  */
    cmin     = oldc*pow(1/cincr,srch2steps_c);
    cmax     = oldc*pow(cincr,srch2steps_c-1);
    c2       = cmin;
    maxstepc = 2*srch2steps_c+2;

    for (i = 1; i < maxstepc; i++)
       {
        valc = computethevaluefunction(
                c2, nopart, valftpoPtr, nopart, assets, wage, AIME, 
                AstatePtr, AIMstatePtr, hcstatePtr, wtransprobPtr, 
                HITtransprobPtr, hctransprobPtr, val, &start, 
                mnlnhc, stdlnhc, hcnodesPtr, hcwgtsPtr, &transfers, 
                pencfsPtr, penaccPtr, penkinkPtr, &noliquidity, PIA, 
                ssbenes, penbal, penbenefits, badheal, survprob, 
                probbad);

        if ( (valc > val) || (noliquidity < illiquidsolution) )
           {
            *crec=c2;
            val=valc;
            illiquidsolution = noliquidity;
           }
       c2 = c2*cincr;

       if (i < maxstepc-1) continue;  /*  Avoid the mess below when possible  */

   /*  Check to see if utility-maximimizing consumption lies at boundary of  */
   /*  search space and adjust space if necessary  */       
       c2 = checkmax2(&cmin, &cmax, crec, &start, &maxstepc, cincr, c2, 
                      srch2steps_c, illiquidsolution);
       }  /*  End loop through consumption  */

    return val;
   }

/*---------------------------------------------------------------------------*/
/*---------------------------------------------------------------------------*/
/*  WORKRULE:  Finds the optimal consumption/hours choice for workers  */
/*             Search over a fixed consumption grid  */

double workrule(
    double assets, double wage, double AIME, double *wtransprobPtr, 
    double *HITtransprobPtr, double *hctransprobPtr, double survprob, 
    double probbad, int csteps, int cnum, double *cdecPtr, double *clevPtr, 
    int hgridn, int hcent, double hstep, double mnlnhc, double stdlnhc,
    double *hcnodesPtr, double *hcwgtsPtr, double *pencfsPtr, 
    double *penaccPtr, double *penkinkPtr, double *AstatePtr, 
    double *AIMstatePtr, double *hcstatePtr, double PIA, double ssbenes, 
    double penbal, double penbenefits, double *crec, double *hrec, 
    int badheal, double ********valftpoPtr)

   {
    double val, valc, transfers, hmin2, hmax2, hmin3, hmin4, c2, hc, hcstar,
           ReentryCost;
    int start, noliquidity, illiquidsolution, oldclev, loc, loloc, hic, 
        hihic, cmaxlevi, i, j;

    val   = 0.0;  /*  U(c,h) + BxE_t{V[t+1]}:  returned to calling function  */
    valc  = 0.0;
    start = 1;
    noliquidity = 0;
    illiquidsolution = 0;
    transfers = 0.0;

    hmin2 = hstep;
    ReentryCost = 0;
    if ((HasEntryCost==1) && (HasTenure==0))
       {
        ReentryCost = EntryPenalty; 
       }
    hmax2 = leispref - fixcost - (leisprefbad*badheal) - ReentryCost - hcent*hstep;
    hmax2 = hstep*floor(hmax2/hstep);  /*  Round, so that (overall) hours grid does not vary with c2  */

    hmin3 = 0.0;
    hmin4 = hmin3 + hstep;

    oldclev  = (int) floor(rounder+*clevPtr);
    loc      = oldclev - csteps;
    if (loc < 1) loc = 1;
    loloc    = loc - xvstep;
    if (loloc < 0) loloc = 0;
    hic      = loc + 2*csteps;
    if (hic >(cnum-2)) hic = (cnum-2); 
    hihic    = hic + xvstep;
    if (hihic > (cnum-1)) hihic = cnum-1;
    cmaxlevi = loloc;  /*  cmaxlev is a double  */
    c2       = cdecPtr[cmaxlevi];
    *crec    = c2;
    *hrec    = 0;

/*  Focus on consumption indices in neighborhood of old cmaxlev  */
/*  Reorient search if any point +-3 indices from the neighborhood's  */
/*  edges outperforms points in interior of neighborhood  */

/*  val can be a large negative num, safest to initialize this way */

    val = computethevaluefunction(
            *crec, *hrec, valftpoPtr, doespart, assets, wage, AIME, 
            AstatePtr, AIMstatePtr, hcstatePtr, wtransprobPtr, 
            HITtransprobPtr, hctransprobPtr, val, &start, 
            mnlnhc, stdlnhc, hcnodesPtr, hcwgtsPtr, &transfers, 
            pencfsPtr, penaccPtr, penkinkPtr, &illiquidsolution, 
            PIA, ssbenes, penbal, penbenefits, badheal, 
            survprob, probbad);

/*  start = 0;  */

    for (i = loloc; i < (hihic+1); i++)
       {
        c2 = cdecPtr[i];

    /*  Must check lower work hours levels because of the earnings   */
    /*  test, and higher levels because of the effect on AIME, but   */
    /*  use the basic WxMU_C = MU_L condition as a guide.            */

        hcstar = leispref - fixcost - (leisprefbad*badheal) - ReentryCost 
                 - (Onemgamma/_gamma)*c2/wage;
        hcstar = hstep*ceil(hcstar/hstep);  /*  Round, so that (overall) hours grid does not vary with c2  */
        if (hcstar < hmin2) hcstar = hmin2;
        if (hcstar > hmax2) hcstar = hmax2;
        hc = hmax2;

        for (j = 1; (j < (hgridn+1)) && (hc > hmin4); j++)
           {
            hc=hcstar-(j-(hcent+1))*hstep;

            valc = computethevaluefunction(
                    c2, hc, valftpoPtr, doespart, assets, wage, AIME,
                    AstatePtr, AIMstatePtr, hcstatePtr, wtransprobPtr, 
                    HITtransprobPtr, hctransprobPtr, val, &start, 
                    mnlnhc, stdlnhc, hcnodesPtr, hcwgtsPtr, &transfers, 
                    pencfsPtr, penaccPtr, penkinkPtr, &noliquidity, 
                    PIA, ssbenes, penbal, penbenefits, badheal, 
                    survprob, probbad);

            if ( (valc > val) || (noliquidity < illiquidsolution) )
               {
                cmaxlevi=i;
                *crec=c2;
                if (transfers > rounder)
                   {
                    *crec=cfloor;
                    cmaxlevi=cfloorindx;
                   }
                *hrec=hc;
                val=valc;
                illiquidsolution = noliquidity;
               }
/*          start = 0;  */
           }  /*  End loop through hours  */

        if (i == hihic)
           {
        /*  Check to see if utility-maximimizing consumption lies at boundary of  */
        /*  search space and adjust space if necessary  */
            start = checkmax(cmaxlevi, cnum, csteps, &loloc, &loc, &hic, &hihic, 
                             &i, illiquidsolution);
           }
       }

    *clevPtr = cmaxlevi;
    return val;
   }

/*---------------------------------------------------------------------------*/
/*---------------------------------------------------------------------------*/
/*  WORKRULE2:  Finds the optimal consumption/hours choice for workers  */
/*              Search centered around optimal decision from earlier search  */

double workrule2(
    double assets, double wage, double AIME, double *wtransprobPtr, 
    double *HITtransprobPtr, double *hctransprobPtr, double survprob, 
    double probbad, double cincr, double hincr, int srch2steps_c, 
    int srch2steps_h, double mnlnhc, double stdlnhc, double *hcnodesPtr, 
    double *hcwgtsPtr, double *pencfsPtr, double *penaccPtr, 
    double *penkinkPtr, double *AstatePtr, double *AIMstatePtr, 
    double *hcstatePtr, double PIA, double ssbenes, double penbal, 
    double penbenefits, double *crec, double *hrec, int badheal, 
    double ********valftpoPtr)

   {
    double val, valc, oldc, oldh, transfers, cmin, cmax, c2, hc, 
           hmin, hmin2, hmax, hmax2, ReentryCost;
    int start, noliquidity, illiquidsolution, maxstepc, maxsteph, i, j; 
	
    val   = 0.0;  /*  U(c,h) + BxE_t{V[t+1]}:  returned to calling function  */
    valc  = 0.0;
    oldc  = *crec;
    oldh  = *hrec;
    start = 1;
    noliquidity = 0;
    illiquidsolution = 0;
    transfers = 0.0;

/*  Focus on consumption/hours in neighborhood of oldc and oldh  */
/*  Reorient search if endpoints outperform neighborhood  */

/*  val can be a large negative num, safest to initialize this way  */

    *crec = oldc;
    *hrec = oldh;
    val = computethevaluefunction(
            *crec, *hrec, valftpoPtr, doespart, assets, wage, AIME,
            AstatePtr, AIMstatePtr, hcstatePtr, wtransprobPtr, 
            HITtransprobPtr, hctransprobPtr, val, &start, 
            mnlnhc, stdlnhc, hcnodesPtr, hcwgtsPtr, &transfers, 
            pencfsPtr, penaccPtr, penkinkPtr, &illiquidsolution, 
            PIA, ssbenes, penbal, penbenefits, badheal, 
            survprob, probbad);

    hmin     = oldh - srch2steps_h*hincr;
    hmin2    = 0.0;  
    hmax     = oldh + (srch2steps_h-1)*hincr;
    ReentryCost = 0;
    if ((HasEntryCost==1) && (HasTenure==0))
       {
        ReentryCost = EntryPenalty; 
       }
    hmax2    = leispref-fixcost-(leisprefbad*badheal)-ReentryCost;
    maxsteph = 2*srch2steps_h+2;
    hc       = hmin;
/*  start = 0;  */

    for (i = 1; (i < maxsteph) && (hc < hmax2); i++) /* loop through hours first  */
       {
        if (hc < hmin2)
           {
            hc = hc + hincr;
            continue;
           }

        cmin      = oldc*pow(1/cincr,srch2steps_c);
        cmax      = oldc*pow(cincr,srch2steps_c-1);
        c2        = cmin;
        maxstepc  = 2*srch2steps_c+2;
        transfers = 0.0;

    /*  If transfers > 0, cash on hand cannot support the consumption floor,   */
    /*  and there is no need to check any other consumption level (we assume   */
    /*  that the floor is so low that the agent never saves)   */

        for (j = 1; j < maxstepc; j++)
           {
            valc = computethevaluefunction(
                    c2, hc, valftpoPtr, doespart, assets, wage, AIME, 
                    AstatePtr, AIMstatePtr, hcstatePtr, wtransprobPtr, 
                    HITtransprobPtr, hctransprobPtr, val, &start, 
                    mnlnhc, stdlnhc, hcnodesPtr, hcwgtsPtr, &transfers, 
                    pencfsPtr, penaccPtr, penkinkPtr, &noliquidity, 
                    PIA, ssbenes, penbal, penbenefits, badheal, 
                    survprob, probbad);

            if ( (valc > val) || (noliquidity < illiquidsolution) )
               {
                *crec = c2;
                if (transfers > rounder) *crec = cfloor;
                *hrec = hc;
                val = valc;
                illiquidsolution = noliquidity;
               }

            if (transfers > rounder) break;
/*          start = 0;  */
            c2 = c2*cincr;

            if (j < maxstepc-1) continue;  /*  Avoid the mess below when possible  */

        /*  Check to see if utility-maximimizing consumption lies at boundary of  */
        /*  search space and adjust space if necessary  */
			
            c2 = checkmax2(&cmin, &cmax, crec, &start, &maxstepc, cincr, c2, 
                           srch2steps_c, illiquidsolution);

           }  /*  End loop through consumption  */

        hc = hc + hincr;
        if (hc > hmax2) i = (maxsteph-1);
        if (i < maxsteph-1) continue;  /*  Avoid the mess below when possible  */

    /*  Check to see if utility-maximimizing hours lies at boundary of  */
    /*  search space and adjust space if necessary  */

        if (*hrec > hmax)
           {
            if (hmax < (hmax2-hincr))
               {
                maxsteph = maxsteph + srch2steps_h;
                hmin     = hmax - hincr;
                hmax     = hmax + srch2steps_h*hincr;
                continue;
               }
            continue;
           }

        if (*hrec > hmin) continue;

        if (hmin > (hmin2+hincr))  /*  h == hmin  */
           {
            maxsteph = maxsteph+srch2steps_h;
            hmax     = hmin + hincr;           
            hmin     = hmin - hincr*srch2steps_h;
            hc       = hmin;                    
            continue;
           }
        continue;
       }  /*  End loop through hours  */

    return val;
   }

/*---------------------------------------------------------------------------*/
/*---------------------------------------------------------------------------*/
/*  GETRULES:  Entry point for the main program--Decision rules  */
/*             Finds the optimal consumption/hours/participation choices  */

int getrules(
    double *searchparmPtr, double *sscredsPtr, double *AstatePtr, 
    double *WstatePtr, double *AIMstatePtr, double *hcstatePtr, 
    double *cdecPtr, double *wagetransprobPtr, double *HITtransprobwPtr, 
    double *HITtransprobnwPtr, double *hctransprobPtr, 
    double *rawhcnodesPtr, double *hcwgtsPtr, double *pencfsPtr, 
    double *penaccPtr, double *penaccprobPtr, double *penkinkPtr, 
    double *wagetrendPtr, double *probbadPtr, double *mnlnhcwPtr, 
    double *mnlnhcnwPtr, double *stderrhcwPtr, double *stderrhcnwPtr, 
    double ********clevmaxPtr, double ********clevlazPtr, 
    double ********holdclevmaxPtr, double ********holdclevlazPtr, 
    double ********valftpoPtr, double ********maxconPtr, 
    double ********lazconPtr, double ********hoursPtr, 
    double ********holdmaxconPtr, double ********holdlazconPtr, 
    double ********holdhoursPtr, double ********partPtr, 
    double ********valfPtr, double ********pensionPtr, 
    double *survprobPtr, double *PrefParmPtr, 
    int iAssetsmin, int iAssetsmax)

   {
/*  Dereference pointers to parameters  */

    int searchtype   = (int) floor(rounder+searchparmPtr[0]);
    int csteps       = (int) floor(rounder+searchparmPtr[1]);
    int srch2steps_c = (int) floor(rounder+searchparmPtr[2]);
    int srch2steps_h = (int) floor(rounder+searchparmPtr[3]);
    int cnum         = (int) floor(rounder+searchparmPtr[4]);
    int hgridn       = (int) floor(rounder+searchparmPtr[5]);
    int hcent        = (int) floor(rounder+searchparmPtr[6]);
    int cfloorindx   = (int) floor(rounder+searchparmPtr[7]);
    double hstep     = searchparmPtr[8];
    double cincr     = searchparmPtr[9];
    double hincr     = searchparmPtr[10];
    int lookagain    = (int) floor(rounder+searchparmPtr[11]);

/*  Obtain the globals eretadj, cumretadj, bigcred, litcred  */
/*  These are age-specific adjustment factors for early or late SS application  */
    int gotadjfactors = getadj(sscredsPtr);
    int gotPrefParms;
    double cumadj2 = pow(eretadj,(nret-age));

    double assets, wage, AIME, mnlnhcw, mnlnhcnw, stdlnhcw, stdlnhcnw, 
           survprob, probbad;
    double PIAkink[2];
    double valfw, valfnw, crec, hrec, clevmax, clevlaz;
    int iType, iTenure, iHIT, iHS, iHC, iAIME, iWage, iAssets, HITshift;
    int badheal, hitype;
    double *HITtransprobw;
    double *HITtransprobnw;
    double penaccfs[5];
    double *wagetransprob;
    double *hctransprob;
    double PIA = 0;
    double ssbenes = 0; 
    double penbal = 0; 
    double penbenefits = 0;

    double *nppencfsPtr, *nppenkinkPtr;
    nppencfsPtr  = &pencfsPtr[HITnum+3];
    nppenkinkPtr = &penkinkPtr[2*HITnum];

    for(iType=0; iType<PrefTypeNum; iType++)
       {
        PrefType = iType;  /*  Global  */
        gotPrefParms = getprefparms(PrefParmPtr,iType);  /*  Globals  */

        for(iTenure=0; iTenure<(HasEntryCost+1); iTenure++)
           {
            HasTenure = iTenure;  /*  Global  */

            if ((age >= manretage) && (iTenure>0))  /*  Working is not option, so tenure is irrelevant  */
               {
                for(iHIT=0; iHIT<HITnum; iHIT++)
                   {
                    for(iHS=0; iHS<HSnum; iHS++)
                       {
                        for(iHC=0; iHC<hcnum; iHC++)
                           {
                            for(iAIME=0; iAIME<AIMEnum; iAIME++)
                               {
                                for(iWage=0; iWage<_Wnum; iWage++)
                                   {
                                    for(iAssets=iAssetsmin; iAssets<iAssetsmax; iAssets++)
                                       {
                                        holdclevlazPtr[iType][iTenure][iHIT][iHS][iHC][iAIME][iWage][iAssets] 
                                            = holdclevlazPtr[iType][0][iHIT][iHS][iHC][iAIME][iWage][iAssets]; 
                                        holdlazconPtr[iType][iTenure][iHIT][iHS][iHC][iAIME][iWage][iAssets]
                                            = holdlazconPtr[iType][0][iHIT][iHS][iHC][iAIME][iWage][iAssets]; 
                                        holdclevmaxPtr[iType][iTenure][iHIT][iHS][iHC][iAIME][iWage][iAssets]
                                            = holdclevmaxPtr[iType][0][iHIT][iHS][iHC][iAIME][iWage][iAssets];
                                        holdmaxconPtr[iType][iTenure][iHIT][iHS][iHC][iAIME][iWage][iAssets]
                                            = holdmaxconPtr[iType][0][iHIT][iHS][iHC][iAIME][iWage][iAssets];
                                        holdhoursPtr[iType][iTenure][iHIT][iHS][iHC][iAIME][iWage][iAssets]
                                            = holdhoursPtr[iType][0][iHIT][iHS][iHC][iAIME][iWage][iAssets];
                                        partPtr[iType][iTenure][iHIT][iHS][iHC][iAIME][iWage][iAssets]
                                            = partPtr[iType][0][iHIT][iHS][iHC][iAIME][iWage][iAssets];
                                        valfPtr[iType][iTenure][iHIT][iHS][iHC][iAIME][iWage][iAssets]
                                            = valfPtr[iType][0][iHIT][iHS][iHC][iAIME][iWage][iAssets];
                                        pensionPtr[iType][iTenure][iHIT][iHS][iHC][iAIME][iWage][iAssets]
                                            = pensionPtr[iType][0][iHIT][iHS][iHC][iAIME][iWage][iAssets];

                                       }  /*  End loop through assets  */
                                   }  /*  End loop through wages  */
                               }  /*  End loop through AIME  */
                           }  /*  End loop through persistent health costs  */
                       }  /*  End loop through health states  */
                   }  /*  End loop through health insurance types  */
                continue;
               }  /*  End update of irrelevant tenure statuses  */

            for(iHIT=0; iHIT<HITnum; iHIT++)
               {
                hitype         = iHIT;
                HITtransprobw  = &HITtransprobwPtr[iHIT*HITnum];
                HITtransprobnw = &HITtransprobnwPtr[iHIT*HITnum];
                PIAkink[0]     = penkinkPtr[iHIT];
                PIAkink[1]     = penkinkPtr[iHIT+HITnum];
                penaccfs[0]    = penaccPtr[iHIT];
                penaccfs[1]    = penaccPtr[HITnum];
                penaccfs[2]    = penaccprobPtr[iHIT];
                penaccfs[3]    = penaccprobPtr[HITnum+iHIT];
                penaccfs[4]    = penaccprobPtr[2*HITnum];

                for(iHS=0; iHS<HSnum; iHS++)
                   {
                    badheal   = iHS;
                    survprob  = survprobPtr[iHS];
                    probbad   = probbadPtr[iHS];

                    HITshift  = iHIT+iHS*HITnum;
                    mnlnhcw   = mnlnhcwPtr[HITshift];
                    mnlnhcnw  = mnlnhcnwPtr[HITshift];
                    stdlnhcw  = stderrhcwPtr[HITshift];
                    stdlnhcnw = stderrhcnwPtr[HITshift];
	
                    for(iHC=0; iHC<hcnum; iHC++)
                       {
                        hctransprob = &hctransprobPtr[iHC*hcnum];

                        for(iAIME=0; iAIME<AIMEnum; iAIME++)
                           {
                            AIME = AIMstatePtr[iAIME];
                            penbenefits = computepenstuff(AIME, pencfsPtr, &PIA, &ssbenes, 
                                                          &penbal, PIAkink, iHIT);
	
                            for(iWage=0; iWage<_Wnum; iWage++)
                               {
                                wage = exp(WstatePtr[iWage]+wagetrendPtr[badheal]);
                                wagetransprob = &wagetransprobPtr[iWage*_Wnum];
                                for(iAssets=iAssetsmin; iAssets<iAssetsmax; iAssets++)
                                   {
                                    pensionPtr[iType][iTenure][iHIT][iHS][iHC][iAIME][iWage][iAssets] = penbal;
                                    assets  = AstatePtr[iAssets];
                                    clevlaz = clevlazPtr[iType][iTenure][iHIT][iHS][iHC][iAIME][iWage][iAssets] - 1; /* GAUSS to C indexing */
                                    crec    = lazconPtr[iType][iTenure][iHIT][iHS][iHC][iAIME][iWage][iAssets];

                                    if (searchtype<3)
                                       {
                                        valfnw = lazyrule(assets, wage, AIME, wagetransprob, 
                                                          HITtransprobnw, hctransprob, survprob, 
                                                          probbad, csteps, cnum, cdecPtr, &clevlaz,  
                                                          mnlnhcnw, stdlnhcnw, rawhcnodesPtr, 
                                                          hcwgtsPtr, nppencfsPtr, penaccfs, 
                                                          nppenkinkPtr, AstatePtr, AIMstatePtr, 
                                                          hcstatePtr, PIA, ssbenes, penbal, 
                                                          penbenefits, &crec, badheal, valftpoPtr);
                                        if (lookagain==1)
                                            valfnw = lazyrule2(assets, wage, AIME, wagetransprob, 
                                                               HITtransprobnw, hctransprob, survprob, 
                                                               probbad, cincr, srch2steps_c, mnlnhcnw, 
                                                               stdlnhcnw, rawhcnodesPtr, hcwgtsPtr, 
                                                               nppencfsPtr, penaccfs, nppenkinkPtr, 
                                                               AstatePtr, AIMstatePtr, hcstatePtr, 
                                                               PIA, ssbenes, penbal, penbenefits, 
                                                               &crec, badheal, valftpoPtr);
                                       }
                                    else /*  searchtype==3 or 4  */
                                       {
                                        valfnw = lazyrule2(assets, wage, AIME, wagetransprob, 
                                                           HITtransprobnw, hctransprob, survprob, 
                                                           probbad, cincr, srch2steps_c, mnlnhcnw, 
                                                           stdlnhcnw, rawhcnodesPtr, hcwgtsPtr, 
                                                           nppencfsPtr, penaccfs, nppenkinkPtr, 
                                                           AstatePtr, AIMstatePtr, hcstatePtr, 
                                                           PIA, ssbenes, penbal, penbenefits, 
                                                           &crec, badheal, valftpoPtr);
                                       }

                                    holdclevlazPtr[iType][iTenure][iHIT][iHS][iHC][iAIME][iWage][iAssets] = clevlaz + 1;  /*  updated  */
                                    holdlazconPtr[iType][iTenure][iHIT][iHS][iHC][iAIME][iWage][iAssets]  = crec; 

                                    partPtr[iType][iTenure][iHIT][iHS][iHC][iAIME][iWage][iAssets]        = 0;
                                    valfPtr[iType][iTenure][iHIT][iHS][iHC][iAIME][iWage][iAssets]        = valfnw;
                                    holdclevmaxPtr[iType][iTenure][iHIT][iHS][iHC][iAIME][iWage][iAssets] = clevlaz + 1;
                                    holdmaxconPtr[iType][iTenure][iHIT][iHS][iHC][iAIME][iWage][iAssets]  = crec;
                                    holdhoursPtr[iType][iTenure][iHIT][iHS][iHC][iAIME][iWage][iAssets]   = 0;

                                    if (age < manretage)  /*  Working is an option  */
                                       {
                                        clevmax = clevmaxPtr[iType][iTenure][iHIT][iHS][iHC][iAIME][iWage][iAssets] - 1; /* GAUSS to C indexing */
                                        crec    = maxconPtr[iType][iTenure][iHIT][iHS][iHC][iAIME][iWage][iAssets];
                                        hrec    = hoursPtr[iType][iTenure][iHIT][iHS][iHC][iAIME][iWage][iAssets];

                                        if (searchtype<3)
                                           {
                                            valfw = workrule(assets, wage, AIME, wagetransprob, 
                                                             HITtransprobw, hctransprob, survprob, 
                                                             probbad, csteps, cnum, cdecPtr, 
                                                             &clevmax, hgridn, hcent, hstep, 
                                                             mnlnhcw, stdlnhcw, rawhcnodesPtr, 
                                                             hcwgtsPtr, nppencfsPtr, penaccfs, 
                                                             nppenkinkPtr, AstatePtr, AIMstatePtr, 
                                                             hcstatePtr, PIA, ssbenes, penbal, 
                                                             penbenefits, &crec, &hrec, badheal, 
                                                             valftpoPtr);
                                            if (lookagain==1)
                                                valfw = workrule2(assets, wage, AIME, wagetransprob, 
                                                                  HITtransprobw, hctransprob, survprob, 
                                                                  probbad, cincr, hincr, srch2steps_c, 
                                                                  srch2steps_h, mnlnhcw, stdlnhcw, 
                                                                  rawhcnodesPtr, hcwgtsPtr, nppencfsPtr, 
                                                                  penaccfs, nppenkinkPtr, AstatePtr, 
                                                                  AIMstatePtr,  hcstatePtr, PIA, ssbenes, 
                                                                  penbal, penbenefits, &crec, &hrec, 
                                                                  badheal, valftpoPtr);
                                           }
                                        else /*  searchtype==3 or 4  */
                                           {
                                            valfw = workrule2(assets, wage, AIME, wagetransprob, 
                                                              HITtransprobw, hctransprob, survprob, 
                                                              probbad, cincr, hincr, srch2steps_c, 
                                                              srch2steps_h, mnlnhcw, stdlnhcw, 
                                                              rawhcnodesPtr, hcwgtsPtr, nppencfsPtr, 
                                                              penaccfs, nppenkinkPtr, AstatePtr, 
                                                              AIMstatePtr,  hcstatePtr, PIA, ssbenes, 
                                                              penbal, penbenefits, &crec, &hrec, 
                                                              badheal, valftpoPtr);
                                           }

                                        holdclevmaxPtr[iType][iTenure][iHIT][iHS][iHC][iAIME][iWage][iAssets] = clevmax + 1;  /*  updated  */
                                        holdmaxconPtr[iType][iTenure][iHIT][iHS][iHC][iAIME][iWage][iAssets]  = crec;
                                        holdhoursPtr[iType][iTenure][iHIT][iHS][iHC][iAIME][iWage][iAssets]   = hrec;

                                        if (valfw > valfnw)
                                           {
                                            valfPtr[iType][iTenure][iHIT][iHS][iHC][iAIME][iWage][iAssets] = valfw;                                 
                                            if (hrec > 0) partPtr[iType][iTenure][iHIT][iHS][iHC][iAIME][iWage][iAssets] = 1; /* hrec>0 oughta be redundant */
                                           }
                                       }  /*  End working age analysis  */ 
                                   }  /*  End loop through assets  */
                               }  /*  End loop through wages  */
                           }  /*  End loop through AIME  */
                       }  /*  End loop through persistent health costs  */
                   }  /*  End loop through health states  */
               }  /*  End loop through health insurance types  */
           }  /*  End loop through job tenure statuses  */
       }  /*  End loop through preference types  */ 

    return 1;
   }

/*---------------------------------------------------------------------------*/
/*---------------------------------------------------------------------------*/
/*  Below is C++/GAUSS I/O code written by Ken Housinger                     */
/*---------------------------------------------------------------------------*/
/*---------------------------------------------------------------------------*/
/*  This function takes a char* and either reverses bits within bytes or bytes */
/*  themselves or both.  It returns a pointer to the original buffer which is  */
/*  altered. */

unsigned char * gread(unsigned char *inbuf, int bytes, int byte_rev, int bit_rev) 
   {
    unsigned char *tempbuf;
    unsigned char tempbyte, tempbit;
    int i, j;

    tempbuf = (unsigned char *) malloc(bytes);
    for (i = 0; i < bytes; i++) 
       {
        if (byte_rev) *(tempbuf + i) = *(inbuf + bytes - i - 1);
        else *(tempbuf + i) = *(inbuf + i);

        if (bit_rev) 
           {
            tempbyte = 0;
            for (j = 0; j < CHAR_BIT; j++) 
               {
                tempbit = *(tempbuf + i) >> (CHAR_BIT - j - 1);
                tempbit = tempbit << (CHAR_BIT - 1);
                tempbit = tempbit >> (CHAR_BIT - j - 1);
                tempbyte = tempbyte | tempbit;
               }
            *(tempbuf + i) = tempbyte;
           }
       }

    for (i = 0; i < bytes; i++)
        *(inbuf + i) = *(tempbuf + i);
    free(tempbuf);

    return(inbuf);
   }

/*---------------------------------------------------------------------------*/
/*---------------------------------------------------------------------------*/
/*  This function reads a Gauss v5.0 fmt file into a Matrix structure and  */
/*  returns the structure.  */

Matrix gau5read(char *fmt) 
   {
/*  Initialize the matrix to be 1x1 and the byte/bit order to 0.  */
    Matrix mat = {1, 1}; 
    unsigned int i;
    int type, byte, bit;
    unsigned char *cread;
    int bit_rev = 0, byte_rev = 0;
    FILE *fd;

    if (sizeof(int) != 4 || sizeof(double) != 8 || CHAR_BIT != 8) 
       {
        printf("Incompatable machine architecture.\n");
        return (mat);
       }

/*  Allocate enough space to store the header.  */
    cread = (unsigned char *) malloc(BASIC_HEADER_LEN); 
/*  Open *fmt for reading only.  */
    fd = fopen(fmt, "rb"); 
  
/*  Read the basic header (128 bytes) all at once.  */

    fread((void *) cread, 1, BASIC_HEADER_LEN, fd);
    byte = (int) *(cread + (BYTE_POS * sizeof(int)));  /* (0=Backward) */
    bit = (int) *(cread + (BIT_POS * sizeof(int)));    /* (0=Backward) */

/*  To get some system independence, we detect whether we have to reverse */
/*  the bytes or bits or both.  If x and x_SYSTEM match, no reverse is  */
/*  necessary. */

    if ((bit || BIT_SYSTEM) && !(bit && BIT_SYSTEM)) bit_rev=1;
    if ((byte || BYTE_SYSTEM) && !(byte && BYTE_SYSTEM)) byte_rev=1;

    type = *( (int *) gread((cread + (TYPE_POS * sizeof(int))), sizeof(int), 
                             byte_rev, bit_rev) );

/*  If the fmt file type is not a scalar, there are another two */
/*  ints of header giving the values of m and n.  If a matrix, also reset n. */

    if (type > SCALAR) 
       { 
        fread((void *) cread, 1, sizeof(int), fd);
        mat.m = *((unsigned int *) gread(cread, sizeof(int), byte_rev, bit_rev));
        fread((void *) cread, 1, sizeof(int), fd);
        if (type == MATRIX)
          mat.n = *((unsigned int *) gread(cread, sizeof(int), byte_rev, bit_rev));
      } 

/*  Allocate memory for the matrix.  The amount needed is m * n * sizeof(double). */
/*  Next, read in the data all at once.  Then use gread to reverse */
/*  bytes/bits if necessary. */

    free(cread);

    mat.data = (double *) malloc(mat.m * mat.n * sizeof(double));
    fread((void *) mat.data, sizeof(double), mat.m * mat.n, fd);
    if (byte_rev || bit_rev)
      for(i = 0; i < mat.m * mat.n; i++)
        gread((unsigned char *) mat.data + (i * sizeof(double)), sizeof(double), 
               byte_rev, bit_rev);

    fclose(fd);

    return (mat);
   }

/*---------------------------------------------------------------------------*/
/*---------------------------------------------------------------------------*/
/*  This function writes a Gauss v5.0 fmt file from a Matrix structure. */

void gau5write(char *fmt, Matrix mat) 
   {
/*  This ugly mess is the basic header. */

    unsigned int header[(BASIC_HEADER_LEN / sizeof(int)) + 2] = 
        {0xffffffff, 0, 0xffffffff, 0, 0xffffffff, 0, 0,
         0xabcdef01,1, 0, 1, 1008, sizeof(double), 0, 1,
         SCALAR, 1, 0, BASIC_HEADER_LEN};

    FILE *fd;

    if (sizeof(int) != 4 || sizeof(double) != 8 || CHAR_BIT != 8) 
       {
        printf("Incompatable machine architecture.\n");
        return;
       }

/*  If forward byte, make 6th int 0xffffffff. */
/*  If forward bit, make 7th int 0xffffffff. */

    if (BYTE_SYSTEM) header[BYTE_POS] = 0xffffffff;
    if (BIT_SYSTEM) header[BIT_POS] = 0xffffffff;

/*  If not a scalar, increase the 16th int by 1 and the 19th int (header */ 
/*  length) by 8 (2 * sizeof(int)).  Also, set m in int 33. */

    if (!(mat.m * mat.n == 1)) 
       {
        header[TYPE_POS] += 1;
        header[HEADER_LEN_POS] += (2 * sizeof(int));
        header[M_POS] = mat.m;

    /*  If not a vector (and not a scalar), increase the 16th int by 1 again */
    /*  and set m in int 34. */
        if (!(mat.n == 1)) 
           {
            header[TYPE_POS] += 1;
            header[N_POS] = mat.n;
           }
       }
  /*
  **Open fmt for writing and create if it does not exist.  If you create it,
  **make it a regular file with permissions 0640.  See comment in gau5read
  **for detail on how read (and similarly, write) work.
  **
  **Order: Write the basic header
  **       If not a scalar, write the other 2 ints of header 
  **       Write the m * n elements of data
  */

    fd = fopen(fmt, "wb"); 
    if ((mat.m * mat.n == 1))
      fwrite((void *) header, 1, BASIC_HEADER_LEN, fd);
    else
      fwrite((void *) header, 1, BASIC_HEADER_LEN + (2 * sizeof(int)), fd);
    fwrite((void *) mat.data, sizeof(double), mat.m * mat.n, fd);
    fclose(fd);
   }  
