#include <string.h>
#include <limits.h>
#include <stdlib.h>

/*---------------------------------------------------------------------------*/
/*---------------------------------------------------------------------------*/
/*  The next three procedures are C/GAUSS I/O routines by Ken Housinger      */

#define BASIC_HEADER_LEN 128 /*Length of a scalar header*/
#define BYTE_POS 5           /*Int offset in header of byte order*/
#define BIT_POS 6            /*Int offset in header of bit order*/
#define TYPE_POS 15          /*Int offset in header of matrix type*/
#define HEADER_LEN_POS 18    /*Int offset in header of header length*/
#define M_POS 32             /*Int offset in header of value for m*/
#define N_POS 33             /*Int offset in header of value for n*/
#define SCALAR 0             /*Value of type that indicates a scalar*/
#define MATRIX 2             /*Value of type that indicates a matrix*/
#define BIT_SYSTEM 0         /*Value of bit order (0=backwards/i386)*/
#define BYTE_SYSTEM 0        /*Value of byte order (0=backwards/i386)*/

/*
**This is my Matrix opaque datatype.  The gauss fmt format has implied 
**dimensionality, but is just a list of doubles.  I could rewrite the
**functions to use pointers-to-pointers to improve the interface with
**Eric's functions, but that would add to the complxity of my code.  I
**suggest writing an intermediate function that turns *data into **data
**to cope with 2-dimensional arrays.
*/ 
typedef struct {
  unsigned int m;
  unsigned int n;
  double *data;
} Matrix;
/*  
**Currently, these functions require full filenames (you must provide the  
**the ".fmt").                                                           
*/
unsigned char * gread(unsigned char *inbuf, int bytes, int byte_reverse, int bit_reverse);

Matrix gau5read(char *fmt);

void gau5write(char *fmt, Matrix mat);

/*---------------------------------------------------------------------------*/
/*---------------------------------------------------------------------------*/

Matrix zeromat(unsigned int recsize);

double ********zeromat2();

void switchem(double **dataPtr, double **tempPtr);

double ********VectorToMatrix(double *dataVec);

int Locate(double *Xarray, double x, int DIM);

double U(double c, double h, double particip, int badheal);

double beq(double assets);

double ass(double AfterTaxIncome, double c, double laborincome, double assets, 
           double *reduc, double ssbenes);

double computelaborincome(double wage, double h);

double computeAfterTaxIncome(double laborincome, double assets, double *MTR,
                             int badheal, double penbenefits);

double lossy(double valfunction, double assets, int noliquidity);

double interp(double Avar, double AIMEvar, double *AstatePtr, 
              double *AIMstatePtr, int prtype, int tentype, int hitype, 
              int badheal, int hctype, int wagetype, double ********VvarPtr);

double computeAIME(double AIME, double laborincome);

double computePIA(double AIME);

double findAIME(double PIA, int imposemax);

double predictpenben(double PIA, double *pencfsPtr, double PIAkink[2], int HItype);

double findPIA(double totben, double *pencfsPtr, double PIAkink[2], 
               double totbenkink[2], int HItype); 

double computepenaccrue(double laborincome, double *penaccPtr);

int getadj(double *sscredsPtr);

double computepenstuff(double AIME, double *pencfsPtr, double *PIA, 
                       double *ssbenes, double *penbal, double PIAkink[2],
                       int HItype);

double computethevaluefunction(
    double *c, double h, double ********valftpoPtr, double particip, double assets, 
    double wage, double AIME, double *AstatePtr, double *AIMstatePtr, 
    double *hcstatePtr, double *wtransprobPtr, double *HITtransprobPtr, 
    double *hctransprobPtr, double oldval, int *start, double mnlnhc, 
    double stdlnhc, double *hcnodesPtr, double *hcwgtsPtr, double *transfers, 
    double *pencfsPtr, double *penaccPtr, double *penkinkPtr, double PIA,  
    double ssbenes, double penbal, double penbenefits, int badheal, 
    double survprob, double probbad) ;

int getprefparms(double *PrefParmPtr,int ThisPrefType);

int globderef(double *interpparmPtr, double *gridptnumPtr, 
              double *ageparmPtr, double *miscparmPtr, 
              double *applyPtr, double *survprobPtr, double *agePtr, 
              double *AIMEbkPtr, double *mprPtr, double *PIAbkPtr);
     
double lazyrule(
    double assets, double wage, double AIME, double *wtransprobPtr, 
    double *HITtransprobPtr, double *hctransprobPtr, double survprob, 
    double probbad, int cnum, double *cdecPtr, double *clevPtr, 
    double mnlnhc, double stdlnhc, double *hcnodesPtr, double *hcwgtsPtr, 
    double *pencfsPtr, double *penaccPtr, double *penkinkPtr, 
    double *AstatePtr, double *AIMstatePtr, double *hcstatePtr,  
    double PIA, double ssbenes, double penbal, double penbenefits,  
    double *crec, int badheal, double ********valftpoPtr);

double workrule(
    double assets, double wage, double AIME, double *wtransprobPtr, 
    double *HITtransprobPtr, double *hctransprobPtr, double survprob, 
    double probbad, int cnum, double *cdecPtr, double *clevPtr, 
    int hgridn, int hcent, double hstep, double mnlnhc, double stdlnhc,
    double *hcnodesPtr, double *hcwgtsPtr, double *pencfsPtr, 
    double *penaccPtr, double *penkinkPtr, double *AstatePtr, 
    double *AIMstatePtr, double *hcstatePtr, double PIA, double ssbenes, 
    double penbal, double penbenefits, double *crec, double *hrec, 
    int badheal, double ********valftpoPtr);

double workrule2(
    double assets, double wage, double AIME, double *wtransprobPtr, 
    double *HITtransprobPtr, double *hctransprobPtr, double survprob, 
    double probbad, double hincr,  int srch2steps_h, double mnlnhc, 
	double stdlnhc, double *hcnodesPtr, double *hcwgtsPtr, double *pencfsPtr, 
    double *penaccPtr, double *penkinkPtr, double *AstatePtr, 
    double *AIMstatePtr, double *hcstatePtr, double PIA, double ssbenes, 
    double penbal, double penbenefits, double *crec, double *hrec, 
    int badheal, double ********valftpoPtr);

int getrules(
    double *searchparmPtr, double *sscredsPtr, double *AstatePtr, 
    double *WstatePtr, double *AIMstatePtr, double *hcstatePtr, 
    double *cdecPtr, double *wagetransprobPtr, double *HITtransprobwPtr, 
    double *HITtransprobnwPtr, double *hctransprobPtr, 
    double *rawhcnodesPtr, double *hcwgtsPtr, double *pencfsPtr, 
    double *penaccPtr, double *penaccprobPtr, double *penkinkPtr, 
    double *wagetrendPtr, double *probbadPtr, double *mnlnhcwPtr, 
    double *mnlnhcnwPtr, double *stderrhcwPtr, double *stderrhcnwPtr, 
    double ********clevmaxPtr, double ********clevlazPtr, 
    double ********holdclevmaxPtr, double ********holdclevlazPtr, 
    double ********valftpoPtr, double ********maxconPtr, 
    double ********lazconPtr, double ********hoursPtr, 
    double ********holdmaxconPtr, double ********holdlazconPtr, 
    double ********holdhoursPtr, double ********partPtr, 
    double ********valfPtr, double ********pensionPtr, 
    double *survprobPtr, double *PrefParmPtr, 
    int iAssetsmin, int iAssetsmax);
