#include <stdio.h>
#include <math.h>
#include "sim63.h"
#include <mpi.h>
#pragma hdrstop

#define useMPI 2 /* 0 => PC, 1 => Unix, 2 => cluster */
#define HSnum 2   /*  health states  */
#define rounder 0.01
#define negpen 1000       /*  scales penalty for violating liquidity constraint  */
#define miserpen 1000000  /*  scales penalty for exceeding asset maximum  */
#define xvstep 3          /*  Gap between "main body" of search area and extreme values  */
#define nopart 0
#define doespart 1
#define notransfers 0
#define dieage 95
#define penavailage 95    /*  Pensions are treated as fully liquid from here on  */

/*  Tax brackets and marginal tax rates  */
#define tbk1 6250
#define tbk2 40200
#define tbk3 68400
#define tbk4 93950
#define tbk5 148250
#define tbk6 284700
#define mtr1 0.0765
#define mtr2 0.2616
#define mtr3 0.4119
#define mtr4 0.3499
#define mtr5 0.3834
#define mtr6 0.4360
#define mtr7 0.4761

/*  After-tax income at bracket points:  */
/*  ati(j) = ati(j-1) + [1-**MTRrecrec(j)]*[tbk(j)-tbk(j-1)]  */
#define ati1 5771.88
#define ati2 30840.56
#define ati3 47424.98
#define ati4 64035.03
#define ati5 97516.41
#define ati6 174474.21

/*  AIME and PIA  */
#define AIMEage 60
#define startwork 25
#define AIMEyrs 35      /*  AIMEage - startwork  */
#define disabincometest 3600  

/*  HRS Coefficients for spousal income equation  */
 
#define spicns   476640.6
#define spiage1 -24223.72
#define spiage2  460.255 
#define spiage3 -3.785257
#define spiage4  0.011295
#define spiage5  0.0  
#define spibadh  1030.71  
#define spibadhage -14.60367
#define cpi8798  1.0  

/*  PSID Coefficients for spousal income equation  */ 
/*
#define spicns  -35856.04
#define spiage1  3579.82
#define spiage2 -137.8337 
#define spiage3  2.791257
#define spiage4  -.0273952  
#define spiage5  0.0001012  
#define spibadh  0.0  
#define spibadhage 0.0
#define cpi8798  1.354  
*/
/*----------------------------------GLOBALS----------------------------------*/
/*  Used in interpolation and quadrature  */
    double Amovein1, Amovein2, AIMEmovein1, AIMEmovein2, _Aint, _AIMEint;
    int Anint, AIMEnint, _Anum, _Wnum, AIMEnum, hcnum, ndnumHC, HITnum,
        PrefTypeNum, HasEntryCost;
/*  Used in the utility function and bequest functions  */
    double conspref, _nu, _gamma, leispref, leisprefbad, fixcost, L, 
           bequest, Onemnu, Onemgamma, beqk, EntryPenalty, keepjob;
/*  Used in the budget constraint  */
    double earnlev, taxfrac, ror, cfloor, _Afloor, AIMEaccrue, agep,   
           AIMEmax, AIMEbk1, AIMEbk2, mpr1, mpr2, mpr3, PIA1, PIA2, 
           penbalfrac, nppenbalfrac, penbenfrac, eretadj, cumeretadj,  
           bigcred, litcred, nospinc, AIMElowfrac;
    int nret, etstage, penbensstart, manretage, mappage, eretage, medicage,
        age, apply, tiedwage, cfloorindx;
/*  Used in computing the value function  */
    double _beta, survprobpen;
/*  Miscellaneous  */
    double _Amax, fracAR1, fracwn, rhohc, hminsim, aggwagegrowth;

/*---------------------------------------------------------------------------*/
/*---------------------------------------------------------------------------*/
/*  MAIN:  entry point for the main program  */

int main(int argc, char *argv[]) 
   {
    unsigned int recsize, gotsims;
    int i, gotderef, rank, size, iSimsmin, iSimsmax, simspermachine, numsims;
    double spm2;
    double *tempvec;
    char fullpath[70];

    Matrix interpparmPtr, gridptnumPtr, prefparmPtr, ageparmPtr, miscparmPtr, 
           sscredsPtr, AstatePtr, AIMstatePtr, AIMEbkPtr, mprPtr, PIAbkPtr, 
           penaccPtr, penaccprobPtr, numsimsPtr, agePtr, survprobPtr, 
           HITtransprobwPtr, HITtransprobnwPtr, mnlnhcwPtr, mnlnhcnwPtr, 
           stderrlnhcwPtr, stderrlnhcnwPtr, apprulePtr, maxconarulePtr, 
           lazconarulePtr, maxhouarulePtr, partarulePtr, maxconnrulePtr, 
           lazconnrulePtr, maxhounrulePtr, partnrulePtr, PrefTypesimPtr, 
           TenuresimPtr, asimPtr, AIMEsimPtr, pensimPtr, appsimPtr, HITsimPtr, 
           HITinnovPtr, WsimPtr, WtypesimPtr, WfracsimPtr, healsimPtr, 
           hctypesimPtr, hcfracsimPtr, loghealthcostsPtr, healthcostsPtr, 
           csimPtr, hsimPtr, partsimPtr, MTRsimPtr, transfersPtr, holdasimPtr, 
           holdAIMEsimPtr, holdpensimPtr, holdappsimPtr, holdHITsimPtr,
           penaccruePtr;

    char rootdir[] = "/home/jjones/ssretire_final/iofiles/";  /*  Unix or Cluster  */
    if (useMPI==0)  /*  PC  */
       {
        strcpy(rootdir,"d:\\ssretire\\iofiles\\");
       }

    rank = 1;
    size = 1;

    if (useMPI==2)
       {
        MPI_Init(&argc, &argv);
        MPI_Comm_size(MPI_COMM_WORLD, &size);
        MPI_Comm_rank(MPI_COMM_WORLD, &rank);
       }

/*  Read in parameter vectors from the *.fmt files  */
/*  Start with globals  */

    interpparmPtr = gau5read(strcat(strcpy(fullpath,rootdir),"interpol.fmt"));
    gridptnumPtr  = gau5read(strcat(strcpy(fullpath,rootdir),"gridpnum.fmt"));
    prefparmPtr   = gau5read(strcat(strcpy(fullpath,rootdir),"prefparm.fmt"));
    ageparmPtr    = gau5read(strcat(strcpy(fullpath,rootdir),"ageparm.fmt"));
    miscparmPtr   = gau5read(strcat(strcpy(fullpath,rootdir),"miscparm.fmt"));
    numsimsPtr    = gau5read(strcat(strcpy(fullpath,rootdir),"_nn.fmt"));
    survprobPtr   = gau5read(strcat(strcpy(fullpath,rootdir),"survprob.fmt"));
    agePtr        = gau5read(strcat(strcpy(fullpath,rootdir),"simage.fmt"));
    AIMEbkPtr     = gau5read(strcat(strcpy(fullpath,rootdir),"AIMEbend.fmt"));
    mprPtr        = gau5read(strcat(strcpy(fullpath,rootdir),"mpr.fmt"));
    PIAbkPtr      = gau5read(strcat(strcpy(fullpath,rootdir),"PIAbend.fmt"));
    sscredsPtr    = gau5read(strcat(strcpy(fullpath,rootdir),"sscreds.fmt"));
  
    AstatePtr     = gau5read(strcat(strcpy(fullpath,rootdir),"Astate.fmt"));
    AIMstatePtr   = gau5read(strcat(strcpy(fullpath,rootdir),"AIMstate.fmt"));
    HITtransprobwPtr  = gau5read(strcat(strcpy(fullpath,rootdir),"hitranw.fmt")); 
    HITtransprobnwPtr = gau5read(strcat(strcpy(fullpath,rootdir),"hitrannw.fmt")); 

    penaccPtr       = gau5read(strcat(strcpy(fullpath,rootdir),"penacc.fmt"));
    penaccprobPtr   = gau5read(strcat(strcpy(fullpath,rootdir),"paccprob.fmt"));
    mnlnhcwPtr      = gau5read(strcat(strcpy(fullpath,rootdir),"mnlnhcw.fmt")); 
    mnlnhcnwPtr     = gau5read(strcat(strcpy(fullpath,rootdir),"mnlnhcnw.fmt"));
    stderrlnhcwPtr  = gau5read(strcat(strcpy(fullpath,rootdir),"stdhcw.fmt")); 
    stderrlnhcnwPtr = gau5read(strcat(strcpy(fullpath,rootdir),"stdhcnw.fmt")); 

    apprulePtr      = gau5read(strcat(strcpy(fullpath,rootdir),"appmati.fmt")); 
    maxconarulePtr  = gau5read(strcat(strcpy(fullpath,rootdir),"maxconai.fmt")); 
    lazconarulePtr  = gau5read(strcat(strcpy(fullpath,rootdir),"lazconai.fmt")); 
    maxhouarulePtr  = gau5read(strcat(strcpy(fullpath,rootdir),"maxhouai.fmt")); 
    partarulePtr    = gau5read(strcat(strcpy(fullpath,rootdir),"partai.fmt")); 

    maxconnrulePtr  = gau5read(strcat(strcpy(fullpath,rootdir),"maxconni.fmt")); 
    lazconnrulePtr  = gau5read(strcat(strcpy(fullpath,rootdir),"lazconni.fmt")); 
    maxhounrulePtr  = gau5read(strcat(strcpy(fullpath,rootdir),"maxhouni.fmt")); 
    partnrulePtr    = gau5read(strcat(strcpy(fullpath,rootdir),"partni.fmt")); 

    PrefTypesimPtr  = gau5read(strcat(strcpy(fullpath,rootdir),"ptypesim.fmt")); 
    TenuresimPtr    = gau5read(strcat(strcpy(fullpath,rootdir),"tensimi.fmt")); 	
    asimPtr         = gau5read(strcat(strcpy(fullpath,rootdir),"asimpo.fmt")); 
    AIMEsimPtr      = gau5read(strcat(strcpy(fullpath,rootdir),"AIMsimpo.fmt")); 
    pensimPtr       = gau5read(strcat(strcpy(fullpath,rootdir),"pensimpo.fmt")); 
    appsimPtr       = gau5read(strcat(strcpy(fullpath,rootdir),"appsimi.fmt")); 
    HITsimPtr       = gau5read(strcat(strcpy(fullpath,rootdir),"hitsimpo.fmt")); 
    HITinnovPtr     = gau5read(strcat(strcpy(fullpath,rootdir),"hitinnvi.fmt")); 
    WsimPtr         = gau5read(strcat(strcpy(fullpath,rootdir),"wsimi.fmt")); 
    WtypesimPtr     = gau5read(strcat(strcpy(fullpath,rootdir),"wtypsimi.fmt")); 
    WfracsimPtr     = gau5read(strcat(strcpy(fullpath,rootdir),"wfrcsimi.fmt")); 
    healsimPtr      = gau5read(strcat(strcpy(fullpath,rootdir),"healsimi.fmt")); 
    hctypesimPtr    = gau5read(strcat(strcpy(fullpath,rootdir),"hctypei.fmt")); 
    hcfracsimPtr    = gau5read(strcat(strcpy(fullpath,rootdir),"hcfraci.fmt")); 
    loghealthcostsPtr = gau5read(strcat(strcpy(fullpath,rootdir),"lnhcsimi.fmt")); 

/*  Dereference a bunch of C++ globals  */
    gotderef = globderef(interpparmPtr.data, gridptnumPtr.data, ageparmPtr.data, 
                         miscparmPtr.data, survprobPtr.data, agePtr.data, 
                         AIMEbkPtr.data, mprPtr.data, PIAbkPtr.data);

/*  Now set up output vectors  */
	
    recsize = (int) *numsimsPtr.data;

    healthcostsPtr = zeromat(recsize);
    csimPtr        = zeromat(recsize);
    hsimPtr        = zeromat(recsize);
    partsimPtr     = zeromat(recsize);
    MTRsimPtr      = zeromat(recsize);
    transfersPtr   = zeromat(recsize);
    penaccruePtr   = zeromat(recsize);
    holdasimPtr    = zeromat(recsize);
    holdAIMEsimPtr = zeromat(recsize);
    holdpensimPtr  = zeromat(recsize);
    holdappsimPtr  = zeromat(recsize);
    holdHITsimPtr  = zeromat(recsize);

    tempvec  = (double *)calloc(recsize,sizeof(double));

/*  Now we get the decision rules, which will be stored  */
/*  vector parts of the maxtrix structures  */

    numsims = (int) floor(rounder+*numsimsPtr.data);

    if (useMPI==2)
       {
        spm2 = ((double) numsims) / ((double) size);
        simspermachine = (int) ceil(spm2);
        iSimsmin = rank*simspermachine;
        iSimsmax = (rank+1)*simspermachine;
        if (iSimsmax > numsims) iSimsmax = numsims;
    /*  If we sum across machines, we need zero entries for simulations not handled  */
    /*  by this machine.  Hence, we make only iSimsmin-iSimsmax non-zero.  */
        for(i = iSimsmin; i < iSimsmax; ++i)
           {
            holdasimPtr.data[i]    = asimPtr.data[i];
            holdAIMEsimPtr.data[i] = AIMEsimPtr.data[i];
            holdpensimPtr.data[i]  = pensimPtr.data[i];
            holdappsimPtr.data[i]  = appsimPtr.data[i];
            holdHITsimPtr.data[i]  = HITsimPtr.data[i];
           }
       }	
    else
       {
        iSimsmin=0;
        iSimsmax=numsims;
        holdasimPtr.data    = asimPtr.data;
        holdAIMEsimPtr.data = AIMEsimPtr.data;
        holdpensimPtr.data  = pensimPtr.data;
        holdappsimPtr.data  = appsimPtr.data;
        holdHITsimPtr.data  = HITsimPtr.data;
       }

    gotsims = dosims(prefparmPtr.data, sscredsPtr.data, AstatePtr.data, 
                     AIMstatePtr.data, penaccPtr.data, penaccprobPtr.data, 
                     numsimsPtr.data, HITtransprobwPtr.data, 
                     HITtransprobnwPtr.data, mnlnhcwPtr.data, mnlnhcnwPtr.data, 
                     stderrlnhcwPtr.data, stderrlnhcnwPtr.data,
                     apprulePtr.data, maxconarulePtr.data, lazconarulePtr.data, 
                     maxhouarulePtr.data, partarulePtr.data, maxconnrulePtr.data, 
                     lazconnrulePtr.data, maxhounrulePtr.data, partnrulePtr.data, 
                     PrefTypesimPtr.data, TenuresimPtr.data, holdasimPtr.data, 
                     holdAIMEsimPtr.data, holdpensimPtr.data, holdappsimPtr.data, 
                     holdHITsimPtr.data, HITinnovPtr.data, WsimPtr.data, 
                     WtypesimPtr.data, WfracsimPtr.data, healsimPtr.data, 
                     hctypesimPtr.data, hcfracsimPtr.data, loghealthcostsPtr.data, 
                     healthcostsPtr.data, csimPtr.data, hsimPtr.data, partsimPtr.data, 
                     MTRsimPtr.data, transfersPtr.data, penaccruePtr.data, 
                     iSimsmin, iSimsmax);

    if (useMPI==2)
       {
        MPI_Reduce((void *) holdasimPtr.data, (void *) tempvec, recsize, MPI_DOUBLE,
                   MPI_SUM, size - 1, MPI_COMM_WORLD);
        switchem(&holdasimPtr.data, &tempvec);

        MPI_Reduce((void *) holdAIMEsimPtr.data, (void *) tempvec, recsize, MPI_DOUBLE,
                   MPI_SUM, size - 1, MPI_COMM_WORLD);
        switchem(&holdAIMEsimPtr.data, &tempvec);
	
        MPI_Reduce((void *) holdpensimPtr.data, (void *) tempvec, recsize, MPI_DOUBLE,
                   MPI_SUM, size - 1, MPI_COMM_WORLD);
        switchem(&holdpensimPtr.data, &tempvec);

        MPI_Reduce((void *) holdappsimPtr.data, (void *) tempvec, recsize, MPI_DOUBLE,
                   MPI_SUM, size - 1, MPI_COMM_WORLD);
        switchem(&holdappsimPtr.data, &tempvec);

        MPI_Reduce((void *) holdHITsimPtr.data, (void *) tempvec, recsize, MPI_DOUBLE,
                   MPI_SUM, size - 1, MPI_COMM_WORLD);
        switchem(&holdHITsimPtr.data, &tempvec);

        MPI_Reduce((void *) hsimPtr.data, (void *) tempvec, recsize, MPI_DOUBLE,
	           MPI_SUM, size - 1, MPI_COMM_WORLD);
        switchem(&hsimPtr.data, &tempvec);

        MPI_Reduce((void *) csimPtr.data, (void *) tempvec, recsize, MPI_DOUBLE,
                   MPI_SUM, size - 1, MPI_COMM_WORLD);
        switchem(&csimPtr.data, &tempvec);

        MPI_Reduce((void *) healthcostsPtr.data, (void *) tempvec, recsize, MPI_DOUBLE,
                   MPI_SUM, size - 1, MPI_COMM_WORLD);
        switchem(&healthcostsPtr.data, &tempvec);

        MPI_Reduce((void *) partsimPtr.data, (void *) tempvec, recsize, MPI_DOUBLE,
                   MPI_SUM, size - 1, MPI_COMM_WORLD);
        switchem(&partsimPtr.data, &tempvec);

        MPI_Reduce((void *) MTRsimPtr.data, (void *) tempvec, recsize, MPI_DOUBLE,
                   MPI_SUM, size - 1, MPI_COMM_WORLD);
        switchem(&MTRsimPtr.data, &tempvec);

        MPI_Reduce((void *) transfersPtr.data, (void *) tempvec, recsize, MPI_DOUBLE,
                   MPI_SUM, size - 1, MPI_COMM_WORLD);
        switchem(&transfersPtr.data, &tempvec);

        MPI_Reduce((void *) penaccruePtr.data, (void *) tempvec, recsize, MPI_DOUBLE,
                   MPI_SUM, size - 1, MPI_COMM_WORLD);
        switchem(&penaccruePtr.data, &tempvec);

       }

/*  Write the data to *.fmt files that GAUSS can read  */
    if ( (useMPI<2) || (rank == (size - 1)) )
       {
        gau5write(strcat(strcpy(fullpath,rootdir),"asimpo.fmt"), holdasimPtr); 
        gau5write(strcat(strcpy(fullpath,rootdir),"AIMsimpo.fmt"), holdAIMEsimPtr); 
        gau5write(strcat(strcpy(fullpath,rootdir),"pensimpo.fmt"), holdpensimPtr); 
        gau5write(strcat(strcpy(fullpath,rootdir),"appsimi.fmt"), holdappsimPtr); 
        gau5write(strcat(strcpy(fullpath,rootdir),"hitsimpo.fmt"), holdHITsimPtr); 

        gau5write(strcat(strcpy(fullpath,rootdir),"hcsimi.fmt"), healthcostsPtr); 
        gau5write(strcat(strcpy(fullpath,rootdir),"csimi.fmt"), csimPtr); 
        gau5write(strcat(strcpy(fullpath,rootdir),"hsimi.fmt"), hsimPtr); 
        gau5write(strcat(strcpy(fullpath,rootdir),"partsimi.fmt"), partsimPtr); 
        gau5write(strcat(strcpy(fullpath,rootdir),"MTRsimi.fmt"), MTRsimPtr);
        gau5write(strcat(strcpy(fullpath,rootdir),"trsimi.fmt"), transfersPtr);
        gau5write(strcat(strcpy(fullpath,rootdir),"paccsimi.fmt"), penaccruePtr);
       }

    MPI_Finalize();

    return (0);
   }

/*---------------------------------------------------------------------------*/
/*---------------------------------------------------------------------------*/
/*  ZEROMAT:  Initializes matrix structure and fills it with zeros  */

Matrix zeromat(unsigned int recsize)
   {
    Matrix mat = {1,1};
    mat.m = recsize;
    mat.n = 1;
    mat.data = (double *)calloc(recsize,sizeof(double));
    return mat;
   }

/*---------------------------------------------------------------------------*/
/*---------------------------------------------------------------------------*/
/*  SWITCHEM:  Switches two pointer addresses  */

void switchem(double **dataPtr, double **tempPtr)
   {
    double *tempPtr2;
    tempPtr2 = *dataPtr;  /*  dataPtr, tempPtr are pointers to pointers.  */
    *dataPtr = *tempPtr;  /*  Dereferencing them yields memory addresses, */
    *tempPtr = tempPtr2;  /*  rather than values.  */
   }

/*---------------------------------------------------------------------------*/
/*---------------------------------------------------------------------------*/
/*  VECTORTOMATRIX:  Converts column vector into multidimensional array.     */
/*                   This is NOT a smart program.  The user will have to     */
/*                   ensure compatibility.                                   */

double ********VectorToMatrix(double *dataVec)
   {
    int i, iType, iTenure, iHIT, iHS, iHC, iAIME, iWage;
    double ********dataMat;

    i = 0;  /*  Position in vector  */
    dataMat = (double ********)malloc(PrefTypeNum*sizeof(double *******));

    for(iType=0; iType<PrefTypeNum; iType++)
       {
        dataMat[iType] = (double *******)malloc((HasEntryCost+1)*sizeof(double ******));
        for(iTenure=0; iTenure<(HasEntryCost+1); iTenure++)
           {
            dataMat[iType][iTenure] = (double ******)malloc(HITnum*sizeof(double *****));

            for(iHIT=0; iHIT<HITnum; iHIT++)
               {
                dataMat[iType][iTenure][iHIT] = (double *****)malloc(HSnum*sizeof(double ****));
                for(iHS=0; iHS<HSnum; iHS++)
                   {
                    dataMat[iType][iTenure][iHIT][iHS] = 
                        (double ****)malloc(hcnum*sizeof(double ***));
                    for(iHC=0; iHC<hcnum; iHC++)
                       {
                        dataMat[iType][iTenure][iHIT][iHS][iHC] = 
                            (double ***)malloc(AIMEnum*sizeof(double **));
                        for(iAIME=0; iAIME<AIMEnum; iAIME++)
                           {
                            dataMat[iType][iTenure][iHIT][iHS][iHC][iAIME] = 
                                (double **)malloc(_Wnum*sizeof(double *));
                            for(iWage=0; iWage<_Wnum; iWage++) 
                               {
                            /*  Let dataMat share an _Anum-sized block of memory with datavec.  */
                            /*  This block of the state space will vary only by asset level.  */
                                i = iType*(_Anum*_Wnum*AIMEnum*hcnum*HSnum*HITnum*(HasEntryCost+1)) +
                                    iTenure*(_Anum*_Wnum*AIMEnum*hcnum*HSnum*HITnum) +
                                    iHIT*(_Anum*_Wnum*AIMEnum*hcnum*HSnum) + 
                                    iHS*(_Anum*_Wnum*AIMEnum*hcnum) +
                                    iHC*(_Anum*_Wnum*AIMEnum) + 
                                    iAIME*(_Anum*_Wnum) + iWage*_Anum;
                                dataMat[iType][iTenure][iHIT][iHS][iHC][iAIME][iWage] = &dataVec[i];
                               }  /*  End loop through wages  */
                           }  /*  End loop through AIME  */
                       }  /*  End loop through persistent health costs  */
                   }  /*  End loop through health states  */
               }  /*  End loop through health insurance types  */ 
           }  /*  End loop through job tenure statuses  */
       }  /*  End loop through preference types  */ 

    return dataMat;
   }

/*---------------------------------------------------------------------------*/
/*---------------------------------------------------------------------------*/
/*  ASS: Finds next period's assets, while checking earnings test  */
 
double ass(double AfterTaxIncome, double c, double laborincome, double assets, 
           double *reduc, double ssbenes)
   {
    double nextass;

/*  Didn't apply for benefits  */
    if (apply==0)
       {
        nextass = assets+AfterTaxIncome+ssbenes-c;
    /*  ssbenes > 0 via disability  */
        *reduc=0;
    /*  if (nextass>_Amax) nextass=_Amax;  */
        return nextass;
        }

/*  Earnings below earnings test threshhold levels  */
    if (laborincome < earnlev) 
       {
        nextass=assets+ssbenes+AfterTaxIncome-c;
        *reduc=0;
    /*  if (nextass > _Amax) nextass=_Amax;  */
        return nextass;
       }

/*  Only part of benefits are reduced  */
    if ( ((laborincome-earnlev)*taxfrac) < ssbenes ) 
       {
        double ssearntest=(laborincome-earnlev)*taxfrac; /*  Lost benefits  */
        *reduc=ssearntest/(ssbenes); /*  fraction of all benefits reduced  */
        ssbenes=ssbenes-ssearntest;
        nextass=assets+AfterTaxIncome+ssbenes-c;  /*  ssbenes are post-reduction  */
    /*  if (nextass>_Amax) nextass=_Amax;  */
        return nextass;
       }

/*  All benefits are reduced  */
    nextass=assets+AfterTaxIncome-c; 
    *reduc=1;
    ssbenes=0;
/*  if (nextass>_Amax) nextass=_Amax;  */
    return nextass;
   }

/*---------------------------------------------------------------------------*/
/*---------------------------------------------------------------------------*/
/*  COMPUTELABORINCOME  */

double computelaborincome(double wage, double h)
   {
    if (tiedwage==0) return (wage*h);
    if ((tiedwage==1) && (h>0)) return (wage*pow(h,1.415)*pow(2000,-.415));
    return 0;  /*  if tiedwage==1 and h==0  */
   }

//*---------------------------------------------------------------------------*/
/*---------------------------------------------------------------------------*/
/*  COMPUTEAFTERTAXINCOME  */

double computeAfterTaxIncome(double laborincome, double assets, double *MTR,
                             double badhealp, double penbenefits)
   {
    double spousalinc, Y;

    spousalinc = spicns + spiage1*agep + spiage2*pow(agep,2) 
                 + spiage3*pow(agep,3) + spiage4*pow(agep,4) 
                 + spiage5*pow(agep,5) + spibadh*badhealp
                 + spibadhage*badhealp*agep;
    spousalinc = spousalinc*cpi8798*(1-nospinc);

    if (spousalinc < 0) spousalinc = 0;

    Y = ror*assets + laborincome + spousalinc + penbenefits;

    if (Y<0) 
       {
        *MTR=0;
        return Y;
       }

    if (Y<tbk1) 
       {
        *MTR=mtr1;
        return ((1-mtr1)*Y);
       }

    if (Y<tbk2) 
       {
        *MTR=mtr2;
        return (ati1+(1-mtr2)*(Y-tbk1));
       }

    if (Y<tbk3)
       {
        *MTR=mtr3;
        return (ati2+(1-mtr3)*(Y-tbk2));
       }

    if (Y<tbk4) 
       {
        *MTR=mtr4;
        return (ati3+(1-mtr4)*(Y-tbk3));
       }

    if (Y<tbk5) 
       {
        *MTR=mtr5;
        return (ati4+(1-mtr5)*(Y-tbk4));
       }

    if (Y<tbk6) 
       {
        *MTR=mtr6;
        return (ati5+(1-mtr6)*(Y-tbk5));
       }

    *MTR=mtr7;
    return ati6+(1-mtr7)*(Y-tbk6);
   }

/*---------------------------------------------------------------------------*/
/*---------------------------------------------------------------------------*/
/*  INTERP:  Linearly interpolates between points on the AIME/asset/hcar grid  */

double interp(double Avar, double AIMEvar, double *AstatePtr, 
              double *AIMstatePtr, int prtype, int tentype, int hitype, 
              int badheal, int hctype, int wagetype, double ********VvarPtr)

   {
    double Amove, Alev, AIMEmove, AIMElev, Achange, AIMEchange, 
           Afrac, AIMEfrac, LoAssetLoAIME,  LoAssetHiAIME, 
           HiAssetLoAIME, HiAssetHiAIME, retr;
    int AlevShort, AIMEShort;

/*  First find the appropriate interpolation factors by seeing where the  */
/*  grid point lies in relation to the grid breaks  */
/*  With the interpolation factors, we can also find the grid point indices  */ 
/*  corresponding to aVar, AIMEvar and hcvar  */

    if (Avar<_Aint)
       {
        Amove= Amovein1;
        Alev = ((Avar-AstatePtr[0])*Amove);
       }
    else
       {
        Amove=Amovein2;
        Alev=((Avar-AstatePtr[Anint-1])*Amove)+Anint-1;
       }

    if (AIMEvar<_AIMEint)
       {
        AIMEmove=AIMEmovein1;
        AIMElev=((AIMEvar-AIMstatePtr[0])*AIMEmove);
       }
    else
       {
        AIMEmove=AIMEmovein2;
        AIMElev=((AIMEvar-AIMstatePtr[AIMEnint-1])*AIMEmove)+AIMEnint-1;
       }

/*  Find the grid points bounding avar, wvar and AIMEvar from below  */
/*  Recall C++ indices run from 0 to *num-1  */

    AlevShort = (int) floor(Alev);
    AIMEShort = (int) floor(AIMElev);

    if (AlevShort>_Anum-2) AlevShort=_Anum-2; 
    if (AIMEShort>AIMEnum-2) AIMEShort=AIMEnum-2;

    if (AlevShort<0) AlevShort=0;
    if (AIMEShort<0) AIMEShort=0;

/*  Find the weights to be placed along each dimension  */

    Achange    = Avar-AstatePtr[AlevShort];
    AIMEchange = AIMEvar-AIMstatePtr[AIMEShort];
    Afrac      = Achange*Amove;
    AIMEfrac   = AIMEchange*AIMEmove;

/*  Evaluate at the grid points bounding avar, AIMEvar and hcvar  */

    LoAssetLoAIME = VvarPtr[prtype][tentype][hitype][badheal][hctype][AIMEShort][wagetype][AlevShort];
    LoAssetHiAIME = VvarPtr[prtype][tentype][hitype][badheal][hctype][AIMEShort+1][wagetype][AlevShort];
    HiAssetLoAIME = VvarPtr[prtype][tentype][hitype][badheal][hctype][AIMEShort][wagetype][AlevShort+1];
    HiAssetHiAIME = VvarPtr[prtype][tentype][hitype][badheal][hctype][AIMEShort+1][wagetype][AlevShort+1];

/*  Weighted average  */
    retr = Afrac*AIMEfrac*         HiAssetHiAIME +
           Afrac*(1-AIMEfrac)*     HiAssetLoAIME +
           (1-Afrac)*AIMEfrac*     LoAssetHiAIME +
           (1-Afrac)*(1-AIMEfrac)* LoAssetLoAIME ;

    return retr;
   }


/*---------------------------------------------------------------------------*/
/*---------------------------------------------------------------------------*/
/*  COMPUTEAIME  */

double computeAIME(double AIME, double laborincome)
   {
    double nextAIME, LowYearIncome;

    if (laborincome>AIMEmax) laborincome=AIMEmax;
    if (age<(AIMEage+1)) AIME=AIME*(1+aggwagegrowth); /* Index year increases until age 60 */
    if (AIME>AIMEmax) AIME=AIMEmax;

    LowYearIncome = AIMElowfrac*AIME;

    if (laborincome>LowYearIncome)
       {
        nextAIME= AIME + ((laborincome-LowYearIncome)*AIMEaccrue);
        if (nextAIME>AIMEmax) nextAIME=AIMEmax;
        return nextAIME; 
       }

    nextAIME=AIME;
    return nextAIME;
   }

/*---------------------------------------------------------------------------*/
/*---------------------------------------------------------------------------*/
/*         below are the bend points for AIME and PIA, respectively          */
/*             AIME                    PIA                                   */
/*             5724                    AIME*.9 = 5151.6                      */
/*             34500                   5151.6 +.32*(AIME-5724) = 14359.9     */
/*             34500+                  14359.9 +.15*(AIME-22392)             */                        
/*         Pension benefits are imputed by:                                  */
/*             pb = p0 + p1*PIA + p2*max{0,PIA-6751.6}                       */
/*---------------------------------------------------------------------------*/
/*---------------------------------------------------------------------------*/
/*  COMPUTEPIA: convert AIME to PIA  */

double computePIA(double AIME)
   {
    double PIA;

    if (AIME<AIMEbk1) 
       {
        PIA=mpr1*AIME;
        return PIA;
       }

    if (AIME<AIMEbk2)
       {
        PIA=PIA1+mpr2*(AIME-AIMEbk1);
        return PIA;
       }

    PIA=PIA2+mpr3*(AIME-AIMEbk2);
    return PIA;
   }

/*---------------------------------------------------------------------------*/
/*---------------------------------------------------------------------------*/
/*  FINDAIME:  find the AIME associated with a given PIA  */

double findAIME(double PIA, int imposemax)
   {
    double AIMEstar;
    if (PIA<0)
       {
        return 0.0;
       }
    if (PIA<PIA1)
       {
        return (PIA/mpr1);
       }
    if (PIA<PIA2)
       {
        return (AIMEbk1+(PIA-PIA1)/mpr2);
       }

    AIMEstar=AIMEbk2+(PIA-PIA2)/mpr3;
    if ( (AIMEstar>AIMEmax) && (imposemax==1) ) AIMEstar = AIMEmax;
    return AIMEstar;
   }

/*---------------------------------------------------------------------------*/
/*---------------------------------------------------------------------------*/
/*  COMPUTEPENACCRUE  */

double computepenaccrue(double laborincome, double *penaccPtr)
   {
    double penaccruerate, penaccrue, penaccrueprob, lli;

    penaccruerate = 0.0;
    if (laborincome>0)
       {
		lli = log(laborincome);
        penaccruerate = penaccPtr[0] + penaccPtr[1]*lli;
        if (penaccruerate <-0.1) penaccruerate = -0.1; 
	    penaccrueprob = penaccPtr[2]+penaccPtr[3]*lli + penaccPtr[4]*lli*lli;
	    penaccrueprob = exp(penaccrueprob)/(1+exp(penaccrueprob));
        penaccruerate = penaccruerate*penaccrueprob;
       }
    penaccrue = penaccruerate*laborincome;

    return penaccrue;
   }

/*---------------------------------------------------------------------------*/
/*---------------------------------------------------------------------------*/
/*  GETADJ:  find adjustment factors for early/late retirement and the       */
/*           crediting back of the earnings test.                            */
/*           NOTE:  Works around a benchmark nret age of 65                  */

int getadj(double *sscredsPtr)
   {
    double cumadj2;
    double bigcred62, bigcred63, bigcred64, bigcred65, bigcred66,
           eretadj62, eretadj63, eretadj64, eretadj65, eretadj66,
           litcred65, litcred66, litcred67, litcred68, litcred69;

    bigcred62 = sscredsPtr[0];  /*  Parameters for crediting back SS benefits lost to earnings test  */
    bigcred63 = sscredsPtr[1];
    bigcred64 = sscredsPtr[2];
    bigcred65 = sscredsPtr[3];
    bigcred66 = sscredsPtr[4];
    eretadj62 = sscredsPtr[5];  /*  early application penalty factors  */
    eretadj63 = sscredsPtr[6];
    eretadj64 = sscredsPtr[7];
    eretadj65 = sscredsPtr[8];
    eretadj66 = sscredsPtr[9],
    litcred65 = sscredsPtr[10];  /*  Parameters for crediting back SS benefits lost to earnings test  */
    litcred66 = sscredsPtr[11];
    litcred67 = sscredsPtr[12];
    litcred68 = sscredsPtr[13];
    litcred69 = sscredsPtr[14];

    eretadj=1;  /*  These are all globals  */
    cumeretadj=1;
    bigcred=0;
    litcred=0;
    cumadj2 = 1;

    if (age < nret)
       {
        if (age > 66)  /* nret > 67  */
           {
            eretadj=eretadj66;
            cumeretadj=pow(eretadj66,(nret-age));
            bigcred=bigcred66;            
           }

        if (age==66)
           {
            if (nret > 67) cumadj2 = pow(eretadj66,(nret-67));
            eretadj=eretadj66;
            cumeretadj=eretadj66*cumadj2;
            bigcred=bigcred66;            
           }

        if (age==65)
           {
            if (nret==67) cumadj2 = eretadj66;
            if (nret > 67) cumadj2 = eretadj66*pow(eretadj66,(nret-67));
            eretadj=eretadj65;
            cumeretadj=eretadj65*cumadj2;
            bigcred=bigcred65;            
           }

        if (age < 65)  /*  Cumulative adjustments benchmarked to nret==65  */
           {           /*  This allows for nret > 65  */
            if (nret==66)
               {
                cumadj2 = eretadj65;
               }
            if (nret==67)
               {
                cumadj2 = eretadj65*eretadj66;
               }
            if (nret > 67)
               {
                cumadj2 = eretadj65*eretadj66*pow(eretadj66,(nret-67));
               }
           }

        if (age==64)
           {
            eretadj=eretadj64;
            cumeretadj=eretadj64*cumadj2;
            bigcred=bigcred64;
           }
        if (age==63)
           {
            eretadj=eretadj63;
            cumeretadj=eretadj64*eretadj63*cumadj2;
            bigcred=bigcred63;
           }
        if (age==62)
           {
            eretadj=eretadj62;
            cumeretadj=eretadj64*eretadj63*eretadj62*cumadj2;
            bigcred=bigcred62;
           }
        if (age<62)  /*  Allow for policy experiments  */
           {
            eretadj=eretadj62;
            cumeretadj=eretadj64*eretadj63*eretadj62;
            cumeretadj=cumeretadj*pow(eretadj,(62-age))*cumadj2;
            bigcred=bigcred62;
           }
       }

    if ((age>=nret) && (age<etstage))
       {
        if (age<65)  litcred=litcred65;  /*  Allow for policy experiments  */
        if (age==65) litcred=litcred65;
        if (age==66) litcred=litcred66;
        if (age==67) litcred=litcred67;
        if (age==68) litcred=litcred68;
        if (age==69) litcred=litcred69;
        if (age>69)  litcred=litcred69;  /*  Allow for policy experiments  */
       }

    return 1;
   }

/*---------------------------------------------------------------------------*/
/*---------------------------------------------------------------------------*/
/*  GETPREFPARMS:  Find Type-specific Preference Parameters                  */

int getprefparms(double *PrefParmPtr,int ThisPrefType)
   {
    _beta        = PrefParmPtr[ThisPrefType];
    _gamma       = PrefParmPtr[ThisPrefType+PrefTypeNum];
    _nu          = PrefParmPtr[ThisPrefType+2*PrefTypeNum];
    bequest      = PrefParmPtr[ThisPrefType+3*PrefTypeNum];
    leispref     = PrefParmPtr[ThisPrefType+4*PrefTypeNum];
    leisprefbad  = PrefParmPtr[ThisPrefType+5*PrefTypeNum];
    fixcost      = PrefParmPtr[ThisPrefType+6*PrefTypeNum];
    EntryPenalty = PrefParmPtr[ThisPrefType+7*PrefTypeNum];
    Onemgamma    = PrefParmPtr[ThisPrefType+8*PrefTypeNum];
    Onemnu       = PrefParmPtr[ThisPrefType+9*PrefTypeNum];
    conspref     = PrefParmPtr[ThisPrefType+10*PrefTypeNum];
    beqk         = PrefParmPtr[ThisPrefType+11*PrefTypeNum];
    keepjob      = PrefParmPtr[ThisPrefType+12*PrefTypeNum];

    return 1;
   }                


/*---------------------------------------------------------------------------*/
/*---------------------------------------------------------------------------*/
/*  GLOBDEREF:  Dereference a bunch of pointers to C++ globals  */

int globderef(double *interpparmPtr, double *gridptnumPtr, 
              double *ageparmPtr, double *miscparmPtr, 
              double *survprobPtr, double *agePtr, double *AIMEbkPtr, 
              double *mprPtr, double *PIAbkPtr)
   {
/*  Dereference variables used in quadrature and interpolation  */

    Amovein1     = interpparmPtr[0];
    Amovein2     = interpparmPtr[1];
    AIMEmovein1  = interpparmPtr[2];
    AIMEmovein2  = interpparmPtr[3];
    Anint        = (int) floor(rounder+interpparmPtr[4]);
    AIMEnint     = (int) floor(rounder+interpparmPtr[5]);
    _Aint        = interpparmPtr[6];
    _AIMEint     = interpparmPtr[7];

    _Anum        = (int) floor(rounder+gridptnumPtr[0]);
    _Wnum        = (int) floor(rounder+gridptnumPtr[1]);
    AIMEnum      = (int) floor(rounder+gridptnumPtr[2]);
    HITnum       = (int) floor(rounder+gridptnumPtr[3]);
    hcnum        = (int) floor(rounder+gridptnumPtr[4]);
    ndnumHC      = (int) floor(rounder+gridptnumPtr[5]);
    PrefTypeNum  = (int) floor(rounder+gridptnumPtr[6]);
    HasEntryCost = (int) floor(rounder+gridptnumPtr[7]);

/*  Dereference parameters used in the budget constraint  */

    nret         = (int) floor(rounder+ageparmPtr[0]);
    etstage      = (int) floor(rounder+ageparmPtr[1]);
    penbensstart = (int) floor(rounder+ageparmPtr[2]);
    manretage    = (int) floor(rounder+ageparmPtr[3]);
    mappage      = (int) floor(rounder+ageparmPtr[4]);
    eretage      = (int) floor(rounder+ageparmPtr[5]);
    medicage     = (int) floor(rounder+ageparmPtr[6]);

    earnlev      = miscparmPtr[0];
    taxfrac      = miscparmPtr[1];
    _Amax        = miscparmPtr[2];
    ror          = miscparmPtr[3];
    cfloor       = miscparmPtr[4];
    _Afloor      = miscparmPtr[5];
    tiedwage     = (int) floor(rounder+miscparmPtr[6]); 
    AIMEaccrue   = miscparmPtr[7];
    penbalfrac   = miscparmPtr[8];
    nppenbalfrac = miscparmPtr[9];
    penbenfrac   = miscparmPtr[10];
    fracAR1      = miscparmPtr[11];
    fracwn       = miscparmPtr[12];
    nospinc      = miscparmPtr[13];
    hminsim      = miscparmPtr[14];
    rhohc        = miscparmPtr[15];
    aggwagegrowth = miscparmPtr[17];
    AIMElowfrac  = miscparmPtr[18];

    age          = (int) floor(rounder+*agePtr);   /*  This is an integer  */
    agep         = floor(rounder+*agePtr);         /*  This is a double    */

/*  Dereference parameters used in computing the value function  */
 
    survprobpen = survprobPtr[2];

/*  Dereference vectors of SS- and pension-related parameters  */
    AIMEbk1 = AIMEbkPtr[0];
    AIMEbk2 = AIMEbkPtr[1];
    AIMEmax = AIMEbkPtr[2];
    mpr1    = mprPtr[0];
    mpr2    = mprPtr[1];
    mpr3    = mprPtr[2];
    PIA1    = PIAbkPtr[0];
    PIA2    = PIAbkPtr[1];

    return 1;
   }                

/*---------------------------------------------------------------------------*/
/*---------------------------------------------------------------------------*/
/*  GETNPHITYPE  */

int getnphitype(double hitshk, double *HITtransprobPtr)
   {
    int nphitype, i; 
    double cumulprob;

    nphitype = 0;
    cumulprob = 0;
    for (i = 0; i < HITnum; i++)
       {
        cumulprob = cumulprob + HITtransprobPtr[i];
        if (hitshk < cumulprob)
           {
            nphitype = i+1;
            break;
           }
       }
    return nphitype;
   }

/*---------------------------------------------------------------------------*/
/*---------------------------------------------------------------------------*/
/*  INTERP2:  Linearly interpolates between points on the wage and           */
/*            persistent health costs grid                                   */

double interp2(double Avar, double AIMEvar, double *AstatePtr, 
               double *AIMstatePtr, int prtype, int tentype, int hitype, 
               int badheal, int hctype, double hcfrac, int wagetype, 
               double wagefrac, double ********VvarPtr)

   {
    double loHCloWageVal, loHChiWageVal, hiHCloWageVal, hiHChiWageVal, retr;

    loHCloWageVal = interp(Avar, AIMEvar, AstatePtr, AIMstatePtr, prtype, 
                           tentype, hitype, badheal, hctype, wagetype, 
                           VvarPtr);
    loHChiWageVal = interp(Avar, AIMEvar, AstatePtr, AIMstatePtr, prtype, 
                           tentype, hitype, badheal, hctype, wagetype+1,
                           VvarPtr);

    retr = (1-hcfrac)*wagefrac*     loHChiWageVal
         + (1-hcfrac)*(1-wagefrac)* loHCloWageVal;

    if (hcnum>1)
       {
        hiHCloWageVal = interp(Avar, AIMEvar, AstatePtr, AIMstatePtr, prtype,  
                               tentype, hitype, badheal, hctype+1, wagetype, 
                               VvarPtr);
        hiHChiWageVal = interp(Avar, AIMEvar, AstatePtr, AIMstatePtr, prtype, 
                               tentype, hitype, badheal, hctype+1, wagetype+1, 
                               VvarPtr);

        retr = retr 
             + hcfrac*wagefrac*     hiHChiWageVal
             + hcfrac*(1-wagefrac)* hiHCloWageVal;
       }
    return retr;
   }

/*---------------------------------------------------------------------------*/
/*---------------------------------------------------------------------------*/
/*  SIMAPERSON:  Simulate one person's decisions for a given year  */

int simaperson( 
    double *HITtransprobwPtr, double *HITtransprobnwPtr, double *mnlnhcwPtr, 
    double *mnlnhcnwPtr, double *stderrlnhcwPtr, double *stderrlnhcnwPtr,  
    double *penaccPtr, double *penaccprobPtr, int age, double PrefTypei, 
    double HasTenurei,  double badheali, double hctypei, double hcfraci, 
    double wagetypei, double wagefraci, double wage, double loghealthcosts, 
    double HITinnov, int personnum, double *AstatePtr, double *AIMstatePtr, 
    double ********apprulePtr, double ********maxconarulePtr, 
    double ********lazconarulePtr, double ********maxhouarulePtr, 
    double ********partarulePtr, double ********maxconnrulePtr, 
    double ********lazconnrulePtr, double ********maxhounrulePtr, 
    double ********partnrulePtr, double *assetsPtr, double *AIMEPtr, 
    double *hitypePtr, double *penbalPtr, double *applyPtr, double *crec, 
    double *hrec, double *partrec, double *MTRrec, double *transfersrec, 
    double *healthcostsrec, double *penaccruerec, double *PrefParmPtr)

   {
    int PrefType, HasTenure, hctype, badheal, wagetype, hitype, HITshift, 
        imposemax, gotPrefParms;
    double assets, AIME, penbal, mnlnhc, stderrlnhc, reduc, laborincome, PIA,
           ssbenes, penbenefits, AfterTaxIncome, nextPIA, penaccrue, cashonhand;
    double penaccfs[5];
    double *HITtransprobPtr, *nextAIME, *nextperiodassets, *nextpenbal,
           *nphitype, *thisapply;
	
/*  Dereference pointers  */

    assets    = *assetsPtr;  /*  These are the current state variables  */
    AIME      = *AIMEPtr;
    PrefType  = (int) floor(rounder+PrefTypei)-1;   /*  Switch from GAUSS indexing to C++  */
    gotPrefParms = getprefparms(PrefParmPtr,PrefType);  /*  Global Preference Parameters   */
    HasTenure = (int) floor(rounder+HasTenurei);    /*  HasTenure is a 0-1 Indicator       */
    if (HasEntryCost==0)  HasTenure = 0;            /*  No Tenure/non-Tenure Distinction   */
    badheal   = (int) floor(rounder+badheali);
    wagetype  = (int) floor(rounder+wagetypei)-1;   /*  Switch from GAUSS indexing to C++  */
    if (_Wnum==1)  /*  No wage uncertainty  */
       {
        wagetype  = 0;
        wagefraci = 0.0;
       }
    hctype = (int) floor(rounder+hctypei)-1;  /*  Switch from GAUSS indexing to C++  */
    if (hcnum==1)  /* No persistent health cost uncertainty  */
       {
        hctype  = 0;
        hcfraci = 0.0;
       }
    hitype      = (int) floor(rounder+*hitypePtr)-1;  /*  Switch from GAUSS indexing to C++  */
    penaccfs[0] = penaccPtr[hitype];
    penaccfs[1] = penaccPtr[HITnum];
    penaccfs[2] = penaccprobPtr[hitype];
    penaccfs[3] = penaccprobPtr[HITnum+hitype];
    penaccfs[4] = penaccprobPtr[2*HITnum];

    penbal      = *penbalPtr;
    apply       = (int) floor(rounder+*applyPtr);     /*  This is a global  */

    nextAIME    = AIMEPtr;  /*  These are future state variables  */
    nextpenbal  = penbalPtr;
    nphitype    = hitypePtr;
    thisapply   = applyPtr;  
    nextperiodassets = assetsPtr;

    *transfersrec = 0.0;
    *thisapply    = floor(rounder+*thisapply);
    imposemax     = 1;
    reduc         = 0;

/*  Apply and thisapply currently are set to prior period's value                  */
/*  If the worker can apply and has not yet done so, check to see if she does now  */

    if (age>=eretage)
       {
        if (apply==0)
           {
            *thisapply=interp2(assets, AIME, AstatePtr, AIMstatePtr, 
                               PrefType, HasTenure, hitype, badheal, hctype, 
                               hcfraci, wagetype, wagefraci, apprulePtr);

            if (*thisapply>.5) 
               {
                *thisapply=1;
                apply=1;
               }
            else *thisapply=0; 
           }	
       }  /*  End application check  */
	
/*  Determine whether person is participating, and set hours and consumption  */
/*  Note that hours are actual hours, not hours conditional on working        */

/*  Get index for health care costs.  Distinguish between good/bad health.    */
    HITshift = hitype+badheal*HITnum;  

    if (apply==0)
       {
        *partrec=interp2(assets, AIME, AstatePtr, AIMstatePtr,  
                         PrefType, HasTenure, hitype, badheal, hctype, 
                         hcfraci, wagetype, wagefraci, partnrulePtr);

    /*  Also distinguish between medicare/non-medicare health costs  */
        if (age >= medicage) HITshift = HITshift+2*HITnum;

        if (*partrec<.5)
           {
            *partrec=0; 
            *hrec=0;
            *crec=interp2(assets, AIME, AstatePtr, AIMstatePtr, 
                          PrefType, HasTenure, hitype, badheal, hctype, 
                          hcfraci, wagetype, wagefraci, lazconnrulePtr);

            mnlnhc = mnlnhcnwPtr[HITshift];
            stderrlnhc = stderrlnhcnwPtr[HITshift];
            HITtransprobPtr = &HITtransprobnwPtr[hitype*HITnum];
           }
        else
           {
            *partrec=1; 
            *crec=interp2(assets, AIME, AstatePtr, AIMstatePtr, 
                          PrefType, HasTenure, hitype, badheal, hctype, 
                          hcfraci, wagetype, wagefraci, maxconnrulePtr);
            *hrec=interp2(assets, AIME, AstatePtr, AIMstatePtr, 
                          PrefType, HasTenure, hitype, badheal, hctype, 
                          hcfraci, wagetype, wagefraci, maxhounrulePtr);
            mnlnhc = mnlnhcwPtr[HITshift];
            stderrlnhc = stderrlnhcwPtr[HITshift];
            HITtransprobPtr = &HITtransprobwPtr[hitype*HITnum];
           }
       }

    else  /*  apply == 1  */
       {
        *partrec=interp2(assets, AIME, AstatePtr, AIMstatePtr, 
                         PrefType, HasTenure, hitype, badheal, hctype, 
                         hcfraci, wagetype, wagefraci, partarulePtr);

        if (*partrec<.5)
           {
            *partrec=0; 
            *hrec=0;
            *crec=interp2(assets, AIME, AstatePtr, AIMstatePtr, 
                          PrefType, HasTenure, hitype, badheal, hctype, 
                          hcfraci, wagetype, wagefraci, lazconarulePtr);
            mnlnhc = mnlnhcnwPtr[HITshift];
            stderrlnhc = stderrlnhcnwPtr[HITshift];
            HITtransprobPtr = &HITtransprobnwPtr[hitype*HITnum];
           }
        else
           {
            *partrec=1; 
            *crec=interp2(assets, AIME, AstatePtr, AIMstatePtr, 
                          PrefType, HasTenure, hitype, badheal, hctype, 
                          hcfraci, wagetype, wagefraci, maxconarulePtr);
            *hrec=interp2(assets, AIME, AstatePtr, AIMstatePtr, 
                          PrefType, HasTenure, hitype, badheal, hctype, 
                          hcfraci, wagetype, wagefraci, maxhouarulePtr);
            mnlnhc = mnlnhcwPtr[HITshift];
            stderrlnhc = stderrlnhcwPtr[HITshift];
            HITtransprobPtr = &HITtransprobwPtr[hitype*HITnum];
           }
       }

/*  Find end-of-period health costs and next period's health insurance status.  */
    *healthcostsrec = exp(mnlnhc + loghealthcosts*stderrlnhc);
    *nphitype = getnphitype(HITinnov, HITtransprobPtr);

/*  Compute income and future assets and AIME  */

    if (*hrec > leispref - fixcost - badheali*leisprefbad)
        *hrec = leispref - fixcost - badheali*leisprefbad;

    laborincome = computelaborincome(wage,*hrec);
    PIA = computePIA(AIME);
    ssbenes = apply*PIA;
    penbenefits = penbal*penbenfrac;
    AfterTaxIncome = computeAfterTaxIncome(laborincome, assets, MTRrec,  
                                           badheali, penbenefits);
    *nextAIME=computeAIME(AIME, laborincome);

/*  Use nextPIA to adjust nextAIME to reflect early retirement penalty or  */
/*  crediting-back of benefits lost via earnings test.                     */
    nextPIA = computePIA(*nextAIME);

/*  Adjust current benefits and next period's AIME to reflect early retirement  */
    if ((apply==1) && (age<nret))
       {
        nextPIA = eretadj*nextPIA;
        *nextAIME = findAIME(nextPIA,imposemax);
        ssbenes = ssbenes*cumeretadj;
       }

/*  There is also a benefit to deferring application for nret<=age<mappage  */
    if ((apply==0) && (age<mappage) && (age>=nret))
      {
       nextPIA = (1+litcred)*nextPIA;
       *nextAIME = findAIME(nextPIA,imposemax);
      }

    *nextperiodassets=ass(AfterTaxIncome, *crec, laborincome, assets, &reduc, ssbenes); 
    penaccrue=computepenaccrue(laborincome,penaccfs);
    *penaccruerec = -1;
    if (laborincome>0) *penaccruerec = penaccrue/laborincome;
    *nextpenbal = (1+ror)*penbal + penaccrue - penbenefits;
    *nextpenbal = *nextpenbal/survprobpen;

/*  Check consumption floor.  transfers > 0 indicates floor binds  */
    cashonhand = *nextperiodassets  + *crec;
    if (cashonhand < cfloor)
       {
        *transfersrec = cfloor - cashonhand;
        *crec = cfloor;
        *nextperiodassets = 0.0;
       }

/*  Use reduc to increase AIME of people that lost benefits by earnings test  */
    if ((apply==1) && (age<etstage))
       {
        if (age<nret) nextPIA = (1+bigcred*reduc)*nextPIA;
        else nextPIA = (1+litcred*reduc)*nextPIA;
        *nextAIME = findAIME(nextPIA,imposemax);
       }

/*  Now deduct next period's health costs  */
    *nextperiodassets = *nextperiodassets - *healthcostsrec;

/*  Treat folks with h < hminsim as if they were not working  */
/*  Replicates coding of raw data                             */

    if (*hrec<hminsim)
       {
        *partrec=0; *hrec=0;
       }

    return 1;
   }
   
/*---------------------------------------------------------------------------*/
/*---------------------------------------------------------------------------*/
/*  DOSIMS:  Entry point for the main program--Simulations  */

int dosims(
    double *prefparmPtr, double *sscredsPtr, double *AstatePtr, 
    double *AIMstatePtr, double *penaccPtr, double *penaccprobPtr, 
    double *numsimsPtr, double *HITtransprobwPtr, double *HITtransprobnwPtr, 
    double *mnlnhcwPtr, double *mnlnhcnwPtr, double *stderrlnhcwPtr, 
    double *stderrlnhcnwPtr, double *apprulePtr, double *maxconarulePtr, 
    double *lazconarulePtr, double *maxhouarulePtr, double *partarulePtr, 
    double *maxconnrulePtr, double *lazconnrulePtr, double *maxhounrulePtr, 
    double *partnrulePtr, double *PrefTypesimPtr, double *TenuresimPtr, 
    double *asimPtr, double *AIMEsimPtr, double *pensimPtr, double *appsimPtr, 
    double *HITsimPtr, double *HITinnovPtr, double *WsimPtr, 
    double *WtypesimPtr, double *WfracsimPtr, double *healsimPtr, 
    double *hctypesimPtr, double *hcfracsimPtr, double *loghealthcostsPtr, 
    double *healthcostsPtr, double *csimPtr, double *hsimPtr, 
    double *partsimPtr, double *MTRsimPtr, double *transfersPtr, 
    double *penaccruePtr, int iSimsmin, int iSimsmax)

   {
    double ********appruleMat, ********maxconaruleMat, ********lazconaruleMat, 
           ********maxhouaruleMat, ********partaruleMat, ********maxconnruleMat,
           ********lazconnruleMat, ********maxhounruleMat, ********partnruleMat; 
    int i, gotone;

    int numsims = (int) floor(rounder+*numsimsPtr);

/*  Obtain the globals eretadj, cumretadj, bigcred, litcred  */
/*  These are age-specific adjustment factors for early or late SS application  */
    int gotadjfactors = getadj(sscredsPtr);

    appruleMat     = VectorToMatrix(apprulePtr);
    maxconaruleMat = VectorToMatrix(maxconarulePtr);
	lazconaruleMat = VectorToMatrix(lazconarulePtr);
    maxhouaruleMat = VectorToMatrix(maxhouarulePtr);
    partaruleMat   = VectorToMatrix(partarulePtr);
    maxconnruleMat = VectorToMatrix(maxconnrulePtr);
	lazconnruleMat = VectorToMatrix(lazconnrulePtr);
    maxhounruleMat = VectorToMatrix(maxhounrulePtr);
    partnruleMat   = VectorToMatrix(partnrulePtr);


    for(i = iSimsmin; i < iSimsmax; ++i)
       {
        gotone = simaperson(HITtransprobwPtr, HITtransprobnwPtr, mnlnhcwPtr, 
                            mnlnhcnwPtr, stderrlnhcwPtr, stderrlnhcnwPtr, 
                            penaccPtr, penaccprobPtr, age, PrefTypesimPtr[i], 
                            TenuresimPtr[i], healsimPtr[i], hctypesimPtr[i], 
                            hcfracsimPtr[i], WtypesimPtr[i], WfracsimPtr[i], 
                            WsimPtr[i], loghealthcostsPtr[i], HITinnovPtr[i], 
                            i, AstatePtr, AIMstatePtr, appruleMat, 
                            maxconaruleMat, lazconaruleMat, maxhouaruleMat,
                            partaruleMat, maxconnruleMat, lazconnruleMat, 
                            maxhounruleMat, partnruleMat, &asimPtr[i], 
                            &AIMEsimPtr[i], &HITsimPtr[i], &pensimPtr[i], 
                            &appsimPtr[i], &csimPtr[i], &hsimPtr[i], 
                            &partsimPtr[i], &MTRsimPtr[i], &transfersPtr[i], 
                            &healthcostsPtr[i], &penaccruePtr[i], 
                            prefparmPtr);       
	}

    return 1;
   }

/*---------------------------------------------------------------------------*/
/*---------------------------------------------------------------------------*/
/*  Below is C++/GAUSS I/O code written by Ken Housinger                     */
/*---------------------------------------------------------------------------*/
/*---------------------------------------------------------------------------*/
/*  This function takes a char* and either reverses bits within bytes or bytes */
/*  themselves or both.  It returns a pointer to the original buffer which is */
/*  altered. */

unsigned char * gread(unsigned char *inbuf, int bytes, int byte_rev, int bit_rev) 
   {
    unsigned char *tempbuf;
    unsigned char tempbyte, tempbit;
    int i, j;

    tempbuf = (unsigned char *) malloc(bytes);
    for (i = 0; i < bytes; i++) 
       {
        if (byte_rev) *(tempbuf + i) = *(inbuf + bytes - i - 1);
        else *(tempbuf + i) = *(inbuf + i);

        if (bit_rev) 
           {
            tempbyte = 0;
            for (j = 0; j < CHAR_BIT; j++) 
               {
                tempbit = *(tempbuf + i) >> (CHAR_BIT - j - 1);
                tempbit = tempbit << (CHAR_BIT - 1);
                tempbit = tempbit >> (CHAR_BIT - j - 1);
                tempbyte = tempbyte | tempbit;
               }
            *(tempbuf + i) = tempbyte;
           }
       }

    for (i = 0; i < bytes; i++)
        *(inbuf + i) = *(tempbuf + i);
    free(tempbuf);

    return(inbuf);
   }

/*---------------------------------------------------------------------------*/
/*---------------------------------------------------------------------------*/
/*  This function reads a Gauss v5.0 fmt file into a Matrix structure and  */
/*  returns the structure. */

Matrix gau5read(char *fmt) 
   {
/*  Initialize the matrix to be 1x1 and the byte/bit order to 0.  */
    Matrix mat = {1, 1}; 
    unsigned int i;
    int type, byte, bit;
    unsigned char *cread;
    int bit_rev = 0, byte_rev = 0;
    FILE *fd;

    if (sizeof(int) != 4 || sizeof(double) != 8 || CHAR_BIT != 8) 
       {
        printf("Incompatable machine architecture.\n");
        return (mat);
       }

/*  Allocate enough space to store the header.  */
    cread = (unsigned char *) malloc(BASIC_HEADER_LEN); 
/*  Open *fmt for reading only.  */
    fd = fopen(fmt, "rb"); 
  
/*  Read the basic header (128 bytes) all at once.  */

    fread((void *) cread, 1, BASIC_HEADER_LEN, fd);
    byte = (int) *(cread + (BYTE_POS * sizeof(int)));  /* (0=Backward)  */
    bit = (int) *(cread + (BIT_POS * sizeof(int)));    /* (0=Backward)  */

/*  To get some system independence, we detect whether we have to reverse  */
/*  the bytes or bits or both.  If x and x_SYSTEM match, no reverse is     */
/*  necessary.  */

    if ((bit || BIT_SYSTEM) && !(bit && BIT_SYSTEM)) bit_rev=1;
    if ((byte || BYTE_SYSTEM) && !(byte && BYTE_SYSTEM)) byte_rev=1;

    type = *( (int *) gread((cread + (TYPE_POS * sizeof(int))), sizeof(int), 
                             byte_rev, bit_rev) );

/*  If the fmt file type is not a scalar, there are another two  */
/*  ints of header giving the values of m and n.  If a matrix, also reset n.  */

    if (type > SCALAR) 
       { 
        fread((void *) cread, 1, sizeof(int), fd);
        mat.m = *((unsigned int *) gread(cread, sizeof(int), byte_rev, bit_rev));
        fread((void *) cread, 1, sizeof(int), fd);
        if (type == MATRIX)
          mat.n = *((unsigned int *) gread(cread, sizeof(int), byte_rev, bit_rev));
      } 

/*  Allocate memory for the matrix.  The amount needed is m * n * sizeof(double). */
/*  Next, read in the data all at once.  Then use gread to reverse bytes/bits if  */
/*  necessary.  */

    free(cread);

    mat.data = (double *) malloc(mat.m * mat.n * sizeof(double));
    fread((void *) mat.data, sizeof(double), mat.m * mat.n, fd);
    if (byte_rev || bit_rev)
      for(i = 0; i < mat.m * mat.n; i++)
        gread((unsigned char *) mat.data + (i * sizeof(double)), sizeof(double), 
               byte_rev, bit_rev);

    fclose(fd);

    return (mat);
   }

/*---------------------------------------------------------------------------*/
/*---------------------------------------------------------------------------*/
/*  This function writes a Gauss v5.0 fmt file from a Matrix structure.  */

void gau5write(char *fmt, Matrix mat) 
   {
/*  This ugly mess is the basic header. */

    unsigned int header[(BASIC_HEADER_LEN / sizeof(int)) + 2] = 
        {0xffffffff, 0, 0xffffffff, 0, 0xffffffff, 0, 0,
         0xabcdef01,1, 0, 1, 1008, sizeof(double), 0, 1,
         SCALAR, 1, 0, BASIC_HEADER_LEN};

    FILE *fd;

    if (sizeof(int) != 4 || sizeof(double) != 8 || CHAR_BIT != 8) 
       {
        printf("Incompatable machine architecture.\n");
        return;
       }

/*  If forward byte, make 6th int 0xffffffff.  */
/*  If forward bit, make 7th int 0xffffffff.   */

    if (BYTE_SYSTEM) header[BYTE_POS] = 0xffffffff;
    if (BIT_SYSTEM) header[BIT_POS] = 0xffffffff;

/*  If not a scalar, increase the 16th int by 1 and the 19th int (header  */
/*  length) by 8 (2 * sizeof(int)).  Also, set m in int 33.  */

    if (!(mat.m * mat.n == 1)) 
       {
        header[TYPE_POS] += 1;
        header[HEADER_LEN_POS] += (2 * sizeof(int));
        header[M_POS] = mat.m;

    /*  If not a vector (and not a scalar), increase the 16th int by 1 again   */
    /*  and set m in int 34.  */
        if (!(mat.n == 1)) 
           {
            header[TYPE_POS] += 1;
            header[N_POS] = mat.n;
           }
       }
  /*
  **Open fmt for writing and create if it does not exist.  If you create it,
  **make it a regular file with permissions 0640.  See comment in gau5read
  **for detail on how read (and similarly, write) work.
  **
  **Order: Write the basic header
  **       If not a scalar, write the other 2 ints of header 
  **       Write the m * n elements of data
  */

    fd = fopen(fmt, "wb"); 
    if ((mat.m * mat.n == 1))
      fwrite((void *) header, 1, BASIC_HEADER_LEN, fd);
    else
      fwrite((void *) header, 1, BASIC_HEADER_LEN + (2 * sizeof(int)), fd);
    fwrite((void *) mat.data, sizeof(double), mat.m * mat.n, fd);
    fclose(fd);
   }  

/*---------------------------------------------------------------------------*/
/*---------------------------------------------------------------------------*/
