#include <string.h>
#include <limits.h>
#include <stdlib.h>

/*---------------------------------------------------------------------------*/
/*---------------------------------------------------------------------------*/
/*  The next three procedures are C/GAUSS I/O routines by Ken Housinger      */

#define BASIC_HEADER_LEN 128 /*Length of a scalar header*/
#define BYTE_POS 5           /*Int offset in header of byte order*/
#define BIT_POS 6            /*Int offset in header of bit order*/
#define TYPE_POS 15          /*Int offset in header of matrix type*/
#define HEADER_LEN_POS 18    /*Int offset in header of header length*/
#define M_POS 32             /*Int offset in header of value for m*/
#define N_POS 33             /*Int offset in header of value for n*/
#define SCALAR 0             /*Value of type that indicates a scalar*/
#define MATRIX 2             /*Value of type that indicates a matrix*/
#define BIT_SYSTEM 0         /*Value of bit order (0=backwards/i386)*/
#define BYTE_SYSTEM 0        /*Value of byte order (0=backwards/i386)*/

/*
**This is my Matrix opaque datatype.  The gauss fmt format has implied 
**dimensionality, but is just a list of doubles.  I could rewrite the
**functions to use pointers-to-pointers to improve the interface with
**Eric's functions, but that would add to the complxity of my code.  I
**suggest writing an intermediate function that turns *data into **data
**to cope with 2-dimensional arrays.
*/ 
typedef struct {
  unsigned int m;
  unsigned int n;
  double *data;
} Matrix;
/*  
**Currently, these functions require full filenames (you must provide the  
**the ".fmt").                                                           
*/
unsigned char * gread(unsigned char *inbuf, int bytes, int byte_reverse, int bit_reverse);

Matrix gau5read(char *fmt);

void gau5write(char *fmt, Matrix mat);

/*---------------------------------------------------------------------------*/
/*---------------------------------------------------------------------------*/

Matrix zeromat(unsigned int recsize);

void switchem(double **dataPtr, double **tempPtr);

double ********VectorToMatrix(double *dataVec);

double ass(double AfterTaxIncome, double c, double laborincome, double assets, 
           double *reduc, double ssbenes);

double computelaborincome(double wage, double h);

double computeAfterTaxIncome(double laborincome, double assets, double *MTR,
                             double badhealp, double penbenefits);

double interp(double Avar, double AIMEvar, double *AstatePtr, 
              double *AIMstatePtr, int prtype, int tentype, int hitype, 
              int badheal, int hctype, int wagetype, double ********VvarPtr);

double computeAIME(double AIME, double laborincome);

double computePIA(double AIME);

double findAIME(double PIA, int imposemax);

double computepenaccrue(double laborincome, double *penaccPtr);

int getadj(double *sscredsPtr);

int getprefparms(double *PrefParmPtr,int ThisPrefType);

int globderef(double *interpparmPtr, double *gridptnumPtr, 
              double *ageparmPtr, double *miscparmPtr, 
              double *survprobPtr, double *agePtr, double *AIMEbkPtr, 
              double *mprPtr, double *PIAbkPtr);
     
int getnphitype(double hitshk, double *HITtransprobPtr);

double interp2(double Avar, double AIMEvar, double *AstatePtr, 
               double *AIMstatePtr, int prtype, int tentype, int hitype, 
               int badheal, int hctype, double hcfrac, int wagetype, 
               double wagefrac, double ********VvarPtr);

int simaperson( 
    double *HITtransprobwPtr, double *HITtransprobnwPtr, double *mnlnhcwPtr, 
    double *mnlnhcnwPtr, double *stderrlnhcwPtr, double *stderrlnhcnwPtr,  
    double *penaccPtr, double *penaccprobPtr, int age, double PrefTypei, 
    double HasTenurei,  double badheali, double hctypei, double hcfraci, 
    double wagetypei, double wagefraci, double wage, double loghealthcosts, 
    double HITinnov, int personnum, double *AstatePtr, double *AIMstatePtr, 
    double ********apprulePtr, double ********maxconarulePtr, 
    double ********lazconarulePtr, double ********maxhouarulePtr, 
    double ********partarulePtr, double ********maxconnrulePtr, 
    double ********lazconnrulePtr, double ********maxhounrulePtr, 
    double ********partnrulePtr, double *assetsPtr, double *AIMEPtr, 
    double *hitypePtr, double *penbalPtr, double *applyPtr, double *crec, 
    double *hrec, double *partrec, double *MTRrec, double *transfersrec, 
    double *healthcostsrec, double *penaccruerec, double *PrefParmPtr);

int dosims(
    double *prefparmPtr, double *sscredsPtr, double *AstatePtr, 
    double *AIMstatePtr, double *penaccPtr, double *penaccprobPtr, 
    double *numsimsPtr, double *HITtransprobwPtr, double *HITtransprobnwPtr, 
    double *mnlnhcwPtr, double *mnlnhcnwPtr, double *stderrlnhcwPtr, 
    double *stderrlnhcnwPtr, double *apprulePtr, double *maxconarulePtr, 
    double *lazconarulePtr, double *maxhouarulePtr, double *partarulePtr, 
    double *maxconnrulePtr, double *lazconnrulePtr, double *maxhounrulePtr, 
    double *partnrulePtr, double *PrefTypesimPtr, double *TenuresimPtr, 
    double *asimPtr, double *AIMEsimPtr, double *pensimPtr, double *appsimPtr, 
    double *HITsimPtr, double *HITinnovPtr, double *WsimPtr, 
    double *WtypesimPtr, double *WfracsimPtr, double *healsimPtr, 
    double *hctypesimPtr, double *hcfracsimPtr, double *loghealthcostsPtr, 
    double *healthcostsPtr, double *csimPtr, double *hsimPtr, 
    double *partsimPtr, double *MTRsimPtr, double *transfersPtr, 
    double *penaccruePtr, int iSimsmin, int iSimsmax);
