@-----------------------------------------------------------------------------@
@-----------------------------------------------------------------------------@
/*
GETAGES:  gets ages right
*/
proc(2) = getages(bornage,dieage,dat);
    local j,k;
    k = indexcat(dat[.,1],bornage);
    if scalmiss(k) == 1;
        k = 1;
    endif;
    j = indexcat(dat[.,1],dieage);
    if scalmiss(j) == 1;
        j = rows(dat);
    endif;
retp(k,j); endp;

@-----------------------------------------------------------------------------@
@-----------------------------------------------------------------------------@
/*
FIXOBS:  Adjusts data to have a common set of years
         Adds filler rows if data doesn't cover [bornage,dieage]
*/
proc(1) = fixobs(dat,k,j,lowmiss,himiss);

   local cn, rn, rn2, bage2, dage2;

   cn    = cols(dat);
   bage2 = dat[k,1];
   dage2 = dat[j,1];
   dat   = dat[k:j,.];

   if bage2 > bornage;
       rn  = bage2-bornage;
       dat = (lowmiss*ones(rn,cn))|dat;
       dat[1:rn,1] = seqa(bornage,1,rn);
   endif;
   rn2 = j-k+bage2-bornage+1;                        @ Number of rows in dat @
   if dage2 < dieage;
       rn  = dieage-dage2;
       dat = dat|(himiss*ones(rn,cn));
       dat[rn2+1:lspan,1] = seqa(dage2+1,1,rn);
   endif;

retp(dat); endp;

@-----------------------------------------------------------------------------@
@-----------------------------------------------------------------------------@

proc(14) = getprofs();

    local j, k, filename, dat2, dat3, hwage_fe, uwage_fe, survprof, 
          survdatg,  survdatb, healg, healb, bv, biasvecg, biasvecb, healbias,
          hlogwage, ulogwage, healm, survsim, bhprob, survdat, survdat2, 
          pencfs0, ltbhpr, minprob, rorshk, wlthshk2, mywages;

    format /ro 10,4;

   @---------Choose between tied wage-hours offers or exogenous wages---------@

    filename = datapath $+ "wageprof_hrs_080309.out";
    load dat2[] = ^filename; 
    dat2  = reshape(dat2, rows(dat2)/5, 5);
    {k,j} = getages(bornage,dieage,dat2);
    dat2  = fixobs(dat2,k,j,1,1);

    if tied==0;
        uwage_fe=dat2[.,2];
        hwage_fe=dat2[.,3];
    elseif tied==1; 
        uwage_fe=dat2[.,4];
        hwage_fe=dat2[.,5];
    endif;

    j=1; do until j>lspan-1;
        if scalmiss(hwage_fe[j+1])==1;
            hwage_fe[j+1]=hwage_fe[j];
        endif;
        if scalmiss(uwage_fe[j+1])==1;
            uwage_fe[j+1]=uwage_fe[j];
        endif;
    j=j+1; endo;

    "      age    hwage_fe   uwage_fe";;
    print dat2[.,1]~hwage_fe~uwage_fe;?;
    dat2 = 0;

   @-----------------------Mortality and Survival Rates-----------------------@
    filename = datapath $+ "survprof_080609.out";
    load dat3[] = ^filename;
    dat3     = reshape(dat3,rows(dat3)/4,4);
    dat3[.,1]= dat3[.,1]-1;         @ Convert to one-year-ahead probabilities @ 
    {k,j}    = getages(bornage,dieage,dat3);
    dat3     = fixobs(dat3,k,j,1,0);
    survdatg = dat3[.,2];
    survdatb = dat3[.,3];
    survdat  = dat3[.,4];                @ Unconditional survival probability @
    minprob  = 0.0001*ones(2,1);
    survdatg[lspan-1:lspan,1] = minprob;
    survdatb[lspan-1:lspan,1] = minprob;
    survdat[lspan-1:lspan,1]  = minprob;

    "      age    survdatg   survdatb     sprob" dat3[.,1]~survdatg~survdatb~survdat;?;
    dat3 = 0;

   @---------------------Health Transition probabilities----------------------@
    filename = datapath $+ "hlthprof_080609.out";
    load dat2[] = ^filename; 
    dat2     = reshape(dat2,rows(dat2)/7,7);
    dat2[.,1]= dat2[.,1]-1;         @ Convert to one-year-ahead probabilities @
    {k,j}    = getages(bornage,dieage,dat2);
    dat2     = fixobs(dat2,k,j,0,1);
    healg    = dat2[.,2];                  @ probability of health: good->bad @
    healb    = dat2[.,3];                   @ probability of health: bad->bad @
    "Probability of bad health at t+1 ";
    "      age       healb     healg" dat2[.,1]~healb~healg;?;
    
    {bhprob,survdat2,ltbhpr} = gethp(healg,healb,survdatg,survdatb);

   @-------------------Generate the health and wage shocks--------------------@
    if newmat==1;                                 @ 1 => load in new matrices @
        {healm,survsim} = initsim(healg,healb,survdatg,survdatb);
    elseif newmat==0; 
        {healm,survsim} = initsim0();
    endif;

    "Unconditional Survival probabilities";
    "        age   sprob(data)  sprob(anal)   sprob(sim)";; 
    ageseq~survdat~survdat2~survsim;?;

   "Simulated and analytical probabilities of bad health by age";
    "        age     healm(sim)  analytical Pr(bh_t-1|bh_t)";;
    ageseq~healm~bhprob~ltbhpr;?;

   @------------------Generate bias-adjusted wage profiles--------------------@
  @ load path=^shkpath bv; @                 @ Model puts updated file here @
    load path=^datapath bv = bv_base_030510;
    bv       = bv*(selecadj>0);
    biasvecb = bv[.,1];
    biasvecg = bv[.,2];
    healbias = bv[1,3];
    {hlogwage,ulogwage} = makewage(hwage_fe,uwage_fe,biasvecg,biasvecb,healbias); 

   @------------------Get observed age-specific ROR shocks--------------------@
    filename = datapath $+ "wlthshk2_090502.txt";  
    load wlthshk2[] = ^filename;  
    wlthshk2 = reshape(wlthshk2,rows(wlthshk2)/5,5);
    {k,j}    = getages(bornage,dieage,wlthshk2);
    wlthshk2 = fixobs(wlthshk2,k,j,0,0);

    if rshktype==1;
        rorshk = wlthshk2[.,5];
    elseif rshktype==2;
        rorshk = wlthshk2[.,2];
    elseif rshktype==3;
        rorshk = (wlthshk2[.,5]+wlthshk2[.,2])/2;
    else;
        rorshk = 0*wlthshk2[.,1];
    endif;
    "Rate of return shocks:";;
    ageseq~rorshk;?;

retp(survdatg, survdatb, healg, healb, hwage_fe, uwage_fe, biasvecg, biasvecb, 
     healbias, hlogwage, ulogwage, survsim, survdat, rorshk);
endp;
@-----------------------------------------------------------------------------@
@-----------------------------------------------------------------------------@
proc(6) = gethcdat();

    local filename, hcparm2, cn, rn, ntypes, hcrow, umnlnhc, hmnlnhc, hitrmat, 
          ustdhc, hstdhc, title1, title2, title3, title4, hcttrnd, mnlnhc65, 
          stdhc65, aseq2;

   @------------------Mean health costs and scaling factors-------------------@
   @   Type 1:  no health insurance                                           @
   @   Type 2:  retiree coverage                                              @
   @   Type 3:  coverage tied to working                                      @

    filename = datapath $+ "hcparm_082009.out";
    load hcparm2[12,2] = ^filename; 
    ntypes  = 5;

    hmnlnhc = zeros(lspan,2*HITnum);
    cn      = 5|1|3|5|2|4;
    hcrow   = hcparm2[cn,1];
    hmnlnhc[1:medicage-bornage,.] = ones(medicage-bornage,1)*hcrow';
    cn      = cn+ntypes;
    hcrow   = hcparm2[cn,1];
    hmnlnhc[medicage-bornage+1:lspan,.] = ones(dieage-medicage+1,1)*hcrow';
    hcttrnd = hcparm2[maxc(cn)+2,.].*(ageseq-55);
    hmnlnhc = hmnlnhc+hcttrnd[.,1];
    umnlnhc = hmnlnhc + hcparm2[maxc(cn)+1,1];

    hstdhc  = zeros(lspan,2*HITnum);
    cn      = 5|1|3|5|2|4;
    hcrow   = hcparm2[cn,2];
    hstdhc[1:medicage-bornage,.] = ones(medicage-bornage,1)*hcrow';
    cn      = cn+ntypes;
    hcrow   = hcparm2[cn,2];
    hstdhc[medicage-bornage+1:lspan,.] =  ones(dieage-medicage+1,1)*hcrow';
    hstdhc  = hstdhc + hcttrnd[.,2];
    ustdhc  = hstdhc + hcparm2[maxc(cn)+1,2];
    hmnlnhc = hmnlnhc + (1-hcvarscl)*hstdhc/2;
    umnlnhc = umnlnhc + (1-hcvarscl)*ustdhc/2;
    hstdhc  = sqrt(hcvarscl*hstdhc);
    ustdhc  = sqrt(hcvarscl*ustdhc);

    if noCOBRA==1;
        hmnlnhc[.,6] = hmnlnhc[.,4];
        umnlnhc[.,6] = umnlnhc[.,4];
        hstdhc[.,6]  = hstdhc[.,4];
        ustdhc[.,6]  = ustdhc[.,4];
    endif;

    rn       = 65-bornage+1;
    mnlnhc65 = hmnlnhc[rn-1,.]+hcttrnd[rn,1]-hcttrnd[rn-1,1];
    mnlnhc65 = mnlnhc65|hmnlnhc[rn,.];
    mnlnhc65 = mnlnhc65|(umnlnhc[rn-1,.]+hcttrnd[rn,1]-hcttrnd[rn-1,1]);
    mnlnhc65 = mnlnhc65|umnlnhc[rn,.];

    stdhc65 = hstdhc[rn-1,.]+hcttrnd[rn,2]-hcttrnd[rn-1,2];
    stdhc65 = stdhc65|hstdhc[rn,.];
    stdhc65 = stdhc65|(ustdhc[rn-1,.]+hcttrnd[rn,2]-hcttrnd[rn-1,2]);
    stdhc65 = stdhc65|ustdhc[rn,.];


    title1 = "        Age         None       Retiree  Tied-Working       None       "
             $+ "Retiree       COBRA";
    title2 = "                    ------------------------------UnHealthy" 
             $+"------------------------------";
    title3 = "                    -------------------------------Healthy" 
             $+"-------------------------------";
    title4 = "                    --------------------------------Age 65" 
             $+"-------------------------------";
    aseq2 = 65*ones(4,1);


    " Analytical Mean Health Care Costs ";
    title2; title1;;
    ageseq~exp(umnlnhc + (ustdhc.^2)/2);
    title3; title1;;
    ageseq~exp(hmnlnhc + (hstdhc.^2)/2);?;
    title4; title1;;
    aseq2~exp(mnlnhc65 + (stdhc65.^2)/2);?;

    " Analytical Std. deviation Health Care Costs ";
    title2; title1;;
    ageseq~sqrt( exp(2*umnlnhc+ustdhc.^2).*(exp(ustdhc.^2)-1) );
    title3; title1;;
    ageseq~sqrt( exp(2*hmnlnhc+hstdhc.^2).*(exp(hstdhc.^2)-1) );?;
    title4; title1;;
    aseq2~sqrt( exp(2*mnlnhc65+stdhc65.^2).*(exp(stdhc65.^2)-1) );?;

    " 99.5-th Percentile Health Care Costs ";
    title2; title1;;
    ageseq~exp(umnlnhc + 2.5758293*ustdhc);
    title3; title1;;
    ageseq~exp(hmnlnhc + 2.5758293*hstdhc);?;
    title4; title1;;
    aseq2~exp(mnlnhc65 + 2.5758293*stdhc65);?;

   @----------------Health Insurance Type Transition Matrices-----------------@
   @---------Rows give current type, columns give next period's type----------@
   @---------------Matrices differ by age and employment choice---------------@
    hitrmat = {1 0 0 1 0 0,
               0 1 0 0 1 0,
               0 0 1 1 0 0};
    "Health Insurance Transition Matrix";;hitrmat;?;
    hitrmat = ones(lspan,1).*.hitrmat;

retp(hmnlnhc, umnlnhc, hstdhc, ustdhc, hcttrnd, hitrmat); endp;

@-----------------------------------------------------------------------------@
@-----------------------------------------------------------------------------@
/*
NOHCVAR:  Constructs age profiles for logged health costs under perfect foresight
          Makes sure means of health costs themselves do not change
*/
proc(4) = nohcvar(hmnlnhc, umnlnhc, hstdhc, ustdhc);

    local totvar, rn, rn2, cn, hvarhc, uvarhc;

    rn     = hivage-bornage+1;
    rn2    = lspan-rn+1;
    totvar = fracar1i*ones(rn2,1);                  @ AR1 innovation variance @
           @ recserar(fracar1i*ones(rn2,1),fracar1i,rhohc^2)                  @ 
   @-Using just the innovation variance allows us to match conditional means.-@  
   @---Using the cumulative variance allows us to match unconditional means---@

    totvar = fracar1-totvar;
    cn     = cols(hstdhc);
    totvar = totvar*ones(1,cn);
    totvar = zeros(rn-1,cn)|totvar;

    hvarhc  = (hstdhc^2).*totvar;
    uvarhc  = (ustdhc^2).*totvar;

    hmnlnhc = hmnlnhc + (hstdhc^2)/2 - hvarhc/2;
    umnlnhc = umnlnhc + (ustdhc^2)/2 - uvarhc/2;
    hstdhc  = 0*hstdhc;
    ustdhc  = 0*ustdhc;

    " Analytical Mean Health Care Costs:  1-year averages ";
    "        Age         None       Retiree     Tied-Working    None       ";;
    "Retiree       COBRA        Average";;
    ageseq~exp(umnlnhc + (ustdhc.^2)/2);?;

    " Analytical Std. deviation Health Care Costs:  1-year averages ";
    "        Age         None       Retiree     Tied-Working    None       ";;
    "Retiree       COBRA        Average";;
    ageseq~sqrt( exp(2*umnlnhc+ustdhc.^2).*(exp(ustdhc.^2)-1) );?;

    " 99.5-th Percentile Lifetime Health Care Costs:  1-year averages ";
    "        Age         None       Retiree     Tied-Working    None       ";;
    "Retiree       COBRA        Average";;
    ageseq~exp(umnlnhc + 2.5758293*ustdhc);?;

retp(hmnlnhc, umnlnhc, hstdhc, ustdhc); endp;

@------------------------------------------------------------------------@
@------------------------------------------------------------------------@
proc(4) = samehc(hmnlnhc, umnlnhc, hstdhc, ustdhc, hitype);

    local cn, title1, title2, title3;

    cn = (hitype|(hitype+3)).*.ones(3,1);
 
    hmnlnhc = hmnlnhc[.,cn];
    umnlnhc = umnlnhc[.,cn];
    hstdhc  = hstdhc[.,cn];
    ustdhc  = ustdhc[.,cn];

    title1 = "        Age         None       Retiree  Tied-Working       None       "
             $+ "Retiree       COBRA";
    title2 = "                    ------------------------------UnHealthy" 
             $+"------------------------------";
    title3 = "                    -------------------------------Healthy" 
             $+"-------------------------------";

    " Analytical Mean Health Care Costs ";
    title2; title1;;
    ageseq~exp(umnlnhc + (ustdhc.^2)/2);
    title3; title1;;
    ageseq~exp(hmnlnhc + (hstdhc.^2)/2);?;

retp(hmnlnhc, umnlnhc, hstdhc, ustdhc); endp;

@-----------------------------------------------------------------------------@
@-----------------------------------------------------------------------------@

proc(2) = makewage(hwage_fe, uwage_fe, biasvecg, biasvecb, healbias);

    local hlogwage, ulogwage, wsim50, winnov, wtypeinv, healsim, bhsim, 
          uwmat, hwmat, uhealobs, healobs, hmwage, umwage, mwage, dscale, 
          devsim, logwage, wsim, wtypesim, wfracsim, i, wsimi, wtypsimi, 
          wfrcsimi, hwage_ba, uwage_ba, devmax, devmin, db;

    format 10,4;   

   @-----Construct log wage profiles that are adjusted for selection bias-----@
   @-------Note that bias adjustments are for growth rates (log diffs),-------@
   @------------rather than log levels (hence the cumulative sums)------------@
   
    hlogwage = ln(hwage_fe)+cumsumc(0|biasvecg); 
    ulogwage = ln(uwage_fe)+cumsumc(0|biasvecb);  
    ulogwage = ulogwage - healbias;
     
    "Total bias adjustments (before any fitting to initial distribution)";    
    "      age      healthy  unhealthy";;        
    ageseq~(hlogwage-ln(hwage_fe))~(ulogwage-ln(uwage_fe));?; 
     
   @------Rescale to initial distribution, adjust for health selection--------@  
   @--We assume that initial distribution does NOT capture health selection---@ 

    load path = ^shkpath wsim50, winnov, wtypeinv, healsim;  
    bhsim    = healsim[.,1];                    @ Recall that 1 -> bad health @
    uwmat    = bhsim.*ln(wsim50); 
    hwmat    = (1-bhsim).*ln(wsim50);
    uhealobs = meanc(bhsim);           @ fraction of unhealthy obs at bornage @
    healobs  = 1-uhealobs;
    hmwage   = meanc(hwmat)./healobs;
    umwage   = meanc(uwmat)./uhealobs;

    mwage    = meanc(ln(wsim50));

    dscale   = zeros(_nn,1);       @ center wages around initial distribution @
    if (wscaled==1);
        dscale = mwage - hlogwage[1] - uhealobs*(ulogwage[1]-hlogwage[1]);
    elseif (wscaled==2);
        dscale   = hmwage-hlogwage[1]; @ observed wages for healthy person @
    endif;
    hlogwage = hlogwage + dscale;
    if (wscaled==3);
        dscale = umwage-ulogwage[1];  @ This will nullify effects of healbias @
    endif;
    ulogwage = ulogwage + dscale;
     
    "Wages for healthy and unhealthy at initial age:";
    "  Fixed effects profiles = ";; hwage_fe[1]~uwage_fe[1];
    "  Bias-adjusted profiles = ";; exp(hlogwage[1]~ulogwage[1]);
    "  Initial data distribution (everyone observed) = ";; exp(hmwage~umwage);?;
    
   @-----------Get initial distribution of wage deviations from HRS-----------@ 
   @---------Deviations based on model means, rather than data means----------@
   @-----------------Update deviations using wage shock model-----------------@
    devsim   = ln(wsim50) - bhsim*ulogwage[1] - (1-bhsim)*hlogwage[1];
    devsim   = recserar(winnov', devsim', rhow*ones(1,_nn));
    devsim   = devsim';      @ rows -> simulated individuals, columns -> ages @ 

   @-------------Keep wage deviations within limits of wage grid--------------@ 
    devmax   = wstate[_Wnum]+wXbound*(wstate[_Wnum]-wstate[_Wnum-1]); 
    devmin   = wstate[1]+wXbound*(wstate[1]-wstate[2]);
    "Log wage bounds = ";; devmin;; devmax;?;
       
    db       = devsim.>devmax;
    devsim   = devsim + db.*(devmax - devsim); 
    db       = devsim.<devmin;
    devsim   = devsim + db.*(devmin - devsim);

   @-----------------------Simulate wages for all ages------------------------@
    wsim     = {};
    wtypesim = {};
    wfracsim = {};          
   
    i=1; do until i>lspan;  
        bhsim   = healsim[.,i];
        logwage =         @ deterministic (conditional on health status) part @
                  hlogwage[i]*(1-bhsim) +  ulogwage[i]*bhsim;

       @--------------Fit deviations to Markov chain categories---------------@
        {wtypsimi,wfrcsimi} = discrtze(devsim[.,i],wstate); 
        wsimi   = exp(devsim[.,i]+logwage); 
 
        Wsim     = Wsim~wsimi;
        wtypesim = wtypesim~wtypsimi;
        wfracsim = wfracsim~wfrcsimi;
        if (bornage+i-1) == elasage;
            {wtypsimi,wfrcsimi} = discrtze(devsim[.,i]+elaspct,wstate); 
            wsimi = wsimi*exp(elaspct);
            save path = ^shkpath wtypsimi, wfrcsimi, wsimi;
        endif;

    i=i+1; endo;
@   "  Discrepancy with initial distribution =        ";; maxc(wsim[.,1]-wsim50);;minc(wsim[.,1]-wsim50);?;    @
   @-Mean simulated wages by age and health status, NOT conditional on working-@
    wsim     = ln(wsim);
    uwmat    = healsim.*wsim; 
    hwmat    = (ones(_nn,lspan)-healsim).*wsim;
    uhealobs = meanc(healsim);        @ fraction of unhealthy obs at each age @
    healobs  = 1-uhealobs;
    hmwage   = (meanc(hwmat))./healobs;
    umwage   = (meanc(uwmat))./uhealobs;
    hmwage   = exp(hmwage);
    umwage   = exp(umwage);
    hwage_ba = exp(hlogwage);
    uwage_ba = exp(ulogwage);    
    
    "Wage profiles:  estimated, bias-unadjusted (_wage); estimated, bias adjusted (_ba);";
    "  and simulated, bias-unadjusted (_mwage)";
    "      age       hwage   hwage_ba     hmwage      uwage   uwage_ba     umwage";;
    ageseq~hwage_fe~hwage_ba~hmwage~uwage_fe~uwage_ba~umwage;?; 

    wsim = exp(wsim);

    "      age    mean wages   maximum   minimum  mean type  mean frac";;
    ageseq~meanc(Wsim)~maxc(Wsim)~minc(Wsim)~meanc(wtypesim)~meanc(wfracsim);?; 
    
    "Std. dev of logged wages ... ";; ageseq~stdc(ln(wsim));?;
    save path = ^shkpath wsim, wtypesim, wfracsim; 
    format 12,4; 

retp(hlogwage,ulogwage); endp;

@-----------------------------------------------------------------------------@
@-----------------------------------------------------------------------------@
/* 
MAKEPREF:  Discretizes an index of an individual's willingness to work
*/

proc(1) = makepref(prefdata,gotindx,prefhrdl);

    local prefindx;
    prefindx = (prefdata.>prefhrdl);
    prefindx = (1+prefindx).*gotindx;        @ Missing values are set to zero @

retp(prefindx); endp;

@-----------------------------------------------------------------------------@
@-----------------------------------------------------------------------------@
/* 
GETINIT:  This gives the initial distribution of assets, wages, health,
          AIME, pensions, health insurance type and persistent health costs
          This data has not been run through the fixed-effects filter
*/
proc(4) = getinit;

    local file1, dat1, dat2, dat3, datnone, datret, dattied, badage, label, 
          i, cn, num, brthyr, hit50, age50, asim50, AIME50, pen50, wsim50,  
          part50, part49, hsim50, hc50, toosmall, lnhc50, mnhc50, mnhc, stdhc, 
          stdhc50, hitdum50, hsadj, hsadj50, sqdev50, prefhrdl, pref50;

    file1    = datapath$+"init_partfake_121009.txt";
    load dat1[] = ^file1; 
    dat1     = reshape(dat1, rows(dat1)/16, 16);
    prefhrdl = median(selif(dat1[.,15],dat1[.,16])); 
    pref50   = makepref(dat1[.,15],dat1[.,16],prefhrdl);
    dat1     = dat1~(pref50.==0)~(pref50.==1)~(pref50.==2);

    brthyr = dat1[.,12];
    badage = (brthyr.<inityrs[1]) + (brthyr.>inityrs[2]);
    dat2   = selif(dat1,(1-badage));
    pref50 = makepref(dat2[.,15],dat2[.,16],prefhrdl);
    dat2   = dat2~(pref50.==0)~(pref50.==1)~(pref50.==2);

    HIT50   = 1 + dat2[.,5] + dat2[.,6]*2;
    datnone = selif(dat2, HIT50.==1);
    datret  = selif(dat2, HIT50.==2);
    dattied = selif(dat2, HIT50.==3);

    let string label = {"AIME            ", "Medcost         ",
                        "Wage            ", "Health (1->bad) ",
                        "Retiree Coverage", "Tied Coverage   ",
                        "Age             ", "Assets          ",
                        "Participation   ", "Lagged Part.    ",
                        "Pension Wealth  ", "Birth Year      ",
                        "V3319           ", "V5009           ",
                        "Pref. Index     ", "Got An Index?   ",
                        "Index == 0      ", "Index == 1      ",
                        "Index == 2      "};

    "Data means for original and age-specific 1992 distributions";
    "Item              Total 92   age-spec      none        ret        tied";
    i=1; do until i>cols(dat1);
        label[i,1];;meanc(dat1[.,i]);;meanc(dat2[.,i]);;meanc(datnone[.,i]);; 
        meanc(datret[.,i]);; meanc(dattied[.,i]);
    i=i+1; endo;
    "Observations    ";; rows(dat1);; rows(dat2);; rows(datnone);; 
    rows(datret);; rows(dattied);; ?; 

   @--------Draw assets, wages and health from empirical distribution---------@
    cn     = rows(dat2);
    num    = rndu(_nn,1)*cn;                      @ random draws of indices @
    num    = floor(num)+onesim;
    dat3   = dat2[num,.];

    AIME50 = dat3[.,1];
    hc50   = dat3[.,2];
    wsim50 = dat3[.,3];     
    hsim50 = dat3[.,4]; 
    HIT50  = 1 + dat3[.,5] + dat3[.,6]*2;
    age50  = dat3[.,7];
    asim50 = dat3[.,8];
    part50 = dat3[.,9];
    part49 = dat3[.,10];
    pen50  = dat3[.,11];
    pref50 = makepref(dat3[.,15],dat3[.,16],prefhrdl);

   @---Use HI- conditional means and standard deviations to get normalized----@
   @---log health cost deviations.  Ignore participation, but let HS shift----@
   @-------------------means (effect on std dev is small).--------------------@

    toosmall = (hc50.<10);
    hc50     = (1-toosmall).*hc50+toosmall.*10;
    lnhc50   = ln(hc50);
    mnhc50   = zeros(_nn,1);
    hsadj50  = zeros(_nn,1);
    stdhc50  = zeros(_nn,1);

    i=1; do until i>HITnum;
        hitdum50 = (HIT50.==i);
        hsadj    = umnlnhc[1,i]-hmnlnhc[1,i]; 
        hsadj50  = hsadj50 + hitdum50.*(hsim50*hsadj);
        mnhc     = meanc(hitdum50.*(lnhc50-hsadj50))/meanc(hitdum50);
        mnhc50   = mnhc50 + hitdum50*mnhc;      
        sqdev50  = (lnhc50-mnhc50-hsadj50)^2;
        stdhc    = meanc(hitdum50.*sqdev50)/meanc(hitdum50);
        stdhc    = sqrt(stdhc);
        stdhc50  = stdhc50 + hitdum50*stdhc;
    i=i+1; endo;

    lnhc50 = (lnhc50-mnhc50-hsadj50)./stdhc50;     @ Rescale to unit variance @

    save path = ^shkpath wsim50, hit50, pen50, AIME50, part49, pref50, age50, 
                         lnhc50, prefhrdl;

retp(asim50,wsim50,hsim50,lnhc50); endp;

@-----------------------------------------------------------------------------@
@-----------------------------------------------------------------------------@
/*
INITSIM:  Simulate sequences of health and survival shocks and associated 
          wage growth shocks
*/
proc(2) = initsim(healg,healb,survg,survb);

    local hsimpo, ii, healhh, healhu, healuh, healuu, healshk, healsim, 
          healm, survshk, alivetpo, alivesim, ghealsim, bhealsim, wsim50, 
          asim50, hsim50, hcinnovp, hcinnovt, hitinnov, hcarsim, lnhcsim, 
          wtypeinv, winnov, hctypsim, hcfrcsim, hctypei, hcfraci, hcsim50, 
          oldassts, asstsign, lnabasst, varasst, bta, asstnoiz, PTinnov,
          survsim;

   @-------------Give hsimpo the intitial stationary distribution-------------@ 
   @-------------See Hamilton, p. 683 (w/ p_11 = hb, p_22 = 1-hg)-------------@
@   healshk  = rndu(_nn,1);                                                   @
@   hsimpo   = healshk .< ( onesim*healg[1,1]/(1-healb[1,1]+healg[1,1]) );    @

    {asim50, wsim50, hsim50, hcsim50} = getinit;

    healsim  = hsim50;
    alivesim = onesim;     

    ii=1; do until ii > (lspan-1);     @ Update with transition probabilities @
                                  @ for appropriate year, noting that 1 = bad @
        healshk  = rndu(_nn,1);
        hsimpo   = (healsim[.,ii].==0).*( healshk.<healg[ii,1] ) +
                   (healsim[.,ii].==1).*( healshk.<healb[ii,1] );
        survshk  = rndu(_nn,1);
        alivetpo = (alivesim[.,ii].==1).*
                   ( (healsim[.,ii].==0).*( survshk.<survg[ii,1] ) +
                     (healsim[.,ii].==1).*( survshk.<survb[ii,1] ) );
        healsim  = healsim~hsimpo;
        alivesim = alivesim~alivetpo;
    ii=ii+1; endo;

    if hhattrit == 0;
        alivesim = ones(_nn,lspan);
    endif;

    healm    = meanc(healsim);
    bhealsim = healsim.*alivesim;
    ghealsim = (1-healsim).*alivesim;
    survsim  = meanc(alivesim);
    save path = ^shkpath healsim, alivesim, ghealsim, bhealsim;  
    clear healsim, alivesim, ghealsim, bhealsim;

   @---------------Now generate matrix of wage type innovations---------------@ 
    wtypeinv = rndu(_nn,lspan);       @ Innovations to feed into Markov Chain @
    save path = ^shkpath wtypeinv;  
    clear wtypeinv;

   @-----Now generate matrix of wage innovations:  first column not used------@ 
    winnov = rndn(_nn, lspan)*stderrw; 
    "stderrw = " stderrw;
    save path = ^shkpath winnov;  
    clear winnov;

   @--------Now alter asset distribution to allow for measurment error--------@ 
    asstsign = -1+2*(asim50.>=0);
    oldassts = abs(asim50).<1;                           @ recode zero assets @
    oldassts = asim50.*(1-oldassts) + oldassts.*asstsign;
    lnabasst = ln(abs(oldassts));
    varasst  = vcx(lnabasst) - assterr;
    bta      = varasst/(varasst+assterr);
    asim50   = lnabasst*bta + (1-bta)*meanc(lnabasst);
    asim50   = asim50 + sqrt(bta*assterr)*rndn(_nn,1);
    asim50   = asstsign.*exp(asim50);
    asstnoiz = exp(sqrt(assterr)*rndn(_nn,lspan));
    asstnoiz[.,1] = oldassts./asim50;
    format /ro 12,4; 
    "mean assets and std dev of measured and 'true' assets, age 50 = "
        meanc(oldassts)~stdc(oldassts)~meanc(asim50)~stdc(asim50);?;
    "Std. Deviation of Measurement error";; ageseq~stdc(asstnoiz);?;
    save path= ^shkpath asim50, asstnoiz;
    clear asim50, asstnoiz;

   @---------------Now generate innovations to preference type----------------@
    PTinnov = rndu(_nn,1);
    save path= ^shkpath PTinnov;
    clear PTinnov;

   @-----Now generate health cost and health insurance status innovations-----@
    hcinnovp = rndn(lspan+1,_nn)*sqrt(fracar1i);
    hcinnovt = rndn(_nn,lspan+1)*sqrt(fracwn);
    hitinnov = rndu(_nn, lspan);

    hcarsim = fracar1*hcsim50';    @ Project onto data to get ar(1) component @
    hcarsim = hcarsim+sqrt(fracar1*(1-fracar1))*rndn(1,_nn);
    hcarsim = recserar(hcinnovp, hcarsim, rhohc*ones(1,_nn));
    hcarsim = hcarsim';
    lnhcsim = hcarsim+hcinnovt;

   @--Recall that time-t shocks are observed after time-t decisions, so that--@
   @-----the time-t-1 value of the AR(1) component is the relevant state------@
    lnhcsim = lnhcsim[.,2:lspan+1];
    hcarsim = hcarsim[.,1:lspan];
    clear hcinnovp, hcinnovt;
    
   @--------------------Discretize simulated AR(1) series---------------------@
   @-------------Note that discretized variable is unit-variance--------------@
    hctypsim={};
    hcfrcsim={};
    ii=1; do until ii>lspan;
        {hctypei,hcfraci} = discrtze(hcarsim[.,ii]/sqrt(fracar1),hcstate);  
        hctypsim = hctypsim~hctypei;
        hcfrcsim = hcfrcsim~hcfraci;
    ii=ii+1; endo;

    "Mean and std dev of logged health costs and their autoregressive component";
    "        age    mean(lnhc)    std(lnhc)  mean(lnhcar)  std(lnhcar)  mean(type)  mean(frac)";;
    ageseq~meanc(lnhcsim)~stdc(lnhcsim)~meanc(hcarsim)~stdc(hcarsim)~meanc(hctypsim)~meanc(hcfrcsim);?;

    save path= ^shkpath lnhcsim, hitinnov, hctypsim, hcfrcsim;
		 
retp(healm,survsim); endp;

@-----------------------------------------------------------------------------@
@-----------------------------------------------------------------------------@
/*
INITSIM0:  load in simulated sequences of health shocks and associated 
           wage growth shocks
*/
proc(2) = initsim0();
    local asim50, asstnoiz, lnhcsim, healsim, healm, alivesim, survsim;

    load path= ^shkpath asim50, asstnoiz;
    "Mean and std dev of 'true' (filtered) assets, age 50 = "
        meanc(asim50)~stdc(asim50);?;
    "Std. Deviation of Measurement error";; ageseq~stdc(asstnoiz);?;
    clear asim50, asstnoiz;

    load path= ^shkpath lnhcsim;
    "Mean and std dev of logged health costs";
    "        age    mean(lnhc)    std(lnhc)";;
    ageseq~meanc(lnhcsim)~stdc(lnhcsim);?;
    clear lnhcsim;

    load path = ^shkpath healsim, alivesim; 
    healm   = meanc(healsim);
    survsim = meanc(alivesim);
		 
retp(healm, survsim); endp;

@-----------------------------------------------------------------------------@
@-----------------------------------------------------------------------------@
/*
GETDATA:  load up the data that is used in the MSM criterion functions, and
          generate some summary statistics
*/

proc(18)=getdata();

   @-----------------prefix "h" => healthy, "u" => unhealthy------------------@
   @-----------------"h" => hours, "a" => assets, "p" => LFP------------------@

    local i, j, k, file1, ncols, rldata, datindx1, datindx2, rldata2, hswgts, 
          rn, cn, totobs, nyrs, goodages, datmns, datmnsa, obsmat, wgtvec,  
          ulfpr, hlfpr, lfprn, lfprr, lfprt, lfpr, uhd, hhd, asstd,  
          qndat, qcmdat, qcndat, qnhaz, qcmhaz, qcnhaz, 
          uhdat, uhdatn, uhdatr, uhdatt, hhdat, hhdatn, hhdatr, hhdatt,
          updat, updatn, updatr, updatt, hpdat, hpdatn, hpdatr, hpdatt,
          pdatn, pdatr, pdatt, pdat, xdatn, xdatr, xdatt, agpdat, prefdata,
          prefwgt0, prefwgt1, prefwgt2, pdat0, pdat1, pdat2, lfpr0, lfpr1,
          lfpr2, hazdatn, hazdatr, hazdatt, hazdat, agdathaz, adat, 
          uhwgt, uhwgtn, uhwgtr, uhwgtt, hhwgt, hhwgtn, hhwgtr, hhwgtt,
          upwgt, upwgtn, upwgtr, upwgtt, hpwgt, hpwgtn, hpwgtr, hpwgtt,
          pwgtn, pwgtr, pwgtt, pwgt, hwgtn, hwgtr, hwgtt, hazwgt, hazwgtn,
          hazwgtr, hazwgtt, assobs,            
          missmat, graphmat, ageseq2, rldata3, ppmat, missmat2, grphmat2,
          grphmat3, rn60, rn67, rn60s, pwgtn60, pwgtr60, pwgtt60, pwgt60, 
          pdatn60, pdatr60, pdatt60, pdat60, agpdat60, lfprn60, lfprr60, 
          lfprt60, lfpr60, summer, totlfpr, prefhrdl;

    format /ro 12,4; 
  
    nyrs     = lastdat-initdat+1;
    ncols    = 13*nyrs+3;    @ Last 3 columns are YOB, Prefindex and gotindex @
    datindx1 = bornage-initdat+1;
    datindx2 = bornage-initdat+datayrs;
    rn60     = 60-bornage+1;
    rn67     = minc( (rn60+(sumage-60)) |datayrs);
    rn60s    = rn67-rn60+1;

   @---------------Health Status Weights for Computing Moments----------------@

    file1    = datapath$+"wgtdata_111309.out";
    load hswgts[] = ^file1;
    hswgts   = reshape(hswgts, rows(hswgts)/ncols, ncols); 
    goodages = (hswgts[.,ncols-2].<mainyrs[1])+ (hswgts[.,ncols-2].>mainyrs[2]);
    goodages = 1-goodages;
    hswgts   = selif(hswgts,goodages);
    hswgts   = hswgts';   @ Now rows give years and columns give observations @
    "number of years of observations = " nyrs;
    "The initial hswgts matrix is " rows(hswgts) " by " cols(hswgts);

    assobs = hswgts[datindx1:datindx2,.];
    assobs = clnwgts(assobs,1);
    uhwgtn = hswgts[nyrs+datindx1:nyrs+datindx2,.];
    uhwgtr = hswgts[2*nyrs+datindx1:2*nyrs+datindx2,.];
    uhwgtt = hswgts[3*nyrs+datindx1:3*nyrs+datindx2,.]; 
    hhwgtn = hswgts[4*nyrs+datindx1:4*nyrs+datindx2,.];
    hhwgtr = hswgts[5*nyrs+datindx1:5*nyrs+datindx2,.];
    hhwgtt = hswgts[6*nyrs+datindx1:6*nyrs+datindx2,.]; 
    upwgtn = hswgts[7*nyrs+datindx1:7*nyrs+datindx2,.];
    upwgtr = hswgts[8*nyrs+datindx1:8*nyrs+datindx2,.];
    upwgtt = hswgts[9*nyrs+datindx1:9*nyrs+datindx2,.];
    hpwgtn = hswgts[10*nyrs+datindx1:10*nyrs+datindx2,.];
    hpwgtr = hswgts[11*nyrs+datindx1:11*nyrs+datindx2,.];
    hpwgtt = hswgts[12*nyrs+datindx1:12*nyrs+datindx2,.];
    clear hswgts;

   @----------Aggregate participation and hours across health status----------@
    pwgt  = upwgtn+upwgtr+upwgtt+hpwgtn+hpwgtr+hpwgtt;
    pwgt  = clnwgts(pwgt,0);
    upwgt = upwgtn+upwgtr+upwgtt;
    upwgt = clnwgts(upwgt,1);
    hpwgt = hpwgtn+hpwgtr+hpwgtt;
    hpwgt = clnwgts(hpwgt,2);

    pwgtn = upwgtn+hpwgtn;
    pwgtn = clnwgts(pwgtn,3);
    pwgtr = upwgtr+hpwgtr;
    pwgtr = clnwgts(pwgtr,4);
    pwgtt = upwgtt+hpwgtt;
    pwgtt = clnwgts(pwgtt,5);

    hwgtn = uhwgtn+hhwgtn;
    hwgtn = clnwgts(hwgtn,6);
    hwgtr = uhwgtr+hhwgtr;
    hwgtr = clnwgts(hwgtr,7);
    hwgtt = uhwgtt+hhwgtt;
    hwgtt = clnwgts(hwgtt,8);

    uhwgt = uhwgtn+uhwgtr+uhwgtt;
    uhwgt = clnwgts(uhwgt,9);
    hhwgt = hhwgtn+hhwgtr+hhwgtt;
    hhwgt = clnwgts(hhwgt,10);

   @-------------------------Get Hours and Asset Data-------------------------@

    file1    = datapath$+"maindata_111309.out";
    load rldata[] = ^file1;
    ncols    = 13*nyrs+3;              @ Last 2 columns are YOB and Prefindex @
    rldata   = reshape(rldata, rows(rldata)/ncols, ncols);
    goodages = (rldata[.,ncols-2].<mainyrs[1])+ (rldata[.,ncols-2].>mainyrs[2]);
    goodages = 1-goodages;
    rldata   = selif(rldata,goodages);
    totobs   = rows(rldata);
    rldata   = rldata';
    "The initial rldata matrix is " rows(rldata) " by " cols(rldata);

    adat   = rldata[datindx1:datindx2,.];
    uhdatn = rldata[nyrs+datindx1:nyrs+datindx2,.];
    uhdatr = rldata[2*nyrs+datindx1:2*nyrs+datindx2,.];  
    uhdatt = rldata[3*nyrs+datindx1:3*nyrs+datindx2,.];
    hhdatn = rldata[4*nyrs+datindx1:4*nyrs+datindx2,.];
    hhdatr = rldata[5*nyrs+datindx1:5*nyrs+datindx2,.];  
    hhdatt = rldata[6*nyrs+datindx1:6*nyrs+datindx2,.];
    updatn = rldata[7*nyrs+datindx1:7*nyrs+datindx2,.];
    updatr = rldata[8*nyrs+datindx1:8*nyrs+datindx2,.];
    updatt = rldata[9*nyrs+datindx1:9*nyrs+datindx2,.];
    hpdatn = rldata[10*nyrs+datindx1:10*nyrs+datindx2,.];
    hpdatr = rldata[11*nyrs+datindx1:11*nyrs+datindx2,.];
    hpdatt = rldata[12*nyrs+datindx1:12*nyrs+datindx2,.];
    prefdata = rldata[ncols-1:ncols,.];
    clear rldata;

   @----------Aggregate participation and hours across health status----------@
   @----To get values conditional on health status, we average across HIT-----@
   @----To get values conditional on HIT, we average across health status-----@
   @--------This keeps weights from being double-multiplied elsewhere---------@

    updat = updatn.*pwgtn+updatr.*pwgtr+updatt.*pwgtt;
    hpdat = hpdatn.*pwgtn+hpdatr.*pwgtr+hpdatt.*pwgtt;
    pdatn = updatn.*upwgt+hpdatn.*hpwgt;
    pdatr = updatr.*upwgt+hpdatr.*hpwgt;
    pdatt = updatt.*upwgt+hpdatt.*hpwgt;
    pdat  = pdatn.*pwgtn+pdatr.*pwgtr+pdatt.*pwgtt;
    pdat  = pdat.>0.5;       @ Ensures that p = 0 or 1, this is NOT always OK @

    load path = ^shkpath prefhrdl;
@prefhrdl = median(selif(prefdata[1,.]', prefdata[2,.]')); @
    prefdata = makepref(prefdata[1,.]', prefdata[2,.]',prefhrdl)';

    clear upwgtn, hpwgtn, upwgtr, hpwgtr, upwgtt, hpwgtt;
    clear updatn, updatr, updatt, hpdatn, hpdatr, hpdatt;

    prefwgt0 = ones(datayrs,1)*(prefdata.==0);
    prefwgt1 = ones(datayrs,1)*(prefdata.==1);
    prefwgt2 = ones(datayrs,1)*(prefdata.==2);

    prefwgt0 = prefwgt0.*pwgt;
    prefwgt1 = prefwgt1.*pwgt;
    prefwgt2 = prefwgt2.*pwgt;

    pdat0 = pdat.*prefwgt0;
    pdat1 = pdat.*prefwgt1;
    pdat2 = pdat.*prefwgt2;

    uhdat = uhdatn.*hwgtn+uhdatr.*hwgtr+uhdatt.*hwgtt;
    hhdat = hhdatn.*hwgtn+hhdatr.*hwgtr+hhdatt.*hwgtt; 

    clear uhwgtn, uhwgtr, uhwgtt, hhwgtn, hhwgtr, hhwgtt;
    clear uhdatn, uhdatr, uhdatt, hhdatn, hhdatr, hhdatt;

    save path= ^datapath uhdat, hhdat, updat, hpdat, pdatn, pdatr, pdatt, 
                         pdat0, pdat1, pdat2, adat; 

    asstd  = meanc(adat');
    uhd    = meanc((uhdat.*uhwgt)');
    hhd    = meanc((hhdat.*hhwgt)');
    ulfpr  = meanc((updat.*upwgt)');
    hlfpr  = meanc((hpdat.*hpwgt)');
    lfprn  = meanc((pdatn.*pwgtn)');
    lfprr  = meanc((pdatr.*pwgtr)');
    lfprt  = meanc((pdatt.*pwgtt)');
    lfpr   = meanc(pdat');
    lfpr0  = meanc(pdat0');
    lfpr1  = meanc(pdat1');
    lfpr2  = meanc(pdat2');

   @----------------------Adjust means for missing data-----------------------@
    uhd   = uhd./meanc(uhwgt'); 
    hhd   = hhd./meanc(hhwgt'); 
    ulfpr = ulfpr./meanc(upwgt');
    hlfpr = hlfpr./meanc(hpwgt');
    lfprn = lfprn./meanc(pwgtn');
    lfprr = lfprr./meanc(pwgtr');
    lfprt = lfprt./meanc(pwgtt'); 
    lfpr  = lfpr./meanc(pwgt');
    lfpr0 = lfpr0./meanc(prefwgt0');
    lfpr1 = lfpr1./meanc(prefwgt1'); 
    lfpr2 = lfpr2./meanc(prefwgt2');   
    asstd = asstd./meanc(assobs');
    agpdat = lfpr~lfprn~lfprr~lfprt~lfpr0~lfpr1~lfpr2;

    summer  = zeros(datayrs,1);
    summer[rn60:rn67] = ones(rn60s,1);

    pwgtn60 = ones(datayrs,1)*pwgtn[rn60,.];
    pwgtn60 = pwgtn60.*(pwgt./=0);
    pwgtr60 = ones(datayrs,1)*pwgtr[rn60,.];
    pwgtr60 = pwgtr60.*(pwgt./=0);
    pwgtt60 = ones(datayrs,1)*pwgtt[rn60,.];
    pwgtt60 = pwgtt60.*(pwgt./=0);
    pwgt60  = ones(datayrs,1)*pwgt[rn60,.];
    pwgt60  = pwgt60.*(pwgt./=0);
    pdatn60 = pdat.*pwgtn60;
    pdatr60 = pdat.*pwgtr60;
    pdatt60 = pdat.*pwgtt60;
    pdat60  = pdat.*pwgt60;
    lfprn60 = meanc(pdatn60')./meanc(pwgtn60');
    lfprr60 = meanc(pdatr60')./meanc(pwgtr60');
    lfprt60 = meanc(pdatt60')./meanc(pwgtt60');
    lfpr60  = meanc(pdat60')./meanc(pwgt60');
    agpdat60 = lfpr60~lfprn60~lfprr60~lfprt60;
    totlfpr  = agpdat60'summer;

    expmtrec[1:rn60s,expmtnum]     = lfpr[rn60:rn67];
    expmtrec[rn60s+1,expmtnum]     = lfpr'summer;
    expmtrec[rn60s+2:(2*rn60s+1),expmtnum] = lfprt60[rn60:rn67];
    expmtrec[2*(rn60s+1),expmtnum] = totlfpr[4];

    "Data means (after adjustments for missing obs) " ;
    "       age         uhours       hhours        ulfpr        hlfpr       ";;
    "lfpr-n       lfpr-r       lfpr-t       assets";;
    ageseq2 = seqa(bornage,1,datayrs);
    ageseq2~uhd~hhd~ulfpr~hlfpr~lfprn~lfprr~lfprt~asstd;?;

    "Aggregate LFPR";
    "        Age          All         None       Retiree        Tied";;
    "        pref=0       pref=1       pref=2";; 
    ageseq[1:datayrs,.]~agpdat;?;
    grphmat3 = ageseq[1:datayrs,.]~agpdat[.,1:4];

    "Aggregate LFPR Based on age-60 categories ";
    "        Age          All         None       Retiree        Tied";;
    ageseq[1:datayrs,.]~agpdat60;?;
    "Total 60-69 ";; totlfpr';?;

    pwgtn60 = pwgtn60.*( ones(datayrs,1)*(pdatn60[rn60,.].>0) );
    pwgtr60 = pwgtr60.*( ones(datayrs,1)*(pdatr60[rn60,.].>0) );
    pwgtt60 = pwgtt60.*( ones(datayrs,1)*(pdatt60[rn60,.].>0) );
    pwgt60  = pwgt60.*( ones(datayrs,1)*(pdat60[rn60,.].>0) );
    pdatn60 = pdat.*pwgtn60;
    pdatr60 = pdat.*pwgtr60;
    pdatt60 = pdat.*pwgtt60;
    pdat60  = pdat.*pwgt60;
    lfprn60 = meanc(pdatn60')./meanc(pwgtn60');
    lfprr60 = meanc(pdatr60')./meanc(pwgtr60');
    lfprt60 = meanc(pdatt60')./meanc(pwgtt60');
    lfpr60  = meanc(pdat60')./meanc(pwgt60');
    agpdat60 = lfpr60~lfprn60~lfprr60~lfprt60;
    totlfpr  = agpdat60'summer;

    "Aggregate LFPR Based on age-60 categories (and working)";
    "        Age          All         None       Retiree        Tied";;
    ageseq[1:datayrs,.]~agpdat60;?;
    "Total 60-69 ";; totlfpr';?;

   @-------Now get asset quantiles and asset-quantile-conditional means-------@

    missmat = miss(uhwgt .> 0,0);
    rldata3 = (uhdat.*missmat)';
    missmat = miss(hhwgt.>0,0);
    rldata3 = rldata3~( (hhdat.*missmat)' );
    missmat = miss(upwgt.>0,0);
    rldata3 = rldata3~( (updat.*missmat)' );
    missmat = miss(hpwgt.>0,0);
    rldata3 = rldata3~( (hpdat.*missmat)' );
    missmat = miss(pwgtn.>0,0);
    rldata3 = rldata3~( (pdatn.*missmat)' );
    missmat = miss(pwgtr.>0,0);
    rldata3 = rldata3~( (pdatr.*missmat)' );
    missmat = miss(pwgtt.>0,0);
    rldata3 = rldata3~( (pdatt.*missmat)' );
    missmat = miss(assobs.>0,0);
    rldata3 = rldata3~( (adat.*missmat)' );
    cn = cols(rldata3)/datayrs;

    hswgts = (uhwgt|hhwgt|upwgt|hpwgt|pwgtn|pwgtr|pwgtt|assobs)';
    save path=^datapath uhwgt, hhwgt, upwgt, hpwgt;
    clear uhwgt, hhwgt, upwgt, hpwgt, missmat;

    {qndat,qcmdat,qcndat} = doquant2(rldata3,hswgts,cn,quants,datayrs,bornage,1);
    clear hswgts;
    save path = ^datapath rldata3;   

    if nosaving==0;
        "        age       asset quantiles and cell counts";; qndat;?;
        "Asset-quantile-conditional means "; prlabel; qcmdat;?;
        "Asset-quantile-conditional counts "; prlabel; qcndat;?;
    endif;

    rn     = qnum+1;
    qndat  = qndat[2:datayrs+1,2:rn];                     @ Eliminate headers @
    qcmdat = qcmdat[2:datayrs+1,2:1+rn*cn];               @ Eliminate headers @
    qcndat = qcndat[2:datayrs+1,2:1+rn*cn];               @ Eliminate headers @

    local apdf, acpdf, qptcm, acpdfnum, acdf;
    {apdf, acpdf, qptcm, acpdfnum, acdf} = 
       getkern(rldata3,cn,quants,bornage,qndat);
    clear rldata3, apdf, acpdf, qptcm, acpdfnum, acdf;

    graphmat = ageseq[1:datayrs,.]~uhd[1:datayrs,.]~hhd[1:datayrs,.];
    graphmat = graphmat~ulfpr[1:datayrs,.]~hlfpr[1:datayrs,.];
    graphmat = graphmat~qcmdat[1:datayrs,13:21]~qndat[1:datayrs,1:2];
    graphmat = graphmat~zeros(datayrs,cols(graphmat)-1);
    save path = ^shkpath graphmat;

   @---------------------------Now Get Hazard Rates---------------------------@

    ppmat = ones(1,cols(pdat))|pdat[1:datayrs-1,.];@ Hazard rates conditional @    
    ppmat = (ppmat.>0.5);                 @ on having worked in previous year @

    hazwgt   = pwgt.*ppmat;
    hazwgtn  = pwgtn.*ppmat;
    hazwgtr  = pwgtr.*ppmat;
    hazwgtt  = pwgtt.*ppmat;

    hazdat   = hazwgt.*(1-pdat);
    hazdat   = meanc(hazdat')./meanc(hazwgt');
    hazdatn  = hazwgtn.*(1-pdatn);
    xdatn    = hazdatn;
    hazdatn  = meanc(hazdatn')./meanc(hazwgtn');
    hazdatr  = hazwgtr.*(1-pdatr);
    xdatr    = hazdatr;
    hazdatr  = meanc(hazdatr')./meanc(hazwgtr');
    hazdatt  = hazwgtt.*(1-pdatt);
    xdatt    = hazdatt;
    hazdatt  = meanc(hazdatt')./meanc(hazwgtt');
    agdathaz = hazdat~hazdatn~hazdatr~hazdatt;

    ?;"Aggregate hazard rates, Data"; 
    "        Age          All         None       Retiree        Tied";;
    ageseq[1:datayrs,.]~agdathaz;?;
    grphmat3 = grphmat3~agdathaz;
    grphmat3 = grphmat3~zeros(datayrs,cols(grphmat3)-1);
    save path = ^shkpath grphmat3;

    rldata3  = {};
    hswgts   = {};
    missmat2 = miss(ppmat,0);
    missmat  = miss(pwgtn.>0,0).*missmat2;
    hswgts   = hswgts|(hazwgtn);
    rldata3  = rldata3~( (pdat.*missmat)' );
    missmat  = miss(pwgtr.>0,0).*missmat2;
    hswgts   = hswgts|(hazwgtr);
    rldata3  = rldata3~( (pdat.*missmat)' );
    missmat  = miss(pwgtt.>0,0).*missmat2;
    hswgts   = hswgts|(hazwgtt);
    rldata3  = rldata3~( (pdat.*missmat)' );
    rldata3  = 1 - rldata3;              @ Let 1 indicate exiting the market @
    missmat  = miss(assobs.>0,0);
    hswgts   = (hswgts|assobs)';
    rldata3  = rldata3~( (adat.*missmat)' );
    cn       = cols(rldata3)/datayrs;

    clear missmat, missmat2;
    {qnhaz,qcmhaz,qcnhaz} = doquant2(rldata3,hswgts,cn,quants,datayrs,bornage,1);
    clear rldata3,hswgts;   

    if nosaving==0;
        "        age       asset quantiles and cell counts";; qnhaz;?;
        "Asset-quantile-conditional means (Hazard Rates) "; prlabel3; qcmhaz;?;
        "Asset-quantile-conditional counts "; prlabel3; qcnhaz;?;
    endif;
    qcmhaz = qcmhaz[2:datayrs+1,2:1+rn*cn];               @ Eliminate headers @
    grphmat2  = ageseq[1:datayrs,.]~qcmhaz[1:datayrs,1:9];
    save path = ^shkpath grphmat2;

   @---------Adjust hazard rate data if age-60 HI categories are used---------@
    if exit60==1;
        pwgtn60  = ones(datayrs,1)*pwgtn[2,.];      @ 2=rn60 in baseline case @
        pwgtn60  = pwgtn60.*(pwgt./=0);
        hazwgtn  = pwgtn60.*ppmat;
        pwgtr60  = pwgtr;                        @ Retiree coverage permanent @
        pwgtr60  = pwgtr60.*(pwgt./=0);
        hazwgtr  = pwgtr60.*ppmat;
        pwgtt60  = ones(datayrs,1)*pwgtt[rn60,.];
        pwgtt60  = pwgtt60.*(pwgt./=0);
        hazwgtt  = pwgtt60.*ppmat;
        pwgt60   = pwgtn60+pwgtr60+pwgtt60;
        pwgt60   = pwgt60.*(pwgt./=0);
        hazwgt   = pwgt60.*ppmat;

        hazdat   = hazwgt.*(1-pdat);
        hazdat   = meanc(hazdat')./meanc(hazwgt');
        hazdatn  = hazwgtn.*(1-pdat);
        xdatn    = hazdatn;
        hazdatn  = meanc(hazdatn')./meanc(hazwgtn');
        hazdatr  = hazwgtr.*(1-pdat);
        xdatr    = hazdatr;
        hazdatr  = meanc(hazdatr')./meanc(hazwgtr');
        hazdatt  = hazwgtt.*(1-pdat);
        xdatt    = hazdatt;
        hazdatt  = meanc(hazdatt')./meanc(hazwgtt');
        agdathaz = hazdat~hazdatn~hazdatr~hazdatt;

        ?;"Aggregate hazard rates, Data, using age-60 categories"; 
        "        Age          All         None       Retiree        Tied";;
        ageseq[1:datayrs,.]~agdathaz;?;
    endif;

    save path=^datapath xdatn, xdatr, xdatt;

   @--------Construct Health Status/HI type Weighting Matrix Used for---------@
   @----------------Constructing the Moment Conditions for MSM----------------@
   @------obsmat does NOT adjust for assets falling outside the quantile------@
   @-when we calculate asset-quantile-conditional means:  makemmts does this.-@

    load path=^datapath uhwgt, hhwgt, upwgt, hpwgt;
    hswgts = mkhswgts(uhwgt,hhwgt,upwgt,hpwgt,hazwgtn,hazwgtr,hazwgtt,
                      prefwgt0,prefwgt1,prefwgt2,pwgtn,pwgtr,pwgtt,assobs);
    clear uhwgt, hhwgt, upwgt, hpwgt, hazwgtn, hazwgtr, hazwgtt, prefwgt0,
          prefwgt1, prefwgt2, pwgtn, pwgtr, pwgtt, assobs;
    obsmat = hswgts.>0;              @ Works if missing obs have zero weights @
    wgtvec = meanc(hswgts');   @ Fractions based on probability weights       @
    save path = ^datapath hswgts, obsmat;  
    clear hswgts, obsmat;

   @--------Now create weighted, centered moments for each observation--------@

    rldata = makemmts(totobs,uhd,hhd,ulfpr,hlfpr,hazdatn,hazdatr,hazdatt,
                      lfprn,lfprr,lfprt,lfpr0,lfpr1,lfpr2,qndat,qcmdat,
                      uhdat,hhdat,updat,hpdat,xdatn,xdatr,xdatt,
                      pdat0,pdat1,pdat2,pdatn,pdatr,pdatt,adat);

    load path = ^datapath obsmat;   
    rldata = rldata.*obsmat;  @ zero out missing obs.  hswgts also does this, @
    clear obsmat;                 @ but this makes rldata easier to interpret @
    save path = ^datapath rldata;   

   @------------------------------De-mean rldata------------------------------@
    load path = ^datapath hswgts;   
    datmns=meanc((rldata.*hswgts)');
    clear rldata, hswgts;
    datmnsa=datmns./wgtvec;
     "                 datmns       wgtvec       datmnsa";;
    seqa(1,1,rows(datmnsa))~datmns~wgtvec~datmnsa;?; 

   @----------Construct moment conditions for the weighting matrices----------@
   @-------if necessary, use moment conditions that aggregate slightly--------@
   @-------------------to ensure data has positive variance-------------------@
    if truevcv == 1;
        load path = ^datapath rldata;
        rldata2   = rldata;
        clear rldata;
    else;   
        rldata2 = makmmts2(totobs,uhd,hhd,ulfpr,hlfpr,hazdatn,hazdatr,hazdatt,
                           lfprn,lfprr,lfprt,lfpr0,lfpr1,lfpr2,qndat,qcmdat,
                           uhdat,hhdat,updat,hpdat,xdatn,xdatr,xdatt,
                           pdat0,pdat1,pdat2,pdatn,pdatr,pdatt,adat);
        load path = ^datapath obsmat; 
        rldata2   = rldata2.*obsmat;@ zero out missing obs.  hswgts also does @
                            @ this, but this makes rldata easier to interpret @  
        clear obsmat;
    endif;

    save path = ^datapath rldata2; 
    datmns = meanc(rldata2');
    clear uhdat, hhdat, updat, hpdat, xdatn, xdatr, xdatt, pdatn, pdatr, 
          pdatt, adat, rldata2;
    datmnsa = datmns./wgtvec;

retp(totobs,uhd,hhd,ulfpr,hlfpr,lfprn,lfprr,lfprt,lfpr0,lfpr1,lfpr2,asstd,
     hazdatn,hazdatr,hazdatt,qndat,qcmdat,datmnsa);
endp;

@-----------------------------------------------------------------------------@
@-----------------------------------------------------------------------------@
proc(1) = clnwgts(wgts,item);

    local toosmall, toobig, maxwgt, minwgt;

    maxwgt = maxc(maxc(wgts));
    minwgt = minc(minc(wgts));
    if maxwgt > 1;
        "Warning maximum probability weight is";; maxwgt;; "for weight set";; item;
        toobig = wgts.>1;
        wgts = wgts.*(1-toobig) + toobig;
    endif;
    if minwgt < 0;
        "Warning minimum probability weight is";; minwgt;; "for weight set";; item;
        toosmall = wgts.<0;
        wgts = wgts.*(1-toosmall);
    endif;

retp(wgts); endp;

@-----------------------------------------------------------------------------@
@-----------------------------------------------------------------------------@

proc(3) = compvinv(mnvals,totobs);

    local v, w, vinv, rldata, rldata2, hswgts, rn;

    load path =^datapath rldata2;
    rldata = rldata2 - mnvals;  
    clear rldata2;
    load path =^datapath hswgts;
    rldata = rldata.*hswgts;
    clear hswgts;
    v = rldata*rldata';
    clear rldata;
    v = v/totobs;

   @-----Take the principal diagonal and form a diagonal weighting matrix-----@
    vinv     = inv(v);  
    rn       = rows(v);
    w        = diag(v);
    w        = diagrv(eye(rn), w);
    w        = inv(w);

    format /ro 12,4; 
    "Principal diagonals of V, Vinv (V^{-1}), and W = ";;
    seqa(1,1,rn)~diag(v)~diag(vinv)~diag(w);?;

retp(v,w,vinv); endp;

@-----------------------------------------------------------------------------@
@-----------------------------------------------------------------------------@
/*
GETHP:  Finds unconditional probabilities of being in bad health at each age
        Also finds unconditional survival probabilities
*/
proc(3) = gethp(healg,healb,survdatg,survdatb);

    local bhprob, ghprob, i, rn, sprob, rn2, ltbhpr;

    rn = rows(healb);
    bhprob = zeros(rn,1);
    ghprob = bhprob;
    ltbhpr = bhprob;                               @ Pr(h_t-1=bad|h_t = bad) @

@-Start with invariant prob ala Hamilton, p. 683 (w/ p_11 = hb, p_22 = 1-hg)-@
    bhprob[1,1] = healg[1,1]/(1-healb[1,1]+healg[1,1]);
    ghprob[1,1] = 1-bhprob[1,1];

    i=2; do until i>rn;
        bhprob[i,1] = bhprob[i-1,1]*healb[i-1,1]*survdatb[i-1,1] + 
                      ghprob[i-1,1]*healg[i-1,1]*survdatg[i-1,1];
        ltbhpr[i,1] = (bhprob[i-1,1]*healb[i-1,1]*survdatb[i-1,1])/bhprob[i,1];
        ghprob[i,1] = bhprob[i-1,1]*(1-healb[i-1,1])*survdatb[i-1,1] + 
                      ghprob[i-1,1]*(1-healg[i-1,1])*survdatg[i-1,1];
        bhprob[i,1] = bhprob[i,1]/(bhprob[i,1]+ghprob[i,1]);       
        ghprob[i,1] = 1 - bhprob[i,1];
    i=i+1; endo;
    ltbhpr[1,1] = ltbhpr[2,1];

   @------Assume that healg and healm are probs conditional on surviving------@
    sprob = survdatg.*ghprob + survdatb.*bhprob;

retp(bhprob,sprob,ltbhpr); endp;

@-----------------------------------------------------------------------------@
@-----------------------------------------------------------------------------@

proc(1) = mkhswgts(uhwgt,hhwgt,upwgt,hpwgt,hazwgtn,hazwgtr,hazwgtt,
                   prefwgt0,prefwgt1,prefwgt2,pwgtn,pwgtr,pwgtt,assobs);

    local hswgts,rn,pwgts;
    rn = rows(quants);

   @----Weights for unconditional (actually age/health-conditional) means-----@
    hswgts = uhwgt[mashft:datayrs,.]|hhwgt[mashft:datayrs,.];
    hswgts = hswgts|upwgt[mashft:datayrs,.]|hpwgt[mashft:datayrs,.];

   @-----------------Weights for HI-type-conditional hazards------------------@
   @-----Note that first year might be skipped, as it can be uninformative----@
    hswgts = hswgts|hazwgtn[(mashft+hazshft):datayrs,.];
    hswgts = hswgts|hazwgtr[(mashft+hazshft):datayrs,.];
    hswgts = hswgts|hazwgtt[(mashft+hazshft):datayrs,.];

   @----------------Weights for pref-type-conditional hazards-----------------@
    if prefmoms > 0;
        hswgts = hswgts|prefwgt0[mashft:datayrs,.];
        hswgts = hswgts|prefwgt1[mashft:datayrs,.];
        hswgts = hswgts|prefwgt2[mashft:datayrs,.];  
    endif;

    if nosaving==0;
       @------Weights for each asset-quantile-conditional mean.  Set to-------@
       @---Zero if assets and/or participation are missing (Assobs is 0/1)----@
        pwgts  = pwgtn[mashft:datayrs,.].*assobs[mashft:datayrs,.];
        pwgts  = pwgts|(pwgtr[mashft:datayrs,.].*assobs[mashft:datayrs,.]);
        pwgts  = pwgts|(pwgtt[mashft:datayrs,.].*assobs[mashft:datayrs,.]);
        hswgts = hswgts|(ones(rn+1,1).*.pwgts);  @ ".*." is the kronecker product @

       @----------------Weights for each asset quantile moment----------------@
        hswgts = hswgts|(ones(rn,1).*.assobs[mashft:datayrs,.]);

    elseif nosaving==1;
        hswgts = hswgts|pwgtn[mashft:datayrs,.];
        hswgts = hswgts|pwgtr[mashft:datayrs,.];
    endif;


retp(hswgts); endp;

@-----------------------------------------------------------------------------@
@-----------------------------------------------------------------------------@
/*
COMPUTEG:  Computes the "full" annuity payment for a given pension balance
           The actual payment is the fraction penprob of the full payment
*/
proc(5) = computeg(atror,sprob);

    local gvec, gvec2, gvec3, d, d2, discount, tseq, penprob, penprob0, j, k, 
          AIMEfrc, AIMEfrc0, filename;

    filename = datapath$+"AIMEfrac_012207.txt";
    load AIMEfrc[17,2] = ^filename;

    {k,j}    = getages(bornage,dieage,AIMEfrc);
    AIMEfrc  = fixobs(AIMEfrc,k,j,0,1);
    AIMEfrc0 = AIMEfrc[.,2];  @ Fraction with "full" (35 years) work history @
 
   @--Gvec is computed assuming that no pension payments are made at dieage---@
    sprob    = sprob[1:lspan,1];
    gvec     = 0*sprob;
    discount = gvec;
    tseq     = seqa(0,1,lspan);
    d        = recsercp(sprob[1:lspan-1],zeros(lspan-1,1)); @ Use sprob_t to discount x_t+1 @
    d        = 1|d;
    d2       = (1+atror)^(-tseq);
    discount = d.*d2;         @ Survival-weighted cumulative discount factors @

    filename = datapath$+"penprob_020204.txt";
    load penprob[16,2] = ^filename;

    {k,j}    = getages(bornage,dieage,penprob);
    penprob  = fixobs(penprob,k,j,0,1);
    penprob0 = penprob[.,2];                @ Fraction of "full benefit" paid @
    d2       = discount;
    discount = discount.*penprob0;
    gvec2    = rev(cumsumc(rev(discount)));
    gvec2    = gvec2./d2;           @ EPDV of penprob*constant annual payment @
    gvec2    = gvec2/(1+atror);

    gvec     = gvec2^-1;
    gvec     = gvec.*penprob0;   @ Benefit for a given pension wealth balance @

    discount = d2;        @ Now assume that full benefit is paid every period @
    gvec3    = rev(cumsumc(rev(discount)));
    gvec3    = gvec3./d2;                   @ EPDV of constant annual payment @
    gvec3    = gvec3/(1+atror);

    "Pension conversion factors:  Wealth-to-payment, Payment-to-Wealth";
    "        age       payment       wealth      penprob      wealth2     AIMEfrac";; 
    ageseq~gvec~gvec2~penprob0~gvec3~AIMEfrc0;?;

retp(gvec,gvec2,gvec3,penprob0,AIMEfrc0); endp;

@-----------------------------------------------------------------------------@
@-----------------------------------------------------------------------------@

proc(0) = prlabel;

    "       age          ------------uhours----------          ";;
    "------------hhours----------          ------------upart-----------";;
    "              ------------hpart-----------          ";;
    "----------partnone----------           -----------partret----------";;
    "             ----------partied----------           ";;
    "------------assets----------";;

retp; endp;

@-----------------------------------------------------------------------------@
@-----------------------------------------------------------------------------@

proc(0) = prlabel3;

    "       age          ----------partnone----------            ";;
    "-----------partret----------            ----------partied----------";;
    "           -----------assets-----------";;

retp; endp;

@-----------------------------------------------------------------------------@
@-----------------------------------------------------------------------------@

proc(7) = piacfs(pbscale,piascale,pacscale,samepen); 

/*---------------------------------------------------------------------------*/
/*---------------------------------------------------------------------------*/
/*         Below are the bend points for AIME and PIA, respectively          */
/*             AIME                    PIA                                   */
/*             5724                    AIME*.9 = 5151.6                      */
/*             34500                   5151.6 +.32*(AIME-5724) = 14359.9     */
/*             34500+                  14359.9 +.15*(AIME-34500)             */                       
/*                                                                           */
/*         There is also a "fake" bendpoint at AIME = 20874, PIA =9999.6,    */
/*         for use in pension imputation.                                    */
/*                                                                           */
/*         Pension benefits are imputed by:                                  */
/*             pb = p0(HItype) + p1*PIA + p2*max{0,PIA-9999.6}               */
/*                             + p2*max{0,PIA-14359.9}                       */
/*             p0 differs across HI types                                    */
/*         Pension accrual is imputed by:                                    */
/*             pacc = p0(HItype) + p1*ln(earns)                              */
/*---------------------------------------------------------------------------*/
/*---------------------------------------------------------------------------*/

    local pencfs, pencfs0, file1, penacc, penacc0, mpr, AIMEbend, PIAbend, 
          pbbend0, p0, p1, p2, p3, cn, k, j, paccprob, trubends, mearn;
             
    load path=^datapath pencfs = pencfs_base_030510;  
    {k,j}   = getages(bornage,dieage,pencfs);
    pencfs0 = fixobs(pencfs,k,j,0,0)*pbscale;
    pencfs0 = pencfs0[.,2:cols(pencfs0)];
    if samepen == 1;
        pencfs0[.,1:HITnum-1] = pencfs0[.,HITnum]*ones(1,HITnum-1);
    endif;
    pencfs0[.,HITnum+1:HITnum+3] = pencfs0[.,HITnum+1:HITnum+3]/piascale;         

    mpr        = {0.9, 0.32, 0.32, 0.15};
    mpr        = mpr*piascale;
    AIMEbend   = {5724, 20874, 34500};                         @  1998 rules  @
    AIMEbend   = AIMEbend|AIMEmax;
    PIAbend    = zeros(4,1);
    PIAbend[1] = AIMEbend[1]*mpr[1];
    PIAbend[2] = PIAbend[1] + (AIMEbend[2]-AIMEbend[1])*mpr[2];
    PIAbend[3] = PIAbend[2] + (AIMEbend[3]-AIMEbend[2])*mpr[3];
    PIAbend[4] = PIAbend[3] + (AIMEbend[4]-AIMEbend[3])*mpr[4];

    pbbend0    = (PIAbend[2|3]').*.ones(lspan,HITnum);
    trubends   = 1|3|4;
    mpr        = mpr[trubends];
    AIMEbend   = AIMEbend[trubends];
    PIAbend    = PIAbend[trubends];

   @-------------------Pension accrual over the life cycle--------------------@
    file1 = datapath$+"penacc_072606.out";
    load penacc[19,5] = ^file1;
    cn = {1 0 0 0 0,
          0 1 1 1 0,
          0 0 0 1 0,
          0 0 1 0 0,
          0 0 0 0 1};
    penacc0 = penacc*cn;
    {k,j} = getages(bornage,dieage,penacc0);
    penacc0 = fixobs(penacc0,k,j,0,0)*pacscale;
    penacc0 = penacc0[.,2:HITnum+2];
    if samepen == 1;
        penacc0[.,1:HITnum-1] = penacc0[.,HITnum]*ones(1,HITnum-1);
    endif;

    "    mpr(PIA)     AIMEbends     PIAbends";;
    mpr~AIMEbend~PIAbend;

    ?;"Pension Wealth Imputation Coefficients";
    "        Age         None      Retiree         Tied          PIA    PIA - 10000    PIA - 14360";; 
    ageseq~pencfs0;?;

    "Pension Imputation kinks";
    "         Age       None       Retiree         Tied       ";;
    "None       Retiree         Tied";; 
    ageseq~pbbend0;?;

    mearn = 10.39331;                         @ Mean logged earnings, from data @
    cn = {1 0 0 0,
          0 1 0 0,
          0 0 1 0};
    cn = cn|( (mearn*ones(1,3))~0.72273);

    "Pension Accrual Coefficients";
    "         Age       None       Retiree         Tied    ln(Income)";;
    ageseq~penacc0~(penacc0*cn);?;

    file1 = datapath$+"paccprob_071406.out";
    load paccprob[1,7] = ^file1;

    cn = {1 1 1 0 0 0 0,
          0 0 1 0 0 0 0,
          0 1 0 0 0 0 0,
          0 0 0 0 0 1 0,
          0 0 0 0 1 0 0,
          0 0 0 1 1 1 0,
          0 0 0 0 0 0 1};
    paccprob = (paccprob*cn)';
    if samepen == 1;
        paccprob[1:HITnum-1]          = paccprob[HITnum]*ones(HITnum-1,1);
        paccprob[HITnum+1:2*HITnum-1] = paccprob[2*HITnum]*ones(HITnum-1,1);
    endif;
    "Probability of Pension Accrual";
    "        None       Retiree         Tied   ln(Income)_N   ln(Income)_R ln(Income)_T  ln(Income)^2";      
    paccprob';?; 
    
retp(pencfs0,mpr,AIMEbend,PIAbend,pbbend0,penacc0,paccprob); endp; 

@-----------------------------------------------------------------------------@
@-----------------------------------------------------------------------------@
/*  SSCREDIT:  Computes credit rates for early retirement and late retirement
               and earnings test.  Numbers taken from 2004 Green Book.
*/
proc(4) = sscredit(ruleyear);

    local piafrac, bigcreds, litcreds, eretadj, i, j, k, cred_j, sscreds;

    bigcreds = zeros(5,1);                                       @ Ages 62-66 @
    eretadj  = ones(5,1);                                        @ Ages 62-66 @
    litcreds = bigcreds;                                         @ Ages 65-69 @

   if ruleyear<1996;
        etstage = 70;
   elseif ruleyear==1996;
        etstage = 69;
   elseif ruleyear==1997;
        etstage = 68;
   elseif ruleyear==1998;
        etstage = 67;
   elseif ruleyear==1999;
        etstage = 66;
   else;
        etstage = nret;
   endif;

   @---piafrac gives PIA for ages 62-70 vis-a-vis PIA from retiring at nret---@
    if ruleyear==1995;
        piafrac = 0.8|(1-.4/3)|(1-.2/3)|1|1.045|1.09|1.135|1.18|1.225;
    elseif (ruleyear==1996) or (ruleyear==1997);
        piafrac = 0.8|(1-.4/3)|(1-.2/3)|1|1.05|1.1|1.15|1.2|1.25;
    elseif (ruleyear==1998) or (ruleyear==1999);
        piafrac = 0.8|(1-.4/3)|(1-.2/3)|1|1.055|1.11|1.165|1.22|1.275;
    elseif (ruleyear==2000) or (ruleyear==2001);
        piafrac = 0.8|(1-.4/3)|(1-.2/3)|1|1.06|1.12|1.18|1.24|1.3;
    elseif ruleyear==2004;
        piafrac = (1-.65/3)|(1-.45/3)|(1-.25/3)|(1-.2/9)|(1+.14/3)|(1+.35/3)|(1+.56/3)|(1+.77/3)|(1+.98/3);
    elseif ruleyear==2030;
        piafrac = 0.7|0.75|0.8|(1-.4/3)|(1-.2/3)|1|1.08|1.16|1.24;
    else;
        piafrac = ones(9,1);
    endif;

    i=62; do until i>69;
        j = i-62+1;
        cred_j = piafrac[j+1]/piafrac[j] - 1;
        if i<67;
            eretadj[j]  = 1/(1+cred_j);
            bigcreds[j] = cred_j;
            if (i>64) and (swapcred==1);
                bigcreds[j] = bigcreds[j-1];
                eretadj[j]  = eretadj[j-1];
            endif;
        endif;

        if i>64;
            litcreds[i-64] = cred_j;
        endif;
    i=i+1; endo;

    "Rule Year =             ";; ruleyear;
    "Normal Retirement age = ";; nret;
    "Earnings test ends at = ";; etstage;    
    "        Age      Bigcred       Eretadj";;
    seqa(62,1,5)~bigcreds~eretadj;?;
    "        Age      Litcred";;
    seqa(65,1,5)~litcreds;?;

    sscreds = bigcreds|eretadj|litcreds;

retp(bigcreds,eretadj,litcreds,sscreds); endp;

@-----------------------------------------------------------------------------@
@-----------------------------------------------------------------------------@
/*
    Tauch:    Converts an AR(1) process into a Markov chain             
              Follows Tauchen (Ec. Letters, 1986)                 
    Inputs:
        stdi:       Scalar, std. dev. of innovation
        rho:        Scalar, AR(1) correlation coefficient
        M:          Scalar indicating number of elements in the chain
        Tail:       Scalar indicating percentages left in each tail
    Output:
        pmtx:       MxM transition matrix
                    rows denote time t values, cols time t+1 values
        values:     Mx1 vector of values
*/

proc(3) = tauch(stdi,rho,m,utail,ltail);

    local v_min, v_max, stdy, pmtx, values, yold, ynew, i, j, w, fe, intvals;

    stdy    = sqrt(stdi^2/(1-rho^2));
    v_min   = cdfni(ltail)*stdy;
    v_max   = cdfni(1-utail)*stdy;

    pmtx    = zeros(M,M);
    values  = zeros(M,1);
    values  = seqa(v_min,(v_max-v_min)/(m-1),m);
    intvals = seqa((values[2,1]+values[1,1])/2,(v_max-v_min)/(m-1),m-1);

    i=1; do until i>m;
        yold = values[i,1];
        ynew = values[1,1];
        w = (values[2,1] - values[1,1])/2;
        fe = (ynew - rho*yold);
        pmtx[i,1] = cdfn((fe+w)/stdi);

        j=2; do until j>m-1;
            ynew = values[j,1];
            w = (ynew-values[j-1,1])/2;
            fe = ynew - rho*yold;         
            pmtx[i,j] = cdfn((fe+w)/stdi) - cdfn((fe-w)/stdi);
        j=j+1;  endo;

        ynew =values[m,1];
        w = (values[m,1] - values[m-1,1])/2;
        fe = (ynew - rho*yold);
        pmtx[i,m] = 1 - cdfn((fe-w)/stdi);
    i=i+1; endo;

retp(pmtx,values,intvals); endp;

@-----------------------------------------------------------------------------@
@-----------------------------------------------------------------------------@
/*
    Taukhuss:   Converts an AR(1) process into a Markov chain             
                Follows Tauchen and Hussey(Econometrica, 1991)                 
    Inputs:
        stdi:       Scalar, std. dev. of innovation
        rho:        Scalar, AR(1) correlation coefficient
        M:          Scalar indicating number of elements in the chain
    Output:
        pmtx:       MxM transition matrix
                    rows denote time t values, cols time t+1 values
        values:     Mx1 vector of values
*/

proc(2) = taukhuss(stdi,rho,m);

    local stdi2, stdy, pmtx, values, quadwgts, yold, ynew, i, j, dev, cpdf;

    stdi2 = sqrt(1-rho^2); @ Innovation variance for normalized process @
    pmtx  = zeros(M,M);
    {values,quadwgts} = ghquad(m,stdi2);

    i=1; do until i>m;
        yold = values[i];
        j=1; do until j>m;
            ynew = values[j];
            dev  = (ynew - rho*yold)/stdi2;
            cpdf = exp(-dev*dev/2.0)/sqrt(2.0*PI*stdi2*stdi2);
            pmtx[i,j] = cpdf*quadwgts[j];
            dev  = ynew/stdi2;
            cpdf = exp(-dev*dev/2.0)/sqrt(2.0*PI*stdi2*stdi2); @ T-H's omega(y) = f(y|0) @
            pmtx[i,j] = pmtx[i,j]/cpdf;
        j=j+1;  endo;
        pmtx[i,.] = pmtx[i,.]/sumc(pmtx[i,.]'); @ Normalize so that probs sum to 1 @
    i=i+1; endo;


    stdy = sqrt(stdi^2/(1-rho^2));
    values = values*stdy;

retp(pmtx,values); endp;

@-----------------------------------------------------------------------------@
@-----------------------------------------------------------------------------@
@   NEWDRAW:   Finds State according to Markov Chain and Uniformly           @
@              Distributed Random Number                                     @

proc(1) = newdraw(pmtxcdf,curstate,randnum);

    local newstate,probindx, z;
    probindx = pmtxcdf[curstate,.]';
    z = probindx.<= randnum; 
    newstate = z'z + 1; 
retp(newstate); endp;

@-----------------------------------------------------------------------------@
@-----------------------------------------------------------------------------@
/*
    DISCRTZE:  Converts continous values into weighted indices of discrete    
               categories.                                                    
*/
proc(2) = discrtze(cvals,states);

    local range, rind, types, frac, tfrac, i, n;

    cvals = cvals[.,1];
    n     = rows(states);
    types = ones(rows(cvals),1);
    frac  = types;

    range = (minc(cvals)-1)|states[2];
    rind  = indexcat(cvals,range); @ Indices of deviations that fall in range @
    if ismiss(rind)==0;
        types[rind,1] = ones(rows(rind),1);
        tfrac = cvals[rind,1];
        tfrac = (tfrac-states[1])/(states[2]-states[1]);
        frac[rind,1] = tfrac;
    endif;

    i=2; do until i >(n-2);
        range = states[i:i+1];    
        rind  = indexcat(cvals,range); 
        if ismiss(rind)==0;
            types[rind,1] = i*ones(rows(rind),1);
            tfrac = cvals[rind,1];
            tfrac = (tfrac-states[i])/(states[i+1]-states[i]);
            frac[rind,1] = tfrac;
        endif;
    i=i+1; endo;

    range = states[n-1]|(maxc(cvals)+1);
    rind = indexcat(cvals,range); 

    if ismiss(rind)==0;
        types[rind,1] = (n-1)*ones(rows(rind),1);
        tfrac = cvals[rind,1];
        tfrac = (tfrac-states[n-1])/(states[n]-states[n-1]);
        frac[rind,1] = tfrac;
    endif;

retp(types,frac); endp;
